package client.brickblock.fwimpl;

import client.brickblock.objects.Powerup;
import client.brickblock.objects.Trap;
import client.framework.models.AbstractGame;
import client.framework.models.AbstractPlayer;
import client.framework.views.GameView;

import javax.microedition.lcdui.Image;
import javax.microedition.lcdui.game.Sprite;
import java.util.Vector;

/**
 * This class is the implementation of the
 * abstract class <code>GameView</code>.
 * It is responsible for displaying the game
 * board with all its objects.
 *
 * When a player is pushed (or moves himself) into
 * the trap, he receives -1 points. Power up objects
 * are used; if a player picks up one such, he
 * receives a temporary bonus by getting bigger,
 * faster or stronger.
 *
 * @author Eivind Sorteberg
 * @author Martin Jarrett
 * @version 2.0
 */
public class BBBoard extends GameView {
    //The non-player objects in the game
    private Vector objects;

    /**
     * Constructor. Calls the super constructor,
     * and decides when the first power up object
     * should be placed on the game board.
     *
     * @param model The game's underlying model.
     */
    public BBBoard(AbstractGame model) {
        super(model);
        this.objects = new Vector();
    }

    /**
     * Adds an object to the board.
     *
     * @param object The object added to the board.
     */
    public void addObject(Sprite object) {
        this.objects.addElement(object);
        this.layerManager.insert(object, 0);
        this.fireBoardChanged();
    }

    /**
     * Removes a <code>PowerUp</code> object from the board at
     * the specified position. This method is called when a
     * power up times out, or when it is picked up by one of
     * the remote players.
     *
     * @param name The name of the power up to remove.
     * @param posX The x-position of the power up.
     * @param posY The y-position of the power up.
     * @return The power up containing the specified attributes
     */
    public Powerup findPowerup(String name, int posX, int posY) {
        for (int i = 0; i < objects.size(); i++) {
            Sprite object = (Sprite)objects.elementAt(i);
            if (object instanceof Powerup) {
                Powerup powerup = (Powerup)object;
                if (powerup.getX() == posX && powerup.getY() == posY && name.equals(powerup.getPowerUpName())) {
                    return powerup;
                }
            }
        }
        return null;
    }

    /* **********************************************************************
     * METHODS INHERITED FROM GAMEVIEW                                      *
     ************************************************************************/

    /**
     * Creates a background image that covers the entire
     * game board.
     *
     * @return The background image.
     */
    public Image createBackgroundImage() {
        int amount;
        double inc = 512.0 / getHeight();
        double counter = 0;
        int[] colors = new int[getHeight() * getWidth()];
        for (int row = 0; row < getHeight(); row++) {
            amount = (int)(Math.abs((255.0 - counter)));
            counter += inc;
            int red = 255 << 16;
            int green = 255 << 8;
            int blue = ((255 - amount) + 128) / 2;
            for (int column = 0; column < getWidth(); column++) {
                colors[row * getWidth() + column] = red | green | blue;
            }
        }
        return Image.createRGBImage(colors, getWidth(), getHeight(), false);
    }

    /**
     * Called if a collision is detected with another player.
     * Detects whether the local player is strong enough to
     * push the other player. If he is as strong as, or stronger
     * than the other player, he pushes the other player with
     * with his own movement speed.
     *
     * After the other player has been pushed, all participants
     * are notified.
     *
     * @param collidesWith The player with whom a collision is detected.
     * @param movement The move that leads to the collision.
     * @see GameView#detectPlayerCollision(client.framework.models.AbstractPlayer, int[])
     * @see client.framework.views.GameView#check()
     */
    protected void handlePlayerCollision(AbstractPlayer collidesWith, int[] movement)
    {
        //Gets the local player
        BBPlayer self = (BBPlayer)model.getSelf();

        //Gets the collidee
        BBPlayer other = (BBPlayer)collidesWith;

        //Determined the strength ratio
        double ratio = (double)self.getStrength() / (double)other.getStrength();

        //Can push if he is strong enough
        if (ratio >= 1) {
            movement[0] = (int)((double)movement[0] * ratio);
            movement[1] = (int)((double)movement[1] * ratio);

            // If the push does not result in the pushed player colliding with the wall, allow the push
            if (detectWallCollision(collidesWith, movement) == 0) {
                model.fireSelfMoved(movement);

                //Notify that a force move has occured
                ((BrickBlock)model).fireForceMove(other, movement);
            }
        }
    }

    /**
     * Called if a collision is detected with a game object
     * (<code>PowerUp</code> or <code>Trap</code>). If such a collision
     * is detected, the server is notified.
     *
     * @param collidesWith The object with which the player collides.
     * @param movement The move that leads to the collision.
     * @see client.framework.views.GameView#detectObjectCollision(client.framework.models.AbstractPlayer, int[])
     * @see client.framework.views.GameView#check()
     */
    public void handleObjectCollision(Object collidesWith, int[] movement) {
        //Gets the local player
        BBPlayer self = (BBPlayer)model.getSelf();

        // NOtify the model if the object is a trap
        if (collidesWith instanceof Trap) {
            this.model.fireHitTrap();
        }
        // Notify the model if the object is a power up
        else if (collidesWith instanceof Powerup) {
            ((BrickBlock)model).firePowerupPickedUp((Powerup)collidesWith);
        }
        // Keep moving to keep the game fluent
        self.move(movement[0], movement[1]);
    }

    /**
     * Called continously the game from the <code>check()</code>
     * method in <code>GameView</code>.
     *
     * In this implementation, nothing is performed in this method,
     * as all necessary client checks are performed in the default
     * <code>check()</code> method.
     *
     * @see client.framework.views.GameView#check()
     */
    protected void specialCheck() {
    }

    /**
     * Called whenever the local player moves, to check whether
     * the move causes a collision with one of the game objects.
     * If so, the <code>handleObjectCollision()</code> method
     * is called, and appropriate actions are taken. If no
     * collision is detected, the method returns null.
     *
     * @see BBBoard#handleObjectCollision(Object, int[])
     * @see client.framework.views.GameView#check()
     * @param player The player trying to move
     * @param movement The movement to take place.
     * @return The object with which the player collides.
     */
    public Object detectObjectCollision(AbstractPlayer player, int[] movement) {
        Sprite sprite = player.getSprite();
        sprite.move(movement[0], movement[1]);  // Simulate the move
        Object collidesWith = null;

        //Checks all the game objects
        for (int i = 0; i < objects.size(); i++) {
            Sprite object = (Sprite)objects.elementAt(i);
            if (sprite.collidesWith(object, true)) {
                collidesWith = object;
                break;
            }
        }
        sprite.move(-movement[0], -movement[1]);    // Reset the move
        return collidesWith;
    }

    /**
     * Empties all the class' object lists. Used when a game is over, to
     * ensure that no objects are transferred to the next game.
     */
    protected void clean() {
        for (int i = 0; i < objects.size(); i++) {
            this.layerManager.remove((Sprite)objects.elementAt(i));
        }
        this.objects.removeAllElements();
    }

    /**
     * Removes the specified <code>PowerUp</code> object
     * from the game board. The object is removed from
     * both the displayed screen, as well as the list
     * of game objects.
     *
     * @param powerup The power up to remove.
     */
    public void removePowerup(Powerup powerup) {
        this.layerManager.remove(powerup);
        this.objects.removeElement(powerup);
    }

    /**
     * Method that is called whenever the server has notified that the
     * local player has been pushed. Checks if the push results in the
     * user colliding with a game object (i.e. trap or power up). If so,
     * the server needs to be notified of this event. If not, the
     * player is simply moved.
     *
     * @param movement The movement with which the local player has been pushed.
     * @see BrickBlock#fireForcePushed(String[])
     */
    public void handleSelfPushed(int[] movement) {
        // If the local player is pushed onto an object, the collision needs to be handled
        Object collidesWith = detectObjectCollision(model.getSelf(), movement);
        if (collidesWith != null) {
            handleObjectCollision(collidesWith, movement);
        }
        else {
            model.getSelf().move(movement[0], movement[1]);
        }
    }
}
