package client.brickblock.fwimpl;

import client.brickblock.objects.Powerup;
import client.framework.models.AbstractPlayer;

import java.util.Vector;

/**
 * BrickBlock's <code>AbstractPlayer</code> implemention.
 * In addition to the attributes and functionality found
 * in <code>AbstractPlayer</code>, this class contains
 * information about the player's strength and screen
 * resolution. In addition, this class offers
 * functionality for handling a player's bonuses in terms
 * of active <code>Powerup</code> objects.
 *
 * @author Martin Jarrett
 * @author Eivind Sorteberg
 * @version 2.0
 */
public class BBPlayer extends AbstractPlayer {
    //The player's current strength
    private int strength;

    //The player's active powerups
    private Vector powerups;

    /**
     * Constructor for <code>BBPlayer</code>. This contructor takes
     * the player's strength as input parameter in addition
     * to the parameters from the super constructor.
     *
     * @param id The player's network id.
     * @param name The player's name.
     * @param strength The player's strength.
     */
    public BBPlayer(String id, String name, int strength) {
        super(id, name);

        this.strength = strength;
        this.powerups = new Vector();
    }

    /**
     * Constructor that corresponds to the super constructor.
     * This creates a <code>BBPlayer</code> objects with a default
     * strength of 2.
     *
     * @param id The player's id.
     * @param name The player's name.
     */
    public BBPlayer(String id, String name) {
        this(id, name, 2);
    }

    /**
     * Yet another constructor that creates a nameless player.
     *
     * @param id The player's id.
     */
    public BBPlayer(String id) {
        this(id, "");
    }

    /**
     * Returns the player's strength.
     *
     * @return The player's strength.
     */
    public int getStrength() {
        return this.strength;
    }

    /**
     * Sets the player's strength to the value given as
     * input parameter.
     *
     * @param strength The player's strength.
     */
    public void setStrength(int strength) {
        this.strength = strength;
    }

    /**
     * Called when the player picks up a <code>Powerup</code>
     * object. The power up is activated, and remains active
     * until the server notifies that the power up has been
     * deactivated.
     *
     * @param powerup The <code>Powerup</code> picked up.
     * @see BBBoard#handleObjectCollision(Object, int[])
     */
    public void  pickupPowerup(Powerup powerup) {
        powerup.activate(this);
        powerups.addElement(powerup);
    }

    /**
     * Deactivates the specified power up and removes it from the
     * player object's list of active power ups.
     *
     * Called when the server has determined that the power up
     * has timed out.
     *
     * @param powerup The power up that has been deactivated.
     * @see BrickBlock#notifyAboutSpecialActionReceived(client.framework.models.AbstractPlayer, String, String[][])
     */
    private void removePowerup(Powerup powerup) {
        powerup.deactivate(this);
        powerups.removeElement(powerup);
    }

    /**
     * Deactivates the power up indicated by given attributes
     * and removes it from the player object's list of active power ups.
     *
     * Called when the server has determined that the power up
     * has timed out.
     *
     * @param type The power up type.
     * @param posX The horizontal position of the power up.
     * @param posY The vertical position of the power up.
     * @see BrickBlock#notifyAboutSpecialActionReceived(client.framework.models.AbstractPlayer, String, String[][])
     * @see client.brickblock.objects.Powerup#getPowerUpName()
     */
    public void removePowerup(String type, int posX, int posY) {
        this.removePowerup(this.findPowerup(type, posX, posY));
    }

    /**
     * Finds the power up specified by the given attributes in the player's list of
     * active power up objects.
     *
     * @param type The type of power up.
     * @param posX The horizonal position of the power up.
     * @param posY The vertical position of the power up.
     * @return The power up containing the specified attributes.
     * @see client.brickblock.objects.Powerup#getPowerUpName()
     */
    private Powerup findPowerup(String type, int posX, int posY) {
        for (int i = 0; i < this.powerups.size(); i++) {
            Powerup powerup = (Powerup)powerups.elementAt(i);
            if (type.equals(powerup.getPowerUpName()) && posX == powerup.getX() && posY == powerup.getY()) {
                return powerup;
            }
        }
        return null;
    }
}