package client.brickblock.fwimpl;

import client.brickblock.objects.*;
import client.framework.models.AbstractGame;
import client.framework.models.AbstractPlayer;
import client.framework.views.GameView;

import java.io.IOException;
import java.util.Vector;

/**
 * The implementation of the abstract class <code>AbstractGame</code>,
 * which contains the most important data for running a game in
 * the framework framework.
 *
 * In addition to the default features offered by
 * <code>AbstractGame</code>, this class contain support for
 * sending and receiving information about placing and picking
 * up <code>PowerUp</code> objects.
 *
 * @author Martin Jarrett
 * @author Eivind Sorteberg
 * @version 2.0
 */
public class BrickBlock extends AbstractGame {
    /**
     * A list containing the players that have been pushed, but not yet acknowledged
     * by the server. Used to avoid "double pushing".
     */
    private Vector pushed;

    /**
     * Initialises the framework.
     */
    public void startFramework(String name, String[] addresses, int packetSize) {
        super.startFramework(name, addresses, packetSize);
        this.pushed = new Vector();
    }

    /**
     * Returns a new <code>BBPlayer</code> instance, which is this
     * game's implementation of the <code>AbstractPlayer</code>
     * class.
     *
     * @param address The player's address.
     * @param name The player's name.
     * @return A new <code>BBPlayer</code> instance.
     */
    public AbstractPlayer newPlayer(String address, String name) {
        return new BBPlayer(address, name);
    }

    /**
     * Returns a new <code>GameView</code> instance, which is this
     * game's implementation of the <code>GameView</code> class.
     *
     * @return A new <code>BBBoard</code> using this as its model.
     */
    protected GameView createGameView() {
        return new BBBoard(this);
    }

    /**
     * Returns the active game board used in the game.
     *
     * @return The active game board.
     * @see client.brickblock.fwimpl.BrickBlock#createGameView()
     */
    public BBBoard getGameBoard() {
        return (BBBoard)game;
    }

    /**
     * Returns the name of this game; "BrickBlock".
     *
     * @return A String containing "BrickBlock".
     */
    public String getGameName() {
        return "BrickBlock";
    }

    /* METHODS FOR TRANSMITTING INFO */

    /**
     * Called when the local player picks up a power up. A notification of this
     * is created using the parser and transmitted to the server using the
     * active <code>Communicator</code> object.
     *
     * @param powerup The power up picked up or timed out
     * @see client.framework.network.parser.MessageParser#createMessage(client.framework.network.parser.Action, Object[][])
     * @see client.framework.network.Communicator#sendMessage(Object, boolean)
     */
    public void firePowerupPickedUp(Powerup powerup) {
        Object message = parser.createMessage(BBAction.POWERUP_REMOVED, new Object[][]{powerup.getValues()});
        this.communicator.sendMessage(message, true);
    }

    /**
     * Called when the local player "force moves" (pushes)
     * another player. When this occurs, how much the other player
     * is pushed is transmitted to the server using the active parser
     * and communicator objects.
     *
     * @param other The player that is pushed.
     * @param movement How much the player is pushed.
     * @see client.framework.network.parser.MessageParser#createMessage(client.framework.network.parser.Action, Object[][])
     * @see client.framework.network.Communicator#sendMessage(Object, boolean)
     */
    public void fireForceMove(AbstractPlayer other, int[] movement) {
        pushed.addElement(other.getId());
        other.move(movement[0], movement[1]);
        Object message = parser.createMessage(BBAction.FORCE, new Object[][]{{
                other.getId(), new Integer(movement[0]), new Integer(movement[1])
        }});
        this.communicator.sendMessage(message, true);
    }

    /**
     * This method is responsible for parsing the incoming actions
     * that can not be parsed by the <code>notifyAboutMessageReceived()</code> method in
     * <code>AbstractGame</code>.
     *
     * These actions are related to either positioning power up or trap
     * objects on the board, players picking up power up
     * objects, power ups being deactivated, or one player pushing another on the game
     * board. Dependent of what kind of action is received,
     * different methods are called.
     *
     * @see client.framework.models.AbstractGame#notifyAboutMessageReceived(String)
     * @param sender The sender of the action.
     * @param action The action that has been performed.
     * @param values String containing a single value or a composite value.
     */
    protected void notifyAboutSpecialActionReceived(AbstractPlayer sender, String action,
                                             String[][] values) {

        //Cast to the right kind of object
        BBPlayer player = (BBPlayer)sender;

        // Either the action is power up related ..
        if (BBAction.POWERUP_ADDED.equals(action) || BBAction.POWERUP_REMOVED.equals(action) || BBAction.POWERUP_INACTIVE.equals(action))  {
            //Find the type of power up and its position
            String type = values[0][0];
            int posX = Integer.parseInt(values[0][1]);
            int posY = Integer.parseInt(values[0][2]);

            // If a power up is added to the game board..
            if (BBAction.POWERUP_ADDED.equals(action)) {
                int increment = Integer.parseInt(values[0][3]);
                //Add the power up to the game board
                this.getGameBoard().addObject(Powerup.createPowerup(type, posX, posY, increment));
            }
            // If a power up is removed; either by timing out or being picked up
            else if (BBAction.POWERUP_REMOVED.equals(action)) {
                Powerup powerup = this.getGameBoard().findPowerup(type, posX, posY);
                if (powerup != null) {
                    if (player != null) { // Activate the power up for the affected player
                        player.pickupPowerup(powerup);
                    }
                    this.getGameBoard().removePowerup(powerup);
                }
            }
            // Final option: the power up shall be deactivated for the specified player
            else {
                player.removePowerup(type, posX, posY);
            }
        }
        // .. or a player has been force pushed ..
        else if (BBAction.FORCE.equals(action)) {
            fireForcePushed(values[0]);
            game.fireBoardChanged();
        }
        // .. or a trap has been ad
        else if (BBAction.TRAP_ADDED.equals(action)) {
            int xPos = Integer.parseInt(values[0][0]);
            int yPos = Integer.parseInt(values[0][1]);
            try {
                this.getGameBoard().addObject(new Trap(xPos, yPos));
            }
            catch (IOException e) {
                e.printStackTrace();
            }
        }
    }

    /**
     * Not implemented.
     */
    protected void createSpecialSettings() {}

    /**
     * Not implemented.
     *
     * @param init Whether the check is performed on game initialisation.
     * @return Always true.
     */
    protected boolean handleSpecialSettings(boolean init) {
        return true;
    }

    /**
     * Called when a notification arrives that another
     * player has been pushed. The player is then moved
     * according to the force value. If this player is in the list of recently
     * pushed players, he can be removed from this list.
     *
     * @param value The player that has been pushed, and how much he has been pushed.
     */
    private void fireForcePushed(String[] value) {
        //Get the player that has been pushed
        String playerId = value[0];

        if (pushed.contains(playerId)) {
            // Remove the player from the pushed list
            pushed.removeElement(playerId);
        }
        else {
            //Get the movement
            int[] movement = new int[]{
                    Integer.parseInt(value[1]),
                    Integer.parseInt(value[2])
            };

            // Find out who has been pushed
            AbstractPlayer player = getPlayer(playerId);

            if (player == getSelf()) {
                // Must check for collisions if it is the local player that has been pushed
                ((BBBoard)game).handleSelfPushed(movement);
            }
            else {
                // If the player is not the local player, the movement can be continued
                player.move(movement[0], movement[1]);
            }
        }
    }
}