package client.brickblock.objects;

import client.brickblock.fwimpl.BBPlayer;
import client.framework.Utils;

import javax.microedition.lcdui.Image;
import javax.microedition.lcdui.game.Sprite;
import java.io.IOException;

/**
 * Abstract class used for power up objects. All
 * power up objects have an increment value that determines
 * the effect of the power up,
 * and can be activated and deactivated.
 *
 * This class extends the <code>Sprite</code>, and
 * power up objects can therefore be placed directly
 * on the game board (by the <code>LayerManager</code>.
 *
 * @author Martin Jarrett
 * @author Eivind Sorteberg
 * @version 2.0
 */
public abstract class Powerup extends Sprite {
    /**
     * The value with which the player's affected value is incremented
     */
    protected int increment;

    /**
     * Constructor that assigns an image to the power up
     * object, as well as defining a position and duration.
     *
     * @param imageAddress The location of the image to use.
     * @param xPos The object's x-position on the screen.
     * @param yPos The object's y-position on the screen.
     * @param increment The increment of the power up.
     * @throws IOException If the image file is not found.
     */
    protected Powerup(String imageAddress, int xPos, int yPos, int increment) throws IOException {
        super(Image.createImage(imageAddress));
        this.setPosition(xPos, yPos);
        this.increment = increment;
    }

    /**
     * "Turns on" the power up by increasing one of the player's
     * attributes. How this power up affects the player is
     * dependent on how this method is implemented.
     *
     * @param player The player that is affected by the power up.
     */
    public abstract void activate(BBPlayer player);

    /**
     * "Turns off" the power up by restoring the player's
     * attribute to the value it would have had if the power
     * up had not been picked up.
     *
     * @param player The player that is affected by the power up.
     */
    public abstract void deactivate(BBPlayer player);

    /**
     * Returns a three letter String representation of the power up object.
     *
     * @return The name of the power up object
     */
    public abstract String getPowerUpName();

    /**
     * Creates a power up object with the given attributes. The type
     * of power up object is specified by the three letter name
     * specified in the <code>type</code> attribute. When new implentors
     * of this class is created, the list of types in this method needs
     * to be updated.
     *
     * @param type The type of power up to create.
     * @param xPos The horizontal position of the object.
     * @param yPos The vertical position of the object.
     * @param value The increment value of the object.
     * @return A power up object with the specified attributes.
     */
    public static Powerup createPowerup(String type, int xPos, int yPos, int value) {
        try {
            if (type.equals("SPD")) return new SpeedUp(xPos, yPos, value);
            else if (type.equals("SIZ")) return new SizeUp(xPos, yPos, value);
            else if (type.equals("STR")) return new StrengthUp(xPos, yPos, value);
        } catch (IOException e) {
            /* Do nothing, not likely to happen */
            return null;
        }
        // May occur if new power up objects are added, but not included in this list.
        System.err.println(Utils.buildString(new Object[]{"Illegal: ", type, ", ", new Integer(xPos), ", ", new Integer(yPos), ", ", new Integer(value)}));
        return null;
    }

    /**
     * Returns the attributes of the power up object. Is used to
     * provide a parser with object values.
     *
     * @return An array containing the object's attributes.
     */
    public Object[] getValues() {
        return new Object[]{
                this.getPowerUpName(),
                new Integer(this.getX()),
                new Integer(this.getY())
        };
    }
}
