package client.framework;

import client.framework.models.AbstractGame;
import client.framework.models.AbstractPlayer;
import client.framework.models.Comparable;

import javax.microedition.lcdui.Graphics;
import javax.microedition.lcdui.Image;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

/**
 * This class contains utilities used in different parts of the
 * framework. These utilities are static methods that offer
 * functionality that may be needed several times.
 *
 * The utilities are: sorting of a list, drawing list of players
 * to a graphics object, drawing message list, drawing list
 * of settings and creating an image.
 *
 * @author Eivind Sorteberg
 * @author Martin Jarrett
 * @version 2.0
 */
public abstract class Utils {
    //The height of each line of text
    public static final int LINE_HEIGHT = 15;

    /**
     * Sorts a list of comparable elements so that the element
     * with the highest value is returned at the start of the
     * list and the element with the lowest value is returned
     * at the end (descending).
     *
     * @see Comparable
     * @param list The list to sort. All elements must implement the <code>Comparable</code> interface.
     * @return A list sorted descending.
     */
    public static Vector sort(final Hashtable list) {
        Vector result = new Vector(list.size());

        Enumeration keys = list.keys();
        //Add the first element to the new list
        if (list.size() > 0) result.addElement(list.get(keys.nextElement()));

        //For all the remaining elements..
        while (keys.hasMoreElements()) {
            Comparable object = (Comparable)list.get(keys.nextElement());

            boolean added = false;
            //..compare to the elements in the new list..
            for (int j = 0; j < result.size(); j++) {
                //..and insert at the correct position..
                if (((Comparable)result.elementAt(j)).compareTo(object) < 0) {
                    result.insertElementAt(object, j);
                    added = true;
                    break;
                }
            }
            //..or at the end
            if (!added) result.addElement(object);
//            result.addElement(list.get(keys.nextElement()));
        }
        return result;
    }

    /**
     * Sorts the given list by the players' assigned teams. This method
     * does not consider the players' score, but only groups the players
     * according to their team. In addition, the method inserts an empty
     * "team" player representing the entire team before that team's players.
     *
     * @param playerlist The list of players.
     * @param model The client's main model.
     * @return A list sorted by teams.
     */
    private static Vector createTeamList(final Vector playerlist, AbstractGame model) {
        int[][] teams = AbstractGame.TEAMCOLORS;
        Vector result = new Vector(playerlist.size() + teams.length);
        AbstractPlayer team;
        AbstractPlayer player;

        for (int team_no = 0; team_no < teams.length; team_no++) {
            team = model.newPlayer("", Utils.buildString(new Object[]{"Team #", new Integer(team_no + 1)}));
            team.setColor(teams[team_no][0], teams[team_no][1], teams[team_no][2]);
            result.addElement(team);
            int nofPlayers = playerlist.size();
            int teamScore = 0;
            for (int player_no = 0; player_no < nofPlayers; player_no++) {
                player = (AbstractPlayer)playerlist.elementAt(player_no);
                if (player.getTeam() == team_no) {
                    teamScore += player.getScore().intValue();
                    result.addElement(player);
                }
            }
            team.setScore(teamScore);
        }
        return result;
    }

    /**
     * Draws a list of players to the provided graphics object.
     * This method is used by both <code>GameView</code> and
     * <code>LobbyView</code> to draw their lists of players.
     * If the game is drawn in the lobby view, indicators to show
     * whether the players are ready or not are also drawn.
     *
     * If teams are enabled, the player list is first sorted according to
     * the teams, and the players within each team are sorted by their score.
     *
     * @param graphics The graphics object to use for the drawing.
     * @param screenWidth The available screen width.
     * @param screenHeight The available screen height.
     * @param model The game's model.
     * @param inGame Whether the drawing is performed in-game or not.
     */
    public static void drawPlayerList(Graphics graphics, int screenWidth, int screenHeight, AbstractGame model, boolean inGame) {
        int readySize = 7;  //The size of the ready indicator

        Vector sortedlist = sort(model.getPlayerlist());

        if (model.isTeamsEnabled()) {
            sortedlist = createTeamList(sortedlist, model);
        }
        int line = 0;
        int player_no;
        boolean teamsEnabled = model.isTeamsEnabled();

        //For each player..
        for (player_no = 0; player_no < sortedlist.size(); player_no++) {
            //To ensure there is room for the list
            if (LINE_HEIGHT * line + LINE_HEIGHT < screenHeight) {
                int offset = 2;     // The left side margin for text

                //Get the player..
                AbstractPlayer player = (AbstractPlayer)sortedlist.elementAt(player_no);

                if (teamsEnabled) {
                    if (line != 0 && player.getName().startsWith("Team #")) {
                        line++;
                    }
                }

                //..and the color
                int r = player.getColor()[0];
                int g = player.getColor()[1];
                int b = player.getColor()[2];

                //Drawing lines in the player's color if in the lobby
                graphics.setColor(r, g, b);
                if (!inGame) graphics.fillRect(0, line * LINE_HEIGHT, screenWidth, LINE_HEIGHT);

                //Drawing "ready" indicators
                if (!inGame) {
                    boolean ready = player.isReady();
                    graphics.setColor(ready ? 0 : 255, ready ? 255 : 0, 0);
                    graphics.fillArc((LINE_HEIGHT - readySize) / 2, line * LINE_HEIGHT + (LINE_HEIGHT - readySize) / 2, readySize, readySize, 0, 360);

                    //Drawing "ready" outlines
                    graphics.setColor(0, 0, 0);
                    graphics.drawArc((LINE_HEIGHT - readySize) / 2, line * LINE_HEIGHT + (LINE_HEIGHT - readySize) / 2, readySize, readySize, 0, 360);

                    //Inverting the color for the player name
                    r = 255 - r;
                    g = 255 - g;
                    b = 255 - b;

                    offset += readySize * 2;
                }

                //Drawing player names and scores
                graphics.setColor(r, g, b);

                //Add a " (Me)" at the end of the local player
                String name = player.getName();
                if (player.equals(model.getSelf())) name = Utils.buildString(new Object[]{name, " (Me)"});

                graphics.drawString(name, offset, line * LINE_HEIGHT, 0);
                graphics.drawString(buildString(new Object[]{"[", player.getScore(), "]"}), screenWidth - 2, line * LINE_HEIGHT, Graphics.TOP | Graphics.RIGHT);
            }
            line++;
        }
    }

    /**
     * Creates a monocolored <code>Image</code> object.
     * This method is used to create user sprites and
     * background images.
     *
     * @param width The width of the image.
     * @param height The height of the image.
     * @param red The amound of red.
     * @param green The amount of green.
     * @param blue The amount of blue.
     * @return An <code>Image</code> object.
     */
    public static Image createImage(int width, int height, int red, int green, int blue) {
        Image image = Image.createImage(width, height);
        Graphics g = image.getGraphics();
        g.setColor(red, green, blue);
        g.fillRect(0, 0, width, height);
        return image;
    }

    /**
     * Builds a string using a <code>StringBuffer</code> object. This
     * method is faster than using the '+' operator when building many
     * and large strings.
     *
     * @param objects The contents of the string.
     * @return A string object with the specified content.
     */
    public static String buildString(Object[] objects) {
        StringBuffer buffer = new StringBuffer();
        for (int i = 0; i < objects.length; i++) {
            Object o = objects[i];
            buffer.append(o.toString());
        }
        return buffer.toString();
    }
}
