package client.framework.models;

import client.brickblock.fwimpl.BBPlayer;
import client.framework.network.ClientSocketCommunicator;
import client.framework.network.Communicator;
import client.framework.network.parser.Action;
import client.framework.network.parser.ActionTriplet;
import client.framework.network.parser.MessageParser;
import client.framework.network.parser.StringParser;
import client.framework.views.*;
import client.framework.Utils;

import javax.microedition.lcdui.Display;
import javax.microedition.lcdui.Displayable;
import javax.microedition.lcdui.Alert;
import javax.microedition.lcdui.AlertType;
import javax.microedition.midlet.MIDlet;
import javax.microedition.midlet.MIDletStateChangeException;
import java.io.IOException;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

/**
 * This class is the main model for the framework. It contains the state
 * information for the game, and is responsible for displaying the correct
 * displayable element at all times.
 *
 * In addition, this class is responsible for communicating using a
 * <code>Communicator</code>, object as well as handling the actions received
 * from the <code>Communicator</code>.
 *
 * This class is the framework's implementation of MIDlet, and
 * contains the <code>startApp</code> method, which is a J2ME
 * application's main method.
 *
 * Finally, this class contains the two threads responsible for the running
 * of a game in the framework.
 *
 * @author Eivind Sorteberg
 * @author Martin Jarrett
 * @version 2.0
 * @noinspection MissedExecutable
 */
public abstract class AbstractGame extends MIDlet {
    // How often the game board is updated locally
    private static final int GAME_FPS = 50;

    // Whether or not the application uses movement prediction
    public static final boolean MOVEMENT_PREDICTION = true;

    //Message separators
    public static final String SEP_ACTION = ":";

    // The team colors: red and blue
    public static final int[][] TEAMCOLORS = {{255, 0, 0}, {0, 0, 255}};

    // The model's message parser instance
    protected MessageParser parser;

    // Whether or not teams are enabled (the only setting handled by the client)
    private static final String TEAMS_ON = "Teams enabled";

    // How often periodic information is transmitted to other participants
    private final int sendInterval = 100;

    // The communicator object
    protected Communicator communicator;

    // The representation of the local player's player object
    protected AbstractPlayer self;

    // Lists containing the players and settings for this game
    private Hashtable players;
    private Hashtable settings;

    /** @noinspection MismatchedQueryAndUpdateOfCollection*/
    // Message list used for debugging purposes
    private Vector messages;

    //The two main views
    protected LobbyView lobby;
    protected GameView game;
    
    //Threads used for local updating and networking
    private LocalThread localThread;
    private PositionThread positionThread;
    private Vector sessions;
    private SessionWindow sessionWindow;
    private long timeOfDeath;

    /**
     * Called after the player has specified a name.
     * Initializes all lists, displays the lobby view and
     * initialises the communication object.
     *
     * @param name The name of the local player.
     * @param addresses The address(es) of the server(s) to connect to
     * @param packetSize The maximum packet size (when using UDP)
     */
    public void startFramework(String name, String[] addresses, int packetSize) {
        //Initializes the list of players
        this.players = new Hashtable();

        //Initializes the list of messages to display
        this.messages = new Vector();

        this.sessions = new Vector();

        this.settings = new Hashtable();

        //Creates the lobby view
        this.lobby = new LobbyView(this);

        this.game = createGameView();

        this.sessionWindow = new SessionWindow(this);
        this.setView(this.sessionWindow);

        //Creates the local player
        this.self = newPlayer(-1 + "", name);
        this.addPlayer(self);

        //Gets the screen resolution of the local phone
        self.setScreenResolution(new int[]{this.game.getWidth(), this.game.getHeight()});

        // Creates the message parser instance
        this.parser = new StringParser();

        //Initializes the communicator object
        this.communicator = new ClientSocketCommunicator(this, packetSize);
        this.communicator.connect(addresses);

        this.timeOfDeath = 0;
    }

    /**
     * Creates a game view of the implementing game's specified type, and
     * returns this.
     *
     * @return The game view that the game uses to draw its game objects.
     */
    protected abstract GameView createGameView();

    /**
     * Returns the model's <code>MessageParser</code> implementation.
     *
     * @return An implementation of the <code>MessageParser</code> interface.
     */
    public MessageParser getParser() {
        return parser;
    }

    /**
     * Returns a list containing the players currently
     * connected to the game.
     *
     * @return The players currently in the game.
     */
    public Hashtable getPlayerlist() {
        return this.players;
    }

    /**
     * Returns the list of active settings, where each setting is
     * of the form <NAME>:<VALUE>.
     *
     * @return The settings for the game.
     */
    public Hashtable getSettings() {
        return this.settings;
    }

    /**
     * Adds a setting to the game. All settings must be of the form
     * <NAME>:<VALUE>, where the name is the name of the setting. The name is
     * used as the setting's ID, so each setting must have a unique name.
     * The value can be either a String, an Integer or a Boolean object.
     *
     * @param setting The name of the new setting.
     * @param value The value of the new setting.
     */
    public void addSetting(String setting, Object value) {
        this.settings.put(setting, value);
    }

    /**
     * Returns the local player representation.
     *
     * @return The local player object.
     */
    public AbstractPlayer getSelf() {
        return this.self;
    }

    /**
     * Adds a player to the player list.
     *
     * @param player The new player.
     */
    public void addPlayer(AbstractPlayer player) {
            this.players.put(player.getId(), player);
            this.firePlayerListChanged();
            if (this.game.isRunning()) {
                this.game.notifyAboutPlayerAdded(player);
            }
    }

    /**
     * Removes one of the game participants from the list
     * of players, and from the game board if a game is running
     *
     * @param player The player to remove.
     */
    public void removePlayer(AbstractPlayer player) {
        this.players.remove(player.getId());
        if (this.game != null) {
            this.game.removePlayer(player);
        }
        this.firePlayerListChanged();
    }

    /**
     * Returns the number of players participating in the
     * game.
     *
     * @return The size of the player vector.
     */
    public int getNumberOfPlayers() {
        return this.players.size();
    }

    /**
     * Shows a message for a given amount of time, before changing to another screen.
     *
     * @param title The title of the message to display.
     * @param message The message to display.
     * @param duration_secs How long the message is displayed.
     * @param nextScreen The screen to display after the message has been shown.
     */
    public void showAlert(String title, String message, int duration_secs, Displayable nextScreen) {
        Alert alert = new Alert(title, message, null, AlertType.ERROR);
        alert.setTimeout(duration_secs * 1000);
        Display.getDisplay(this).setCurrent(alert, nextScreen);
    }

    /**
     * Changes the view currently displayed to the one specified.
     * This method is used to toggle between game, lobby and
     * list views.
     *
     * @param view The view to display.
     */
    public void setView(Displayable view) {
        Display.getDisplay(this).setCurrent(view);
    }

    /**
     * Returns the currently displayed view.
     * @return The currently displayed view.
     */
    public Displayable getView() {
        return Display.getDisplay(this).getCurrent();
    }

    /**
     * Method fired when the list of messages is updated. This list
     * is mostly used for debugging information. Currently, the list is
     * not in use, but it can be used to print messages to work around
     * the problem of mobile phones not having a console.
     *
     * @param message The message to display.
     * @param append Whether the message should append or overwrite previous messages
     */
    public void fireMessagesChanged(String message, boolean append) {
        if (!append) this.messages.removeAllElements();

        this.messages.addElement(message);
    }

    /* MENU SELECTIONS */

    /**
     * Method called when the user selects "Ready" in the game lobby.
     * Toggles ready status on or off, and notifies all other users
     * in the network.
     */
    public void fireReadySelected() {
        this.self.setReady(!self.isReady());
        this.firePlayerListChanged();

        String ready = self.isReady() ? "true" : "false";
        Object message = parser.createMessage(Action.READY, new Object[][]{new Object[]{ready}});
        this.communicator.sendMessage(message, true);
    }

    /**
     * Method called when the user selects "Change settings" in the
     * game lobby. A screen for modifying the current settings is
     * then displayed.
     *
     * @see client.framework.views.SettingsWindow
     */
    public void fireChangeSettingsSelected() {
        SettingsWindow settings = new SettingsWindow(this);
        this.setView(settings);
    }

    /* ***********************/

    /**
     * Method called after returning from the "Change Settings"
     * screen. Checks whether any of the settings have been changed.
     * If so, the new values from the settings are transmitted to
     * all participants.
     *
     * @param settings The values of the settings.
     */
    public void fireSettingsUpdated(Hashtable settings) {
        Enumeration names = settings.keys();

        while (names.hasMoreElements()) {
            String name = (String) names.nextElement();
            Object value = settings.get(name);

            //If a setting has been changed, the new value is transmitted
            if (fireSettingChanged(name, value)) {
                this.sendSetting(name, value);
            }
        }
        this.viewLobby();
    }

    /**
     * Method called to set the lobby view as the active view.
     */
    public void viewLobby() {
        this.setView(lobby);
    }

    /**
     * Method called to make the phone flash and vibrate.
     */
    public void flashAndVibrate() {
        Display.getDisplay(this).vibrate(50);
        Display.getDisplay(this).flashBacklight(100);
    }

    /**
     * Method called when the local player scores. The players score
     * is updated, and all participants are notified about the event.
     */
    public void fireHitTrap() {
        long currentTime = System.currentTimeMillis();
        if (System.currentTimeMillis() > this.timeOfDeath) {
            this.sendScore();
            this.firePlayerListChanged();
            this.timeOfDeath = currentTime + 2000;
        }
    }

    /**
     * Method called when one of the players has signalled that he is ready
     * or not ready. The player's status is then updated.
     *
     * @param player The player sending the signal.
     * @param value "true" if the player is ready, "false" if not.
     */
    public void firePlayerReady(AbstractPlayer player, String value) {
        player.setReady("true".equals(value));
        this.firePlayerListChanged();
    }

    /**
     * Called when the player selects "start game" in the lobby menu.
     * Controls whether all players have signalled they are ready to
     * start the game. If so, a START command is issued to all the
     * participants and the game view is displayed.
     */
    public void startGame() {
        boolean ready = true;

        //Checks if all the players are ready
        Enumeration ids = players.keys();
        while (ids.hasMoreElements()) {
            AbstractPlayer player = getPlayer(ids.nextElement());

            //If a player is not ready, display a error message
            if (!player.isReady()) {
                ready = false;
            }
        }

        //If all players are ready, start the game
        if (ready) {
            //Notify all participants
            Object message = parser.createMessage(Action.START, null);
            this.communicator.sendMessage(message, true);
        }
        else {
            showAlert(
                    "Unable to start game.",
                    "Not all players are ready to start.",
                    3,
                    this.lobby
            );
        }
    }

    /**
     * Method called when one of the players have initiated the game.
     * Changes the view to the GameView, and initiates the neccessary
     * two threads.
     *
     * @see client.framework.views.GameView @param value The game screen resolution
     * @param values The screen resolution for this particular game.
     */
    public void loadGame(String[] values) {
        this.resetScores(); //All players start from scratch

        // Get the screen resolution for the game
        int[] screenResolution = new int[]{
                Integer.parseInt(values[0]),
                Integer.parseInt(values[1])
        };

        // Move self to the intial position
        this.getSelf().setPosition(Integer.parseInt(values[2]), Integer.parseInt(values[3]));
        handleSettings(true);   // Check initial settings
        this.game.startGame(screenResolution);

        this.lobby.setGameStarted(true);
        this.setView(game);         //Displays the gameboard

        //Initiates the local and transmission threads
        this.localThread = new LocalThread();
        this.positionThread = new PositionThread();
        new Thread(this.positionThread).start();
        new Thread(this.localThread).start();
    }

    /**
     * Method called when the local player has moved on the game
     * board.
     *
     * @param movement How much the local player has moved.
     */
    public void fireSelfMoved(int[] movement) {
        this.self.move(movement[0], movement[1]);
        game.fireBoardChanged();
    }

    /**
     * Method that is called whenever the application receives a text message from
     * one of the other participants. The text message is specified in the form
     * <action>":"[<player>]"@"[<values>]. The values may contain one or more attributes.
     * How the values are used depends on the action.
     *
     * @see client.framework.network.parser.Action
     * @param textMessage A string representing the message.
     */
    public void notifyAboutMessageReceived(String textMessage) {
        try {
            if (Character.isDigit(textMessage.charAt(0))) { // Used for the tests
                communicator.sendMessage(Utils.buildString(new Object[]{textMessage, "|"}), false);
                return;
            }

            // Parse the message
            ActionTriplet received = parser.parseMessage(textMessage);

            String action = received.getAction();
            String senderId = received.getSender();
            String values[][] = received.getValues();

            // Find the sender of the action
            AbstractPlayer sender = null;
            if (senderId != null) {
                sender = getPlayer(senderId);
            }

            // Methods are called dependent of the kind of action performed
            if      (Action.SESSION_LIST.equals(action)) fireSessionListReceived(values);
            else if (Action.SETTINGS_LIST.equals(action)) fireSettingsListReceived(values);
            else if (Action.PLAYER.equals(action)) {
                // Never receive player info from "self"
                if (sender == getSelf()) {
                    sender = new BBPlayer(senderId);
                    addPlayer(sender);
                }
                firePlayerInfoReceived(sender, values[0]);
            }
            else if (Action.TEAM.equals(action)) fireTeamAssigned(values[0][0]);
            else if (Action.SCORE.equals(action)) fireScoreReceived(sender, values[0]);
            else if (Action.READY.equals(action)) firePlayerReady(sender, values[0][0]);
            else if (Action.START.equals(action)) loadGame(values[0]);
            else if (Action.SETTING_CHANGED.equals(action)) fireSettingReceived(values[0]);
            else if (Action.POSITION.equals(action)) {
                if (game != null) {
                    firePositionsReceived(values);
                }
            }
            else if (Action.GAME_OVER.equals(action)) stopGame(values[0][0]);
            else if (Action.DISCONNECT.equals(action)) {
                fireDisconnectionReceived(sender);
            }
            else if (Action.ALIVE_REQUEST.equals(action)) fireAliveRequestReceived(values[0][0]);
            else {
                // If no default action matches; check the game specific actions
                notifyAboutSpecialActionReceived(sender, action, values);
            }
        }
        catch (IOException e) {
            System.err.println(e.getMessage());
        }
    }

    /**
     * Called when the client receives an alive request from the server.
     * Returns the id of the request to tell the server that the client
     * is still connected.
     *
     * @param requestId The id of the alive request.
     */
    private void fireAliveRequestReceived(String requestId) {
        Object message = parser.createMessage(Action.ALIVE_REQUEST, new Object[][]{{requestId}});
        communicator.sendMessage(message, false);
    }

    /**
     * Called whenever a player disconnects from the session. The player is
     * then removed from the player list, and from the game board if a game
     * is in progress.
     *
     * If the disconnected player is the local player,
     * the client has stopped transmitting messages to the server because of an
     * error. A notification of this is then displayed, and the application
     * is closed.
     *
     * @param player The player that has disconnected.
     */
    private void fireDisconnectionReceived(AbstractPlayer player) {
        if (player != getSelf()) {
            this.removePlayer(player);
        }
        else {
            this.showAlert(
                    "Connection Problem",
                    "You were disconnected from the server. The application will close.",
                    3000,
                    this.lobby);
            this.quit();
        }
    }

    /**
     * Method that is called when the local player has been assigned
     * to one of the two teams.
     *
     * @param team The team to which the local player has been assigned.
     */
    private void fireTeamAssigned(String team) {
        this.getSelf().setTeam(Integer.parseInt(team));
    }

    /**
     * Called when a list of sessions has been received. Adds the sessions
     * to the list of sessions, and refreshes the session view.
     *
     * @param values The currently active sessions on the server.
     */
    private void fireSessionListReceived(String[][] values) {
        this.sessions.removeAllElements();
        if (values != null) {
            int nofSessions = values.length;
            for (int i = 0; i < nofSessions; i++) {
                this.sessions.addElement(new Session(values[i]));
            }
        }
        this.sessions.addElement(new Session(-1, "New session", 0, 1));
        this.sessionWindow.fireSessionsUpdated();
    }

    /**
     * Ends the game and returns to the lobby. The local threads are
     * stopped to free resources.
     *
     * @param reason The reason for ending the game.
     */
    public void stopGame(String reason) {
        this.game.stopGame();
        this.localThread.stop();
        this.positionThread.stop();
        this.lobby.setGameStarted(false);
        this.showAlert("Game Over", reason, 3, this.lobby);
    }

    /**
     * Called when the contents of the player list have changed, e.g. because
     * of changed score or added or removed players. Notifies the lobby screen
     * of this event.
     */
    public void firePlayerListChanged() {
        this.lobby.notifyAboutPlayerListChanged();
    }

    /* *****************************************************************
     * ABSTRACT METHODS, MUST BE IMPLEMENTED IN CLASSES EXTENDING THIS *
     *******************************************************************/

    /**
     * Returns the name of the game.
     *
     * @return The name of the game.
     */
    public abstract String getGameName();

    /**
     * Returns a new player object of the type used in the game, which
     * is a subclass of AbstractPlayer. This method
     * should create a new player object, for instance like:
     * "return new BBPlayer(address, name, defaultSize, defaultSpeed);"
     *
     * The default size and speed of the Player object should be defined
     * in the implementation of <code>AbstractPlayer</code>.
     *
     * @see AbstractPlayer#AbstractPlayer(String, String, int, int)
     * @param address The player's network address.
     * @param name The name of the player.
     * @return An instance of an AbstractPlayer subclass.
     */
    public abstract AbstractPlayer newPlayer(String address, String name);

    /**
     * Checks whether the state of the game is valid according to
     * the game settings. This method is called contineously throughout
     * the game, and should be implemented the same way as the
     * handleSettings() method.
     *
     * @see AbstractGame#handleSettings(boolean)
     * @param init Whether the method is called on game initialisation or not
     * @return True if all settings are valid, false if not
     */
    protected abstract boolean handleSpecialSettings(boolean init);

    /**
     * Method used to parse and interpret messages received from other participants,
     * that are not among the default actions defined in <code>AbstractGame</code>.
     *
     * For a message, the sender (<code>AbstractPlayer</code>) is specified, and a
     * three-letter word identifying the action leading to the message. The final
     * parameter is a String containing one or more values corresponding to the action.
     *
     * @param player The sender of the message
     * @param action The action specified in the message
     * @param values The value(s) contained in the message
     */
    protected abstract void notifyAboutSpecialActionReceived(AbstractPlayer player, String action, String[][] values);

    /* ***********************************************
     * METHODS INHERITED FROM MIDlet                 *
     *************************************************/

    /**
     * Starts the MIDlet.
     *
     * @throws MIDletStateChangeException
     */
    protected void startApp() throws MIDletStateChangeException {
        StartupWindow startupWindow = new StartupWindow(this);
        setView(startupWindow);
    }

    /**
     * Called when the MIDlet enters its paused state, which
     * can happen for example if the MIDlet is minimized.
     */
    protected void pauseApp() {
    }

    /**
     * Called when the MIDlet is closed. Sends a disconnection notification
     * to the server.
     *
     * @param unconditional If true when this method is called, the MIDlet
     *      must cleanup and release all resources.
     * @throws MIDletStateChangeException
     */
    protected void destroyApp(boolean unconditional)
            throws MIDletStateChangeException {

        if (this.communicator != null) {
            Object message = parser.createMessage(Action.DISCONNECT, null);
            this.communicator.sendMessage(message, true);
            this.communicator.close();   //Closes the communication class
        }
        this.notifyDestroyed();
    }

    /* ********************************************
     * PRIVATE METHODS                            *
     **********************************************/

    /**
     * Method called continously throughout a running game. Compares the
     * predefined settings to the current state of the game. If all settings
     * are valid, the method returns true, if not it returns false. In addition
     * to the default settings specified in the framework, game specific settings
     * may be defined in the game implementation. Handling this settings should
     * be performed in the <AbstractGame.handleSpecialSettings()> method.
     *
     * Some settings should only be checked on game initialisation. This is done
     * by checking if the method is called on game initialisation or not.
     *
     * In the server-client version of the framework, most settings are handled
     * by the server. This method is therefore only used to check whether teams
     * are enabled or not.
     *
     * @param init Whether the check is performed on game initialisation or not.
     * @return True if all settings are valid, false if not.
     */
    private boolean handleSettings(boolean init) {
        //Checks only performed on game initialization
        if (init) {
            checkTeams();   //Handles teams
        }

        //..and additional settings can be specified in the game implementation
        return handleSpecialSettings(init);
    }

    /**
     * Called whenever a value is assigned a setting. The method checks
     * whether the new value for the setting differs from the old value.
     * If so, true is returned.
     *
     * @param setting The setting in question.
     * @param newValue The setting's value.
     * @return True if the value has changed, false if not.
     */
    private boolean fireSettingChanged(String setting, Object newValue) {
        //If the setting exists
        if (settings.containsKey(setting)) {
            Object oldValue = settings.get(setting);

            //Return true if the value has changed
            if (!oldValue.equals(newValue)) {
                settings.put(setting, newValue);
                return true;
            }
        }

        return false;
    }

    /**
     * Called when one of the remote players in the network has moved.
     * The game board is updated with the player positioned in the
     * new position.
     *
     * @param player The player that has moved.
     * @param x The player's x-coordinate.
     * @param y The player's y-coordinate.
     */
    private void firePlayerMoved(AbstractPlayer player, int x, int y) {
        player.setPosition(x, y);
        game.fireBoardChanged();
    }

    /**
     * Sets the scores of all players to zero. This method
     * is called whenever a new game is initiated, to ensure
     * all players start with equal score.
     */
    private void resetScores() {
        Enumeration ids = players.keys();
        while (ids.hasMoreElements()) {
            getPlayer(ids.nextElement()).setScore(0);
        }
    }

    /**
     * Returns the player with the specified id. If none of the players
     * in the player list has this id, the method assumes that the
     * player must be the local player.
     *
     * @param id The id of the player.
     * @return The player having the specified id, or the local player if no players
     * have this id.
     */
    protected AbstractPlayer getPlayer(Object id) {
        AbstractPlayer player = (AbstractPlayer)players.get(id);
        return player != null ? player : getSelf();
    }

    /* MESSAGE SENDING METHODS */

    /**
     * Sends the local player's data to the server.
     *
     * @see AbstractPlayer#getAttributeValues()
     */
    public void sendPlayerInfo() {
        Object message = parser.createMessage(Action.PLAYER, new Object[][]{self.getAttributeValues()});
        this.communicator.sendMessage(message, true);
    }

    /**
     * Sends the local player's position to the server.
     */
    private void sendPosition() {
        Object[] position = new Object[]{new Integer(self.getPosition()[0]), new Integer(self.getPosition()[1])};
        Object message = parser.createMessage(Action.POSITION, new Object[][]{position});
        this.communicator.sendMessage(message, true);
    }

    /**
     * Sends the local player's score to the server.
     */
    private void sendScore() {
        Object message = parser.createMessage(Action.SCORE, null);
        this.communicator.sendMessage(message, true);
    }

    /**
     * Sends the value of a setting to the server.
     *
     * @param name The name of the setting.
     * @param value The setting's value.
     */
    private void sendSetting(String name, Object value) {
        Object message = parser.createMessage(Action.SETTING_CHANGED, new Object[][]{{name, value}});
        communicator.sendMessage(message, true);
    }

    /* MESSAGE RECEIVAL METHODS */

    /**
     * Called when player information is received from one of the
     * other players. The player's attributes are then updated.
     *
     * @see AbstractPlayer#getAttributeValues()
     * @see AbstractPlayer#setAttributeValues(String[])
     * @param player The player sending his attributes.
     * @param values String consisting of all the player's attribute values.
     */
    private void firePlayerInfoReceived(AbstractPlayer player, String[] values) {
        player.setAttributeValues(values);  //Update the player's attributes
        this.firePlayerListChanged();
    }

    /**
     * Called when a player position is received from one of the other
     * players. The player's position is then updated, and the game
     * board is updated.
     *
     * @param values Array containing the position for each player
     */
    private void firePositionsReceived(String[][] values) {
        int nofValues = values.length;
        for (int i = 0; i < nofValues; i++) {
            String[] position = values[i];
            AbstractPlayer player = getPlayer(position[0]);
            if (player != getSelf()) {
                firePlayerMoved(player, Integer.parseInt(position[1]), Integer.parseInt(position[2]));
            }
        }
    }

    /**
     * Called when a score is received from one of the other
     * players. The player's score is then updated, and the player
     * list is updated.
     *
     * @param player The player whose score has been received.
     * @param values The new score.
     */
    private void fireScoreReceived(AbstractPlayer player, String[] values) {
        player.setScore(Integer.parseInt(values[0]));
        firePlayerMoved(player, Integer.parseInt(values[1]), Integer.parseInt(values[2]));
        if (player != getSelf()) {  // "Double position" to avoid calculating a vector leading out of the game board
            firePlayerMoved(player, Integer.parseInt(values[1]), Integer.parseInt(values[2]));
        }
        this.firePlayerListChanged();
    }

    /**
     * Called when a list of settings is received from the server. The settings list
     * is cleared, and the settings with their values are added to the list.
     *
     * @param values Array containing setting names and values.
     */
    private void fireSettingsListReceived(String[][] values) {
        this.settings.clear();
        int nofSettings = values.length;
        for (int i = 0; i < nofSettings; i++) {
            settings.put(values[i][0], getSettingObject(values[i][1]));
        }
    }

    /**
     * Parses the value of a setting to the correct object type. The possible
     * types for values are boolean, int or String.
     *
     * @param value The value that needs to be converted.
     * @return The value of the correct object type.
     */
    private Object getSettingObject(String value) {
        if ("true".equals(value)) return Boolean.TRUE;
        else if ("false".equals(value)) return Boolean.FALSE;
        else if (Character.isDigit(value.charAt(0))) {
            return Integer.valueOf(value);
        }
        else {
            return value;
        }
    }

    /**
     * Called when change of a setting has been performed. The setting
     * is then updated with the new value.
     *
     * @param values String array containing the setting and the new value.
     */
    private void fireSettingReceived(String[] values) {
        //Gets the setting and its value
        fireSettingChanged(values[0], getSettingObject(values[1]));
    }

    /**
     * Checks how many players are connected to a session. This method
     * is only called when trying to join a session, and the number of players
     * in the session is not updated until the session list is refreshed. Thus,
     * more players than allowed may in some cases be allowed to join the session.
     *
     * @param session The session to join
     * @return False if too many players have joined the session, else true.
     */
    private boolean checkNofPlayers(Session session) {
        //If too many players have joined, the game can not start
        return session.getNofPlayers() + 1 <= session.getMaxPlayers();
    }

    /**
     * Checks whether teams are enabled in the game. If so, the players are
     * assigned a team color corresponding to their team.
     */
    private void checkTeams() {
        if (((Boolean)settings.get(TEAMS_ON)).booleanValue()) {
            Enumeration ids = players.keys();
            while (ids.hasMoreElements()) {
                AbstractPlayer player = getPlayer(ids.nextElement());
                //Give each player his team color.
                player.setTeamColor(TEAMCOLORS[player.getTeam()]);
            }
        }
    }

    /**
     * Called when closing the application.
     * @see AbstractGame#destroyApp(boolean)
     */
    public void quit() {
        try {
            this.destroyApp(true);
        } catch (MIDletStateChangeException e) {
            // Do nothing
        }
    }

    /**
     * Returns a list of the found sessions.
     *
     * @return The session list.
     */
    public Vector getSessions() {
        return this.sessions;
    }

    /**
     * Called when a player has selected to join a session. If the player
     * is allowed to join the session (the session is not full), the server
     * is notified.
     *
     * @param selectedSession The session the player wishes to join.
     */
    public void fireSessionSelected(int selectedSession) {
        Session session = (Session)this.sessions.elementAt(selectedSession);
        if (checkNofPlayers(session)) {
            this.sessions.removeAllElements();
            this.sessions = null;
            Object message = parser.createMessage(Action.SESSION_SELECTED, new Object[][]{new Object[]{new Integer(session.getId())}});
            this.communicator.sendMessage(message, true);
            this.setView(this.lobby);
        }
    }

    /**
     * Called when the player wishes to refresh the session list.
     * Causes the client to reconnect to the server.
     */
    public void refreshSessionList() {
        this.sendPlayerInfo();
    }

    /**
     * Checks whether the teams are enabled or not.
     *
     * @return True if teams are enabled; false if not.
     */
    public boolean isTeamsEnabled() {
        Boolean value = (Boolean)this.settings.get(TEAMS_ON);
        return value != null && value.booleanValue();
    }

    /**
     * Called when the player wishes to rejoin the running game
     * after having visited the lobby view.
     */
    public void fireReturnToGameSelected() {
        this.setView(game);
    }

    /********************************************************
     * PRIVATE THREADS                                      *
     ********************************************************/

    /**
     * A <code>Runnable</code> object responsible for the local running
     * of the game. The speed of which the game is updated depends on
     * the <code>GAME_FPS</code> variable defined in <code>AbstractGame</code>.
     *
     * Each time the Thread repeats, the settings are controlled. If all settings
     * are valid the <code>check()</code> method in GameView is called.
     * If the settings are not valid, the game is stopped, and the players return
     * to the lobby.
     *
     * Also, more importantly, this threads causes the local calculations to be performed
     * each time the thread loops. That is, collision detections and movements are performed
     * each time this thread loops.
     *
     * @see client.framework.views.GameView#check()
     * @see AbstractGame#stopGame(String)
     */
    private class LocalThread implements Runnable {
        //Whether the game is active
        private boolean running = true;

        /**
         * The method called when the Thread is started.
         */
        public void run() {
            while (running) {
                try {
                    if (handleSettings(false)) game.check(); // Performs the local calculations.
                    else                       stopGame("Illegal setting");
                    Thread.sleep(1000 / GAME_FPS);
                } catch (InterruptedException e) {
                    e.printStackTrace();
                }
            }
        }

        /**
         * Causes the <code>run()</code> method to break
         */
        public void stop() {
            running = false;
        }
    }

    /**
     * A <code>Runnable</code> object that is resonsible for
     * sending the local player's position to the server
     * if it has been changed. How often the position check is performed
     * depends on the <code>POSITION_INTERVAL</code> variable defined in
     * <code>AbstractGame</code>.
     *
     * If the local player has stopped moving, the player's position is sent
     * to the server exactly twice. After that, the position is not transmitted
     * until the player has started moving again.
     */
    private class PositionThread implements Runnable {
        private boolean sentStandingStill = false;

        //Whether the game is active
        private boolean running = true;

        /**
         * The method called when the Thread is started.
         */
        public void run() {
            int[] lastPosition = new int[]{
                    self.getPosition()[0],
                    self.getPosition()[1]
            };

            while (running) {
                try {
                    //Gets the cu   rrent position
                    int x = self.getPosition()[0];
                    int y = self.getPosition()[1];

                    //If the position has not been changed..
                    if (x == lastPosition[0] && y == lastPosition[1]) {
                        // ..send the position once to signal that the player has stopped
                        if (!sentStandingStill) {
                            sentStandingStill = true;
                            sendPosition();
                        }
                    }
                    // If the position has been changed, send the new position
                    else {
                        sentStandingStill = false;
                        //Send the new position
                        sendPosition();
                        //Store the new position
                        lastPosition[0] = x;
                        lastPosition[1] = y;
                    }

                    Thread.sleep(sendInterval);
                } catch (InterruptedException e) {
                    e.printStackTrace();
                }
            }
        }

        /**
         * Causes the <code>run()</code> method to break.
         */
        public void stop() {
            running = false;
        }
    }
}