package client.framework.models;

import client.framework.Utils;

import javax.microedition.lcdui.Graphics;
import javax.microedition.lcdui.Image;
import javax.microedition.lcdui.game.Sprite;
import java.util.Random;

/**
 * An <code>AbstractPlayer</code> object is a representation of
 * a participant in a framework game. Each player has a <code>Sprite</code>
 * object, which is his visual repesentation on the game board.
 *
 * In addition, the player has different attributes such as a name, id,
 * color, score, and size.
 *
 * This class also contain support for simple movement
 * prediction, where future positions are calculated based
 * on the two last received position updates.
 *
 * @author Eivind Sorteberg
 * @author Martin Jarrett
 * @version 2.0
 */
public abstract class AbstractPlayer implements Comparable {
    //The network id of the player.
    private String id;

    //The player's name
    private String name;

    //The player's RGB color
    private int[] color;

    //The team of which the player is a member (if any)
    private int team;

    //The RGB color of the player's team
    private int[] teamColor;

    //The player's score
    private Integer score;

    //The player's ready status
    private boolean ready;

    //The size of the player's <code>Sprite</code>
    private int size;

    //The speed in which the player is able to move
    private int speed;

    //The <code>Sprite</code> representing the player
    private Sprite sprite;

    // The player's current position
    private int[] position;

    // The last received position
    private int[] lastPosition;

    // The direction in which the player moves.
    private int[] movement;


    //The player's phone's screen resolution
    private int[] screenResolution;

    /**
     * Constructor for AbstractPlayer. All implementing
     * subclasses should call this contructor through a
     * <code>super()</code>-call. The constructor creates an
     * <code>AbstractPlayer</code>-instance, and gives default values
     * to all attributes not specified as input parameters.
     *
     * @param id The player's network id.
     * @param name The player's name.
     * @param size The size of the player's <code>Sprite</code> object.
     * @param speed The speed in which the player is able to move.
     * @see AbstractGame#newPlayer(String, String)
     */
    protected AbstractPlayer(String id, String name, int size, int speed) {
        this.id = id;
        this.name = name;
        this.size = size;
        this.speed = speed;
        this.ready = false;
        this.team = -1;
        this.position = new int[]{-1, -1};
        this.lastPosition = new int[]{-1, -1};
        this.movement = new int[2];
        this.score = new Integer(0);

        Random r = new Random();

        //Generates a random color.
        this.color = new int[3];
        this.color[0] = r.nextInt(256);
        this.color[1] = r.nextInt(256);
        this.color[2] = r.nextInt(256);

        //Creates the player's sprite
        this.sprite = createSprite();

        this.screenResolution = new int[2];
    }

    /**
     * Constructor using default values for size (10) and
     * speed (1).
     *
     * @param id The player's id.
     * @param name The player's name.
     */
    public AbstractPlayer(String id, String name) {
        this(id, name, 10, 1);
    }

    /**
     * Constructor using an empty name and default values
     * for size (10) and speed (1).
     *
     * @param id The id of the player.
     */
    public AbstractPlayer(String id) {
        this(id, "");
    }

    /**
     * Returns the player's id. The id is assigned by the server.
     *
     * @return The player's id.
     */
    public String getId() {
        return this.id;
    }

    /**
     * Returns the player's name.
     *
     * @return The player's name.
     */
    public String getName() {
        return this.name;
    }

    /**
     * Sets the player's name.
     *
     * @param name The player's name.
     */
    public void setName(String name) {
        this.name = name;
    }

    /**
     * Returns the player's color in an array containing values
     * between [0 and 255] for each of the colors red, green and
     * blue.
     *
     * @return The player's color.
     */
    public int[] getColor() {
        return this.color;
    }

    /**
     * Sets the player's color. Values above 255 are set to 255,
     * and values below 0 are set to 0. Each of the colors red,
     * green and blue needs to be defined.
     *
     * @param red The amount of red.
     * @param green The amound of green.
     * @param blue The amount of blue.
     * @see AbstractPlayer#setImage()
     */
    public void setColor(int red, int green, int blue) {
        this.color[0] = checkColor(red);
        this.color[1] = checkColor(green);
        this.color[2] = checkColor(blue);

        //Creates an image corresponding to the new color.
        this.setImage();
    }

    /**
     * Sets the player's team color. If the player does not have a team color,
     * the team color is set to player's color.
     * 
     * @param color The player's team's color
     * @see AbstractPlayer#setImage()
     */
    public void setTeamColor(int[] color) {
        if (this.teamColor == null) {
            this.teamColor = color;
        }
        //Creates an image corresponding to the player's colors.
        this.setImage();
    }

    /**
     * Creates an image for the player's sprite. If the player has no team color,
     * the image is a simple square with just the player's color.
     * If the player does have a team color, the sprite's top right corner is
     * filled with this color.
     */
    public void setImage() {
        Image image = Utils.createImage(size, size, color[0], color[1], color[2]);

        //Creates team color
        if (this.teamColor != null) {
            Graphics graphics = image.getGraphics();
            graphics.setColor(this.teamColor[0], this.teamColor[1], this.teamColor[2]);
            graphics.fillTriangle(0, this.size / 2, this.size / 2, 0, this.size / 2, this.size / 2);
            graphics.fillTriangle(0, this.size / 2, this.size / 2, 0, 0, 0);
        }

        //Assigns the image to the sprite
        this.sprite.setImage(image, this.size, this.size);
    }

    /**
     * Method for displaying the player's name and score. This
     * method is used to create the player lists in the lobby
     * and in-game.
     *
     * @see Utils#drawPlayerList(javax.microedition.lcdui.Graphics, int, int, AbstractGame, boolean)
     * @return The player's name and score.
     */
    public String toString() {
        return new StringBuffer(this.getName()).append("   [").append(this.getScore()).append("]").toString();
    }

    /**
     * Returns the player's score.
     *
     * @return The player's score.
     */
    public Integer getScore() {
        return score;
    }

    /**
     * Sets the player's score to the specified value.
     *
     * @param score The player's score.
     */
    public void setScore(int score) {
        this.score = new Integer(score);
    }

    /**
     * Returns the player's ready status.
     *
     * @return True if the player is ready, false if not.
     */
    public boolean isReady() {
        return ready;
    }

    /**
     * Sets the player's ready status to the specified value.
     *
     * @param ready Whether the player is ready or not.
     */
    public void setReady(boolean ready) {
        this.ready = ready;
    }

    /**
     * Returns the team that the player is member of, if any.
     * If the player is not member of any teams, the method will
     * return -1 (the default team value).
     *
     * @return The player's team number, or -1 if none.
     */
    public int getTeam() {
        return team;
    }

    /**
     * Sets the player as member of the specified team. If
     * this method is never called, the player's team number
     * will be -1.
     *
     * @param team The player's team number.
     */
    public void setTeam(int team) {
        this.team = team;
    }

    /**
     * Checks whether the input object is the same player object
     * as this. If the object is not an <code>AbstractPlayer</code>
     * object, or if the player names and/or addresses differ, the
     * method returns false.
     *
     * @param object The object to compare this to.
     * @return True if the players are the same, false if not.
     */
    public boolean equals(Object object) {
        if (!(object instanceof AbstractPlayer)) return false;
        else {
            AbstractPlayer p2 = (AbstractPlayer)object;

            return  p2.getId().equals(this.getId()) &&
                    p2.getName().equals(this.getName());
        }
    }

    /**
     * Returns the player's sprite representation.
     *
     * @return The player's sprite.
     */
    public Sprite getSprite() {
        return this.sprite;
    }

    /**
     * Sets the screen resolution variable to the screen resolution on
     * the player's phone.
     *
     * @param screenResolution The screen resolution retrieved from the phone.
     */
    public void setScreenResolution(int[] screenResolution) {
        this.screenResolution = screenResolution;
    }

    /**
     * Returns an array containing the player's screen resolution.
     *
     * @return The player's screen resolution.
     */
    public int[] getScreenResolution() {
        return screenResolution;
    }

    /**
     * Sets the x-position of the player.
     *
     * @param x The player's x-position.
     */
    public void setX(int x) {
        this.position[0] = x;
        this.sprite.setPosition(x, this.position[1]);
    }

    /**
     * Sets the y-position of the player.
     *
     * @param y The player's y-position.
     */
    public void setY(int y) {
        this.position[1] = y;
        this.sprite.setPosition(this.position[0], y);
    }

    /**
     * Sets the player's position on the game board to the specified
     * position. When this is done, the player's movement direction
     * is calculated based on the previously received position
     * update.
     *
     * @param x The player's x-position.
     * @param y The player's y-position.
     */
    public void setPosition(int x, int y) {
        if (AbstractGame.MOVEMENT_PREDICTION) {
            this.calculateMovementVector(x, y);
        }
        this.position[0] = x;
        this.position[1] = y;
        this.lastPosition[0] = x;
        this.lastPosition[1] = y;
        this.sprite.setPosition(x, y);
    }

    /**
     * Moves the player a distance determined by the
     * input parameters.
     *
     * @param dx The x-movement.
     * @param dy The y-movement.
     */
    public void move(double dx, double dy) {
        this.position[0] += dx;
        this.position[1] += dy;
        this.sprite.setPosition(position[0], position[1]);
    }

    /**
     * Called to move the player a distance detemined by the player's
     * movement vector. Is called whenever the game thread loops, and the player
     * needs to be moved by the movement prediction algorithm.
     */
    public void move() {
        this.move(this.movement[0], this.movement[1]);
    }

    /**
     * Returns an array containing the player's x- and
     * y-coordinate in that order.
     *
     * @return The player's position.
     */
    public int[] getPosition() {
        return this.position;
    }

    /**
     * Sets the size of the player's sprite to the specified size,
     * and creates the sprite's image.
     *
     * @param size The player's size.
     */
    public void setSize(int size) {
        this.size = size;
        this.setImage();
    }

    /**
     * Returns the size of the player's sprite.
     *
     * @return The player's size.
     */
    public int getSize() {
        return this.size;
    }

    /**
     * Returns the speed in which the player moves on the
     * game board.
     *
     * @return The player's speed.
     */
    public int getSpeed() {
        return speed;
    }

    /**
     * Set's the speed in which the player moves in the
     * game board to the specified value.
     *
     * @param speed The player's speed.
     */
    public void setSpeed(int speed) {
        this.speed = speed;
    }

    /**
     * Returns a String containing the most important player attributes.
     * From this array, a String is created and transmitted to the server
     * when joining connecting to a server, and the player's representation
     * on the other mobile phones is created from this attribute String.
     *
     * @see AbstractGame#sendPlayerInfo()
     * @see AbstractPlayer#setAttributeValues(String[])
     * @return A String array containing the player's attributes.
     */
    public String[] getAttributeValues() {
        return new String[]{
                name,
                Integer.toString(color[0]),
                Integer.toString(color[1]),
                Integer.toString(color[2]),
                score.toString(),
                Integer.toString(team),
                ready ? "true" : "false",
                Integer.toString(position[0]),
                Integer.toString(position[1]),
                Integer.toString(screenResolution[0]),
                Integer.toString(screenResolution[1])
        };
    }

    /**
     * Called when a player's attributes has been received. Parses a String
     * containing attribute values and applies the correct values to the
     * different settings. The attribute String is always created by the
     * <code>getAttributeValues()</code> method, and so the rules from that
     * method are used for parsing.
     *
     * @param values The player's attributes
     * @see AbstractPlayer#getAttributeValues()
     */
    public void setAttributeValues(String[] values) {
        this.name = values[0];
        int red = Integer.parseInt(values[1]);
        int green = Integer.parseInt(values[2]);
        int blue = Integer.parseInt(values[3]);
        this.setColor(red, green, blue);
        this.setScore(Integer.parseInt(values[4]));
        this.setTeam(Integer.parseInt(values[5]));
        this.setReady("true".equals(values[6]));
        int xPos = Integer.parseInt(values[7]);
        int yPos = Integer.parseInt(values[8]);
        this.setPosition(xPos, yPos);
        int screenX = Integer.parseInt(values[9]);
        int screenY = Integer.parseInt(values[9]);
        this.setScreenResolution(new int[]{screenX, screenY});
        this.resetMovement();
    }

    /**
     * Method inherited from the <code>Comparable</code> interface.
     * Is used to sort player objects according to their score. If the
     * object passed as input parameter is not a <code>AbstractPlayer</code>
     * object, 1 is returned. If the input parameter is an
     * <code>AbstractPlayer</code>, the method returns the difference between
     * this player's score and the input parameter's score.
     *
     * @param object The object to compare to.
     * @return Positive value if this player has higher score than the input player,
     *      negative value if the input player has the higher score and 0 if they
     *      have equal score.
     */
    public int compareTo(Comparable object) {
        if (!(object instanceof AbstractPlayer)) return 1;
        else {
            AbstractPlayer other = ((AbstractPlayer)object);
            if (this.getScore().intValue() != other.getScore().intValue()) {
                return this.getScore().intValue() - other.getScore().intValue();
            }
            else {
                return (((AbstractPlayer)object).getName()).compareTo(this.getName());
            }
        }
    }

    /**
     * Resets the player's movement.
     */
    protected void resetMovement() {
        this.movement[0] = 0;
        this.movement[1] = 0;
    }

    /**
     * Calculates the player's movement direction, based on a new and
     * an old position.
     *
     * @param newX The new x-position.
     * @param newY The new y-position.
     */
    private void calculateMovementVector(int newX, int newY) {
        if (this.lastPosition[0] >= 0) {
            int oldX = this.lastPosition[0];
            int oldY = this.lastPosition[1];
            int dx = newX - oldX;
            int dy = newY - oldY;
            this.movement[0] = Math.abs(dx) > 0 ? (Math.abs(dx)) / dx * this.getSpeed() : 0;
            this.movement[1] = Math.abs(dy) > 0 ? (Math.abs(dy)) / dy * this.getSpeed() : 0;
        }
    }

    /**
     * Creates a square <code>Sprite</code> with the player's size and color.
     *
     * @return A <code>Sprite</code> object.
     */
    private Sprite createSprite() {
        return new Sprite(Utils.createImage(size, size, color[0], color[1], color[2]));
    }

    /**
     * Checks whether a color amount is within the boundaries ([0, 255]).
     * If the amout is too low, it is adjusted to 0; if it is too high, it
     * is adjusted to 255.
     *
     * @param amount The color amount to check.
     * @return The adjusted color amount.
     */
    private int checkColor(int amount) {
        if (amount <= 0)        return 0;
        else if (amount >= 255) return 255;
        else                    return amount;
    }
}
