package client.framework.models;

import client.framework.Utils;

/**
 * This class is a model representing a session running on the
 * server to which the client is connected. The session consist
 * of an id, a name, a number of connected players, and a
 * maximum number of players.
 *
 * @version 1.0
 * @author Eivind Sorteberg
 * @author Martin Jarrett
 */
public class Session {
    private int id;
    private String name;
    private int nofPlayers;
    private int maxPlayers;

    /**
     * Constructor that creates a <code>Session</code> object
     * with the specified attributes.
     *
     * @param id The session's id.
     * @param name The session's name.
     * @param nofPlayers The number of currently connected players.
     * @param maxPlayers The maximum number of players.
     */
    public Session(int id, String name, int nofPlayers, int maxPlayers) {
        this.id = id;
        this.name = name;
        this.maxPlayers = maxPlayers;
        this.nofPlayers = nofPlayers;
    }

    /**
     * Creates a new <code>Session</code> object, where the
     * session's attributes are contained in an array.
     *
     * @see Session#Session(int, String, int, int)
     * @param values The session's attributes.
     */
    public Session(String[] values) {
        this.id = Integer.parseInt(values[0]);
        this.name = values[1];
        this.nofPlayers = Integer.parseInt(values[2]);
        this.maxPlayers = Integer.parseInt(values[3]);
    }

    /**
     * Returns a string representation of the session. This method is used
     * by the <code>SessionView</code> to list the server's active sessions,
     * and their attributes.
     *
     * @return A string representation of the session.
     */
    public String toString() {
        String s = this.name;
        if (this.id >= 0) {
            s = Utils.buildString(new Object[]{" (", new Integer(nofPlayers), "/", new Integer(maxPlayers), ")"});
        }
        return s;
    }

    /**
     * Returns the session's id.
     *
     * @return The session's id.
     */
    public int getId() {
        return id;
    }

    /**
     * Returns the session's name.
     *
     * @return The session's name.
     */
    public String getName() {
        return name;
    }

    /**
     * Returns the session's number of currently connected players.
     *
     * @return The number of connected players.
     */
    public int getNofPlayers() {
        return nofPlayers;
    }

    /**
     * Returns the session's maximum number of players. Used to check
     * whether the local player is allowed to connect to the session or
     * not (if it is full).
     *
     * @return The session's maximum number of players.
     */
    public int getMaxPlayers() {
        return maxPlayers;
    }
}
