package client.framework.network;

import client.framework.models.AbstractGame;

import javax.microedition.io.*;
import java.io.*;
import java.util.Vector;

/**
 * This class is a <code>Communicator</code> object that
 * communicates using <code>SocketConnection</code> or
 * <code>UDPDatagramConnection</code> objects for its communication,
 * dependent of whether the transport protocol in use is
 * TCP or UDP, respectively.
 *
 * @version 1.0
 * @author Martin Jarrett
 * @author Eivind Sorteberg
 */
public class ClientSocketCommunicator implements Communicator {
    // How long the receiver thread sleeps before it checks for incoming data
    private static final int RECEIVE_SLEEPTIME = 10;

    // The connection object
    private Connection connection;

    // The model with which the communicator communicates locally
    private AbstractGame model;

    // The thread responsible for sending outgoing messages
    protected SenderThread sender;

    // The size of the UDP data packets
    private int packetSize;

    // The addresses to which the communicator is connected
    private String[] addresses;

    /**
     * Creates a <code>ClientSocketCommunicator</code> object.
     *
     * @param abstractGame The model with which the communicator communicates locally
     * @param packetSize The size of the UDP data packets (not used for TCP communication)
     */
    public ClientSocketCommunicator(AbstractGame abstractGame, int packetSize) {
        this.model = abstractGame;
        this.packetSize = packetSize;
    }

    /**
     * Connects to the specified address(es). The address(es)
     * needs to specify the protocol to use (socket or datagram), the
     * address(es) to the server(s) and the port number of the server.
     * The format of the address(es) is:
     * <protocol>://<server IP>:<server port>.
     *
     * Initialises the sender and receiver threads.
     *
     * @param addresses The address(es) to connect to.
     * @return Always true (assumes successful connection).
     */
    public boolean connect(final String[] addresses) {
        this.addresses = addresses;
        this.startThreads();
        return true;
    }

    /**
     * Not implemented.
     *
     * @param adresses Not used.
     */
    public void disconnect(String[] adresses) {
    }

    /**
     * Sends the specified message to the server.
     *
     * @param message The message to send.
     * @param confirm Not used in this implementation.
     */
    public void sendMessage(Object message, boolean confirm) {
        this.sender.sendMessage(message);
    }

    /**
     * Called when the receiver thread has received a message.
     * Forwards this message to the model.
     *
     * @see AbstractGame#notifyAboutMessageReceived(String)
     * @param message The message that was received.
     */
    public void notifyAboutMessageReceived(Object message) {
        this.model.notifyAboutMessageReceived(message.toString());
    }

    /**
     * Used to check if a connection has been established.
     *
     * @return True if a connection has been established, false if not.
     */
    private boolean isConnected() {
        return this.connection != null;
    }

    /**
     * Called when the connection is closed. In this implementation,
     * this method has no effect.
     */
    public void close() {
        this.disconnect(null);
    }

    /**
     * Not implemented.
     */
    public void searchForNodes() {
    }

    /**
     * Initiates the sender and receiver threads. Called when
     * starting the application.
     */
    private void startThreads() {
        ConnectorThread connectorThread = new ConnectorThread();
        this.sender = new SenderThread();
        ReceiverThread receiverThread = new ReceiverThread();

        new Thread(connectorThread).start();
        new Thread(sender).start();
        new Thread(receiverThread).start();
    }

    /**
     * Thread that connects to the address(es) given when initialising
     * the <code>ClientSocketCommunicator</code> object. When the connection
     * has been established, the local player's attributes are transmitted
     * to the server(s).
     */
    private class ConnectorThread implements Runnable {
        public void run() {
            try {
                for (int i = 0; i < addresses.length; i++) {
                    connection = Connector.open(addresses[i]);
                    model.sendPlayerInfo();
                }
            } catch (IOException e) {
                // Nothing to do
                e.printStackTrace();
            }
        }
    }

    /**
     * Thread responsible for sending messages to the server. Dependent
     * of whether UDP or TCP is used for the communication, the appropriate
     * transmission mechanisms are used.
     */
    protected class SenderThread implements Runnable {
        // A list containing the messages waiting to be sent.
        private Vector sendQueue;

        /**
         * Initialises the send queue.
         */
        private SenderThread() {
            this.sendQueue = new Vector();
        }

        /**
         * Sends messages that are waiting to be sent. When the send queue is empty,
         * this thread sleeps until new messages to be sent arrive.
         */
        public void run() {
            while (!isConnected()) { /* Wait for connection */ }

            try {
                if (connection instanceof UDPDatagramConnection) {
                    UDPDatagramConnection udpConn = (UDPDatagramConnection)connection;
                    synchronized (this) {
                        while (isConnected()) {
                            while (!sendQueue.isEmpty()) {
                                Object action = sendQueue.elementAt(0);
                                sendQueue.removeElement(action);
                                byte[] toSend = action.toString().getBytes();
                                Datagram data = udpConn.newDatagram(toSend, toSend.length);
                                udpConn.send(data);
                                Thread.sleep(10);
                            }
                            wait();
                        }
                    }
                }
                else if (connection instanceof SocketConnection) {
                    model.fireMessagesChanged("connection is SocketConnection", true);
                    SocketConnection sConn = (SocketConnection)connection;
                    OutputStream outputStream = sConn.openOutputStream();
                    OutputStreamWriter writer = new OutputStreamWriter(outputStream);
                    synchronized (this) {
                        while (isConnected()) {
                            while (!sendQueue.isEmpty()) {
                                Object action = sendQueue.elementAt(0);
                                sendQueue.removeElement(action);
                                writer.write(action.toString());
                                writer.flush();
                                Thread.sleep(10);
                            }
                            wait();
                        }
                    }
                }
            }
            catch (IOException e) {
                e.printStackTrace();
                model.fireMessagesChanged(e.getMessage(), true);
            }
            catch (InterruptedException e) {
                e.printStackTrace();
                model.fireMessagesChanged(new StringBuffer("Exception: ").append(e.getMessage()).toString(), true);
            }
        }

        /**
         * Called when a new message is to be sent. Adds the message to the
         * message queue, and notifies the sender thread that messages are
         * waiting to be sent.
         *
         * @param message The message to be sent.
         */
        public void sendMessage(Object message) {
            this.sendQueue.addElement(message);
            synchronized (this) {
                this.notify();
            }
        }
    }

    /**
     * Thread responsible for listening to the connection, and detecting
     * when new messages arrive. When this happens, the thread calls the
     * <code>notifyAboutMessageReceived()</code>.
     *
     * Between each loop in the listen thread, the thread sleeps for an
     * amount of time specified by the <code>RECEIVE_SLEEPTIME</code>
     * variable, to avoid occupying too much resources.
     */
    private class ReceiverThread implements Runnable {
        /**
         * The loop listening for incoming messages on the connection.
         */
        public void run() {
            try {
                while(!isConnected()) {/* Waiting for connection */}

                while (isConnected()) {
                    if (connection instanceof UDPDatagramConnection) {
                        UDPDatagramConnection udpConn = (UDPDatagramConnection)connection;
                        byte[] bytes = new byte[packetSize];
                        Datagram received = udpConn.newDatagram(bytes, bytes.length);
                        udpConn.receive(received);

                        StringBuffer message = new StringBuffer();
                        for (int i = 0; i < bytes.length; i++) {
                            char c = (char)bytes[i];
                            if (c == '|') break;
                            message.append(c);
                        }
                        notifyAboutMessageReceived(message.toString());
                    }
                    else if (connection instanceof SocketConnection) {
                        SocketConnection socketConn = (SocketConnection)connection;
                        InputStream inputStream = socketConn.openInputStream();
                        InputStreamReader reader = new InputStreamReader(inputStream);
                        StringBuffer message = new StringBuffer();
                        while(isConnected()) {
                            while(reader.ready()) {
                                char character = (char)reader.read();
                                if (character == '|') {
                                    notifyAboutMessageReceived(message.toString());
                                    message = new StringBuffer();
                                }
                                else {
                                    message.append(character);
                                }
                            }
                        }
                    }
                    Thread.sleep(RECEIVE_SLEEPTIME);
                }
            } catch (IOException e) {
                e.printStackTrace();
            } catch (InterruptedException e) {
                e.printStackTrace();
            }
        }
    }
}