package client.framework.network;

/**
 * An interface that specifies methods for connecting to and
 * communicating with a server.
 *
 * @version 2.0
 * @author Martin Jarrett
 * @author Eivind Sorteberg
 */
public interface Communicator {
    /**
     * Connects to the server(s) with the specified address(es).
     * The format of these addresses depends on the type of
     * communication specified in the interface implementation.
     *
     * @param addresses The address(es) to connect to.
     * @return Whether the connection was successful or not.
     */
    public boolean connect(String[] addresses);

    /**
     * Disconnects from the specified address(es).
     *
     * @param addresses The address(es) to disconnect from.
     */
    public void disconnect(String[] addresses);

    /**
     * Called when the client's model wants to send a message
     * using its <code>Communicator</code> instance. This message
     * should be created by the <code>MessageParser</code> implementation,
     * to ensure that the format of the message follow the specified format.
     *
     * In implementations of this method, the confirm flag can be used to
     * implement guaranteed delivery of critical messages, similar to
     * that found in the TCP transport protocol.
     *
     * @param message The message to send.
     * @param confirm Whether receival of the message should be confirmed by the receiver.
     */
    public void sendMessage(Object message, boolean confirm);

    /**
     * Called when a message is received. The implementation of this method
     * should notify the model that a message has been received.
     *
     * @param message The message that has been received.
     */
    public void notifyAboutMessageReceived(Object message);

    /**
     * Used to close the communicator object, and free resources.
     */
    public void close();

    /**
     * Searches for network nodes to connect to. Is not used for
     * client-server networks, but can be useful if the framework is
     * used in a peer-to-peer network.
     */
    public void searchForNodes();
}
