package client.framework.network.parser;

import client.framework.Utils;

import java.io.IOException;
import java.util.Vector;

/**
 * An implementation of the <code>MessageParser</code> interface, that
 * creates and parses <code>String</code> messages. The format of these messages
 * is <action>":"[<player>"@"][<value>{","<value>}{";"<value>{","<value>}}]"|" (in EBNF-notation).
 *
 * @version 1.0
 * @author Eivind Sorteberg
 * @author Martin Jarrett
 */
public class StringParser implements MessageParser {
    // The different separators defined for the format
    private static final char SEP_ACTION = ':';
    private static final char SEP_SENDER = '@';
    private static final char SEP_VALUEPAIR = ';';
    private static final char SEP_VALUE = ',';
    private static final char END_OF_MESSAGE = '|';

    /**
     * Parses a given string message into an <code>ActionTriplet</code>
     * wrapper. If the message is not a <code>String</code> object, an
     * <code>IOException</code> is thrown.
     *
     * @param message The string message to parse.
     * @return An <code>ActionTriplet</code> containing the message's contents.
     * @throws IOException If the given message is not a <code>String</code> instance.
     */
    public ActionTriplet parseMessage(Object message) throws IOException {
        if (message instanceof String) {
            return this.parseStringMessage((String)message);
        }
        else throw new IOException(Utils.buildString(new Object[]{"Expected String, found: ", message.getClass().getName()}));
    }

    /**
     * Creates a string message in the specified format
     * from the given action identifier and values. The
     * string message is wrapped as an <code>Object</code>
     * to meet the interface's specification.
     *
     * @param action The action identifier.
     * @param values An array containing the message's values.
     * @return A string message to send to the server.
     */
    public Object createMessage(Action action, Object[][] values) {
        return this.createStringMessage(action, values);
    }

    /**
     * Parses a given string message into an <code>ActionTriplet</code>
     * wrapper. The message is ensured to be a string message when
     * this method is called. An <code>IOException</code> is thrown if
     * the message does not follow the specified message format.
     *
     * @param message The message to parse.
     * @return An <code>ActionTriplet</code> wrapper.
     * @throws IOException If the message does not follow the specified format.
     */
    private ActionTriplet parseStringMessage(String message) throws IOException {
        int indexOfActionSeparator = message.indexOf(SEP_ACTION);

        if (indexOfActionSeparator < 0) throw new IOException("Invalid message: Missing action separator '" + SEP_ACTION + "'");

        String action = message.substring(0, indexOfActionSeparator);
        String sender = null;

        int indexOfSenderSeparator = message.indexOf(SEP_SENDER);
        if (indexOfSenderSeparator >= 0) {
            sender = message.substring(indexOfActionSeparator + 1, indexOfSenderSeparator);
            message = message.substring(indexOfSenderSeparator + 1);
        }
        else {
            message = message.substring(indexOfActionSeparator + 1);
        }
        String[][] attributes = getAttributes(message);
        return new ActionTriplet(action, sender, attributes);
    }

    /**
     * Returns the [<value>{","<value>}{";"<value>{","<value>}}]
     * part of the message in an n*m array, where n is the number of
     * separate value strings and m is the number of values per value string.
     *
     * @param message The values in string format.
     * @return An array containing the separate values.
     */
    private static String[][] getAttributes(String message) {
        if (message.length() > 0) {
            int indexOfLastIdentifier = 0;
            int indexOfNextIdentifier = message.indexOf(SEP_VALUEPAIR);
            Vector arrays = new Vector();
            while (indexOfNextIdentifier >= 0) {
                arrays.addElement(getAttributeString(message.substring(indexOfLastIdentifier, indexOfNextIdentifier)));
                indexOfLastIdentifier = indexOfNextIdentifier + 1;
                indexOfNextIdentifier = message.indexOf(SEP_VALUEPAIR, indexOfLastIdentifier);
            }
            arrays.addElement(getAttributeString(message.substring(indexOfLastIdentifier)));
            int nofArrays = arrays.size();
            String[][] attributes = new String[nofArrays][];

            for (int i = 0; i < nofArrays; i++) {
                attributes[i] = (String[])arrays.elementAt(i);
            }
            return attributes;
        }
        else {
            return null;
        }
    }

    /**
     * Locates and returns each <value>{","<value>} element
     * part of the [<value>{","<value>}{";"<value>{","<value>}}]
     * sequence in an array containing the values.
     *
     * @param attributeString The [<value>{","<value>}{";"<value>{","<value>}}] sequence.
     * @return An array containing each of the values.
     */
    private static String[] getAttributeString(String attributeString) {
        int before = 0;
        int after = attributeString.indexOf(SEP_VALUE);
        Vector vector = new Vector();
        while (after >= 0) {
            vector.addElement(attributeString.substring(before, after));
            before = after + 1;
            after = attributeString.indexOf(SEP_VALUE, before);
        }
        vector.addElement(attributeString.substring(before));
        int nofAttributes = vector.size();
        String[] attributes = new String[vector.size()];
        for (int i = 0; i < nofAttributes; i++) {
            attributes[i] = (String)vector.elementAt(i);
        }
        return attributes;
    }

    /**
     * Creates a string message in the specified format
     * from the given action identifier and values. The
     * string message is wrapped as an <code>Object</code>
     * to meet the interface's specification.
     *
     * @param action The action identifier.
     * @param values An array containing the message's values.
     * @return A string message to send to the server.
     */
    private String createStringMessage(Action action, Object[][] values) {
        StringBuffer buffer = new StringBuffer(action.getName()).append(SEP_ACTION);
        if (values != null) {
            int nofValues = values.length;
            for (int value_no = 0; value_no < nofValues; value_no++) {
                int nofAttributes = values[value_no].length;
                for (int attribute_no = 0; attribute_no < nofAttributes; attribute_no++) {
                    buffer.append(values[value_no][attribute_no]).append(SEP_VALUE);
                }
                buffer.setCharAt(buffer.length() - 1, SEP_VALUEPAIR);
            }
            buffer.deleteCharAt(buffer.length() - 1);
        }
        buffer.append(END_OF_MESSAGE);
        return buffer.toString();
    }
}
