package client.framework.views;

import client.framework.models.AbstractGame;
import client.framework.models.AbstractPlayer;
import client.framework.Utils;

import javax.microedition.lcdui.*;
import javax.microedition.lcdui.game.GameCanvas;
import javax.microedition.lcdui.game.LayerManager;
import javax.microedition.lcdui.game.Sprite;
import javax.microedition.lcdui.game.TiledLayer;
import java.util.Enumeration;
import java.util.Hashtable;

/**
 * The class responsible for running the game and displaying
 * the game board. Contains lists of objects to draw to the
 * game board, as well methods for detecting collisions and
 * handling movement.
 *
 * The class is also responsible for detecting in-game user
 * inputs and handling these.
 *
 * @version 2.0
 * @author Martin Jarrett
 * @author Eivind Sorteberg
 */
public abstract class GameView extends GameCanvas implements CommandListener {
    //The underlying game model
    protected AbstractGame model;

    //Keeps track of the different layers.
    protected LayerManager layerManager;

    //The background on the game board
    protected TiledLayer background;

    //Whether the player list should be drawn on top of the game board
    private boolean showStats;

    //Commands for exiting the game
    private Command exit;

    // The width and height of the game board
    protected int boardWidth;
    protected int boardHeight;

    // The background image on the game board
    private Image bgImage;

    // Variables used to center the game board on the phone's screen
    private int offsetY;
    private int offsetX;

    // Whether a game is currently running.
    private boolean running;

    /**
     * Constructor for the <code>GameView class</code>.
     * Initialises the game objects, creates the game board and
     * initialises the commands.
     *
     * @param model The underlying game model, which handles communication.
     */
    public GameView(AbstractGame model) {
        super(true);
        this.running = false;
        setFullScreenMode(true);

        this.model = model;

        this.layerManager = new LayerManager();
        this.bgImage = createBackgroundImage();

        //Initialises commands
        this.setCommandListener(this);
        this.exit = new Command("Lobby", Command.BACK, 0);
        this.addCommand(exit);
    }

    /**
     * Called whenever a change to the game board has been detected.
     * The game board is updated, and all game objects are redrawn.
     */
    public void fireBoardChanged() {
        this.drawBoard(getGraphics());
    }

    /**
     * Method called continously throughout a game. Checks whether the
     * local player collides with a wall, another player or a game object.
     * If a collision is detected, corresponding actions are carried out.
     */
    public void check() {
        specialCheck();                         //Game specific checks
        checkInput();                           //Detect user input
        moveOpponents();
        int[] movement = calculateMovement();   //Calculate the player's movement

        //If the player moves, detect collisions
        if (movement[0] != 0 || movement[1] != 0) {
            //Detect different sorts of collisions
            AbstractPlayer collidesWithPlayer = detectPlayerCollision(model.getSelf(), movement);
            Object collidesWithObject = detectObjectCollision(model.getSelf(), movement);
            int collidesWithWall = detectWallCollision(model.getSelf(), movement);

            //If no collision is detected, the player is free to move
            if (collidesWithObject == null && collidesWithPlayer == null &&
                    collidesWithWall == 0) {
                model.fireSelfMoved(movement);
            }
            //If collision is detected, actions must be taken
            else {
                model.flashAndVibrate();

                //Player has not moved, but movement will lead to collision.
                //Handle these collisions
                if (collidesWithPlayer != null) {
                    handlePlayerCollision(collidesWithPlayer, movement);
                }
                if (collidesWithObject != null) {
                    handleObjectCollision(collidesWithObject, movement);
                }
                if (collidesWithWall > 0) { // Detect which side of the client's brick collides with the wall
                    if ((collidesWithWall & (2 << 3)) > 0)  model.getSelf().setX(0);
                    if ((collidesWithWall & (2 << 2)) > 0)  model.getSelf().setX(boardWidth - model.getSelf().getSize());
                    if ((collidesWithWall & (2 << 1)) > 0)  model.getSelf().setY(0);
                    if ((collidesWithWall & (2     )) > 0)  model.getSelf().setY(boardHeight - model.getSelf().getSize());
                }
            }
        }
        this.fireBoardChanged();
    }

    /**
     * Returns the size of the game board.
     *
     * @return The size of the game board.
     */
    public int[] getSize() {
        return new int[]{boardWidth, boardHeight};
    }

    /**
     * Moves the opponents without receiving position updates.
     * Uses the movement prediction to move the players according
     * to their movement vectors.
     *
     * @see client.framework.models.AbstractPlayer#move()
     */
    private void moveOpponents() {
        Hashtable players = this.model.getPlayerlist();
        Enumeration ids = players.keys();
        while (ids.hasMoreElements()) {
            AbstractPlayer player = (AbstractPlayer)players.get(ids.nextElement());
            if (!player.equals(this.model.getSelf())) {
                player.move();
            }
        }
    }

    /**
     * Checks whether any keys have been pressed. For handling
     * key events other than directional keys, this method
     * should be overridden.
     *
     * In the default implementation, the only key detected is
     * the "FIRE" key. Pressing this key leads to the player list
     * being drawn on top of the game board.
     *
     * This method is currently a little bugged, and leads to the
     * player list flickering instead of being nicely displayed.
     *
     * @see GameView#check()
     */
    public void checkInput() {
        int keyState = this.getKeyStates();

        //Is "FIRE" pressed?
        if (((keyState & FIRE_PRESSED) != 0) != showStats) {
            showStats = !showStats;
        }
        if (showStats) {
            Graphics graphics = getGraphics();
            Utils.drawPlayerList(graphics, getWidth(), getHeight(), this.model, true);
            flushGraphics();
        }
    }

    /**
     * Draws the game board with its objects on the middle of
     * the mobile phone's screen.
     *
     * @param graphics The <code>GameCanvas</code>' <code>Graphics</code> object.
     */
    public void drawBoard(Graphics graphics) {
        graphics.drawImage(this.bgImage, 0, 0, 0);
        graphics.translate(offsetX, offsetY);

        graphics.setColor(0, 0, 0);
        graphics.drawRect(-1, -1, boardWidth + 1, boardHeight + 1);

        layerManager.paint(graphics, 0, 0);
        flushGraphics();
    }

    /**
     * Method inherited from the <code>CommandListener</code> interface. This
     * method is called whenever the user selects a command displayed
     * at the bottom of the screen. In this implementation, the only
     * command available is "Exit", which returns the player to the
     * lobby.
     *
     * @param command The command causing the method to be called.
     * @param displayable The currently displayed screen.
     */
    public void commandAction(Command command, Displayable displayable) {
        if (command.equals(exit)) {
            model.viewLobby();
        }
    }

    /**
     * Detects whether the local player collides with any of the other players.
     * If so, actions needs to be taken. The method returns the player
     * with which the local player collides if any, or null if no collision
     * is detected.
     *
     * @return The player with which a collision is detected, or null if no
     *  collision is detected.
     * @param mover The player trying to perform the movement.
     * @param movement The movement to take place.
     */
    protected AbstractPlayer detectPlayerCollision(AbstractPlayer mover, int[] movement) {
        //Gets a list of all players
        Hashtable players = model.getPlayerlist();
        Sprite moverSprite = mover.getSprite();
        moverSprite.move(movement[0], movement[1]); // Simulate the move

        AbstractPlayer collidesWith = null;

        Enumeration ids = players.keys();
        //For each of the players in the game..
        while (ids.hasMoreElements()) {
            AbstractPlayer player = (AbstractPlayer)players.get(ids.nextElement());

            if (!mover.equals(player)) {
                //..check for collision
                if (moverSprite.collidesWith(player.getSprite(), true)) {
                    collidesWith = player;
                    break;
                }
            }
        }
        moverSprite.move(-movement[0], -movement[1]);
        return collidesWith;
    }

    /**
     * Initialises the players' sprites. Makes use of each player's
     * <code>Sprite</code> object, and places these on the game board
     * using the <code>layerManager</code> object.
     *
     * @see AbstractPlayer#getSprite()
     */
    protected void initPlayers() {
        Hashtable players = model.getPlayerlist();
        //For each player in the player list..
        Enumeration ids = players.keys();
        while (ids.hasMoreElements()) {
            AbstractPlayer player = (AbstractPlayer)players.get(ids.nextElement());

            //..add the player's sprite to the board
            Sprite sprite = player.getSprite();
            layerManager.insert(sprite, 0);
        }
    }

    /**
     * Abstract method that is used for defining the background
     * of the game board. This method should return a single image
     * that fills the entire screen.
     *
     * @return The background image for the game board.
     */
    protected abstract Image createBackgroundImage();

    /**
     * Method called each time the <code>check()</code> method is
     * called. This method should contain the game specific checks
     * that need to be performed as long as the game is running.
     *
     * @see GameView#check()
     */
    protected abstract void specialCheck();

    /**
     * This method is called whenever a collision with another player is
     * detected, and is responsible for handling this event. The move
     * has not yet been performed, but the <code>movement</code>
     * parameter contains the move that will cause the collision if
     * it is carried out.
     *
     * @param collidesWith The player with which a collision is detected.
     * @param movement The move causing the collision
     * @see GameView#detectPlayerCollision(client.framework.models.AbstractPlayer, int[])
     */
    protected abstract void handlePlayerCollision(AbstractPlayer collidesWith, int[] movement);

    /**
     * This method is called whenever a collision with a game specific
     * object is detected, and is responsible for handling this event. The
     * move has not yet been performed, but the <code>movement</code>
     * parameter contains the move that would cause the collision if
     * it is carried out.
     *
     * @param collidesWith The object with which a collision is detected.
     * @param movement The move causing the collision.
     * @see GameView#detectObjectCollision(client.framework.models.AbstractPlayer, int[])
     */
    public abstract void handleObjectCollision(Object collidesWith, int[] movement);

    /**
     * This method is called continuously throughout the game in order
     * to detect collisions with game objects other than players. The method
     * should run through all objects active in the game and check whether
     * a collision is detected.
     *
     * If a collision is detected, the method should return the colliding object,
     * if not it should return null.
     *
     * @param player The player trying to move
     * @param movement The movement to take place.
     * @return The object with which a collision is detected.
     */
    public abstract Object detectObjectCollision(AbstractPlayer player, int[] movement);

    /**
     * Checks whether any of the movement keys have been pressed.
     * This method handles the default movement, and enables eight
     * movement directions.
     *
     * The returned array is a vector representing the movement in
     * horizontal and vertical direction, respectively.
     *
     * @see GameView#check()
     * @return The player's movement direction.
     */
    private int[] calculateMovement() {
        int keyState = this.getKeyStates();

        //Default is no movement
        int x = 0;
        int y = 0;

        //Detects movement
        if ((keyState & LEFT_PRESSED) != 0)     x--;
        if ((keyState & RIGHT_PRESSED) != 0)    x++;
        if ((keyState & UP_PRESSED) != 0)       y--;
        if ((keyState & DOWN_PRESSED) != 0)     y++;

        int speed = model.getSelf().getSpeed();

        return new int[]{x * speed, y * speed};
    }

    /**
     * Checks whether the player collides with one of the four walls.
     * If so, actions should be taken (usually denying the player the
     * move).
     *
     * @param player The player trying to move
     * @param movement The movement to take place
     * @see GameView#check()
     * @return A four digit binary representation of the collisions, in the order WENS.
     */
    protected int detectWallCollision(AbstractPlayer player, int[] movement) {
        //Get's the local player's sprite
        Sprite sprite = player.getSprite();

        //Determines the player's (new) position
        int xPos = sprite.getX() + movement[0];
        int yPos = sprite.getY() + movement[1];

        //Determines the player's size
        int width = sprite.getWidth();
        int height = sprite.getHeight();

        int collisions = 0;
        if (xPos < 0)                           collisions |= (2 << 3);
        if (xPos > this.boardWidth - width)     collisions |= (2 << 2);
        if (yPos < 0)                           collisions |= (2 << 1);
        if (yPos > this.boardHeight - height)   collisions |= 2;

        return collisions;
    }

    /**
     * Removes the specified player from the game board. Is
     * called whenever a player is disconnected from the session.
     *
     * @param player The player to remove from the game board.
     */
    public void removePlayer(AbstractPlayer player) {
        this.layerManager.remove(player.getSprite());
        this.fireBoardChanged();
    }

    /**
     * Adds the specified player to the game board. Called whenever a new
     * player connects to the session.
     *
     * @param player The player to add to the game board.
     */
    public void notifyAboutPlayerAdded(AbstractPlayer player) {
        this.layerManager.insert(player.getSprite(), 0);
    }

    /**
     * Starts a new game with the specified size of the
     * game board.
     *
     * @param screenResolution An array containing the height and width of the game board.
     */
    public void startGame(int[] screenResolution) {
        this.setFullScreenMode(true);
        this.running = true;
        this.boardWidth = screenResolution[0];
        this.boardHeight = screenResolution[1];

        this.offsetX = (getWidth() - boardWidth) / 2;
        this.offsetY = (getHeight() - boardHeight) / 2;

        initPlayers();
    }

    /**
     * Checks whether a game is currently running.
     *
     * @return True if a game is running, false if not.
     */
    public boolean isRunning() {
        return this.running;
    }

    /**
     * Stops the current game, and sets the running parameter to false.
     * Calls the <code>clean()</code> method to release occupied resources.
     */
    public void stopGame() {
        this.running = false;
        this.clean();
    }

    /**
     * Abstract method that should contain functionality for releasing
     * occupied resources and resetting the game view's state.
     */
    protected abstract void clean();
}