package client.framework.views;

import client.framework.models.AbstractGame;
import client.framework.Utils;

import javax.microedition.lcdui.*;
import javax.microedition.lcdui.game.GameCanvas;

/**
 * The screen shown as default when starting the application.
 * This screen does not contain any information such as
 * player lists, but gets the information its draws through
 * the <code>AbstractGame</code> object.
 *
 * In addition to drawing contents to the screen, this class
 * is responsible for detecting and handling user inputs
 * when this screen is active.
 *
 * @author Martin Jarrett
 * @author Eivind Sorteberg
 * @version 2.0
 */
public class LobbyView extends GameCanvas implements CommandListener {
    //Commands
    private Command startGame;
    private Command exit;
    private Command ready;
    private Command changeSettings;

    //The model containing game information
    private AbstractGame model;
    private Command returnToGame;

    /**
     * Contructor for this class. Initialises the
     * commands and sets the player list as default
     * start-up view.
     *
     * @param model The game model.
     */
    public LobbyView(AbstractGame model) {
        super(false);
        this.model = model;
        this.setCommandListener(this);
        this.activateCommands();
    }

    /**
     * Method implemented from <code>CommandListener</code>. Called
     * whenever the user selects one of the commands, and performs
     * corresponding actions. These actions are either a change of
     * state in this object, or a method found in
     * <code>AbstractGame</code>.
     *
     * @param command The command selected by the user.
     * @param displayable This screen.
     * @see LobbyView#paint(javax.microedition.lcdui.Graphics)
     */
    public void commandAction(Command command, Displayable displayable) {
        if (command == ready)          this.model.fireReadySelected();
        else if (command == startGame)      this.model.startGame();
        else if (command == changeSettings) this.model.fireChangeSettingsSelected();
        else if (command == exit)           this.model.quit();
        else if (command == returnToGame) this.model.fireReturnToGameSelected();
    }

    /**
     * Overrides the class' default paint method and is called whenever
     * the contents of the Lobby change. Dependent
     * of the current state of the Lobby, different contents are drawn.
     * The three possible states are to show players (default), messages
     * or settings.
     *
     * @param graphics The default graphics object used for this GameCanvas
     */
    public void paint(Graphics graphics) {
        super.paint(graphics);

        //Default font for the lobby
        graphics.setFont(Font.getFont(Font.FACE_MONOSPACE, Font.STYLE_PLAIN, Font.SIZE_SMALL));

        //Draw background
        graphics.setColor(255, 255, 255);
        graphics.fillRect(0, 0, getWidth(), getHeight());

        // Draws the player list
        Utils.drawPlayerList(graphics, getWidth(), getHeight(), this.model, false);

    }

    /**
     * Initiates and attaches the commands seen in this screen.
     * Called on class initialisation.
     */
    private void activateCommands() {
        this.startGame =        new Command("Start game", "Start",      Command.SCREEN, 2);
        this.exit =             new Command("Exit", "Exit",            Command.EXIT,   5);
        this.ready =            new Command("Ready", "Ready",           Command.OK,     6);
        this.changeSettings =   new Command("Change settings", "Settings",  Command.SCREEN, 9);
        this.returnToGame =           new Command("Return to game", "Return", Command.OK, 11);

        this.addCommand(startGame);
        this.addCommand(exit);
        this.addCommand(ready);
        this.addCommand(changeSettings);
    }

    public void setGameStarted(boolean started) {
        if (started) {
            this.removeCommand(changeSettings);
            this.removeCommand(ready);
            this.removeCommand(startGame);
            this.addCommand(returnToGame);
        }
        else {
            this.addCommand(changeSettings);
            this.addCommand(ready);
            this.addCommand(startGame);
            this.removeCommand(returnToGame);
        }
    }

    public void notifyAboutPlayerListChanged() {
        this.repaint();
    }
}
