package client.framework.views;

import client.framework.models.AbstractGame;

import javax.microedition.lcdui.*;
import java.util.Vector;

/**
 * A <code>Form</code> extension that lists the currently active
 * on the server when the client has connected to the server. The
 * user may select to join one of the listed sessions, or to
 * create a new session.
 *
 * @version 1.0
 * @author Eivind Sorteberg
 * @author Martin Jarrett
 */
public class SessionWindow extends Form implements CommandListener {
    // The client's main model
    private AbstractGame model;

    // Commands
    private Command select;
    private Command refresh;
    private Command exit;

    // The session list
    private ChoiceGroup choices;

    // Timer used to notify the user when a connection with the server can not be established
    private ConnectionTimer connectionTimer;

    /**
     * Creates the session window.
     *
     * @param model The main model for the client application.
     */
    public SessionWindow(AbstractGame model) {
        super("Available sessions");
        this.connectionTimer = new ConnectionTimer();
        this.connectionTimer.start();
        this.model = model;

        this.select = new Command("Connect", Command.OK, 0);
        this.refresh = new Command("Refresh", "Refresh list", Command.SCREEN, 1);
        this.exit = new Command("Exit", "Exit", Command.EXIT, 2);
        this.setCommandListener(this);
        this.addCommand(refresh);
        this.addCommand(exit);
    }

    /**
     * Called when the client has received a (possibly empty) list of
     * sessions from the server. The sessions are shown on the screen, and
     * the user may select to connect to an active session, or
     * to start a new session.
     */
    private void showSessions() {
        this.connectionTimer.stopSearch();
        this.deleteAll();
        this.removeCommand(select);
        Vector sessions = model.getSessions();
        String[] sessionNames = new String[sessions.size()];
        for (int i = 0; i < sessionNames.length; i++) {
            sessionNames[i] = sessions.elementAt(i).toString();
        }
        if (sessionNames.length > 0) {
            this.choices = new ChoiceGroup("", Choice.EXCLUSIVE, sessionNames, null);
            this.append(choices);
            this.addCommand(select);
        }
    }

    /**
     * Called when the user selects one of the available commands. The
     * result of this depends on the command selected.
     *
     * @param command The command the user selected.
     * @param displayable Not used.
     */
    public void commandAction(Command command, Displayable displayable) {
        if (command.equals(this.select)) { // Connect to the selected session
            int selectedSession = choices.getSelectedIndex();
            this.model.fireSessionSelected(selectedSession);
        }
        else if (command.equals(this.refresh)) { // Refresh the session list
            this.connectionTimer = new ConnectionTimer();
            connectionTimer.start();
            deleteAll();
            this.model.refreshSessionList();
        }
        else if (command.equals(this.exit)) { // Close the application
            this.model.quit();
        }
    }

    /**
     * Called when a new list of sessions is received. Refreshes the session list.
     */
    public void fireSessionsUpdated() {
        this.showSessions();
    }

    /**
     * Thread that is used to notify the user when a connection to the server
     * can not be established. Runs for 10 seconds; if a session list has not
     * been received within this time, a message is displayed that tells
     * the user to try refreshing the list, or to restart the application.
     */
    private class ConnectionTimer extends Thread {
        private StringItem message;
        private boolean searching;

        public ConnectionTimer() {
            this.message = new StringItem("", "");
        }

        public void run() {
            int nofDots = 1;
            long stopTime = System.currentTimeMillis() + 10000;
            deleteAll();
            append(message);
            this.searching = true;
            removeCommand(refresh);
            while (this.searching) {
//                deleteAll();
                nofDots = ((nofDots + 1) % 4);

                StringBuffer label = new StringBuffer("Connecting to server.");
                for (int i = 0; i < nofDots; i++) {
                    label.append(".");
                }
                message.setLabel(label.toString());
                
//
                try {
                    // Causes the "refresh" command to appear up to 500 ms late 
                    Thread.sleep(500);
                } catch (InterruptedException e) {
                    /* Do nothing */
                }
                if (System.currentTimeMillis() >= stopTime && this.searching) {
                    this.message.setLabel("Connection failed");
                    this.message.setText("Connection to server could not be established. " +
                            "Try refreshing the session list. If this does not work, " +
                            "please restart the application and check your server settings.");
                    this.searching = false;
                }
            }
            addCommand(refresh);
        }

        public void stopSearch() {
            this.searching = false;
        }
    }
}
