package client.framework.views;

import client.framework.models.AbstractGame;

import javax.microedition.lcdui.*;
import java.util.Enumeration;
import java.util.Hashtable;

/**
 * This class is a simple form that shows a list
 * of the currently active settings of the game, and
 * enables editing these. An object from this class is
 * only used and displayed if the user selects "Change
 * Settings" in the <code>LobbyView</code> class. When
 * the user have changed the wanted settings, he clicks
 * "OK" and returns to the lobby.
 *
 * @author Eivind Sorteberg
 * @author Martin Jarrett
 * @version 1.0
 */
public class SettingsWindow extends Form implements CommandListener {
    //The containers for the settings
    private Item[] items;

    //To select when done
    private Command ok;

    //The game model
    private AbstractGame model;

    /**
     * Called when a screen for editing settings is needed.
     * This constructor creates a form and attaches components
     * to this form so that the user may edit the settings.
     * When done, the user clicks "OK" and returns to the lobby.
     *
     * @param model The model containing the settings.
     */
    public SettingsWindow(AbstractGame model) {
        super("Settings Editor");
        this.model = model;

        //For when done
        this.ok = new Command("OK", Command.OK, 0);
        this.setCommandListener(this);
        this.addCommand(ok);

        //Place the settings components
        this.placeSettings();
    }

    /**
     * Method implemented from CommandListener. Is called
     * when the user clicks the "OK" command, and closes
     * the form. In this process, the list of settings is
     * sent to the model to check whether any settings have
     * been changed.
     *
     * @param command The only possible command is "OK".
     * @param displayable This form.
     * @see AbstractGame#fireSettingsUpdated(java.util.Hashtable)
     */
    public void commandAction(Command command, Displayable displayable) {
        if (command == ok) {
            closeWindow();
        }
    }

    /**
     * Gets the settings from the game model, creates
     * an appropriate editable component for each of
     * these settings and adds the component to the
     * form.
      */
    private void placeSettings() {
        //Get the settings
        Hashtable settings = model.getSettings();

        //Array for the different components
        this.items = new Item[settings.size()];

        int index = 0;
        Enumeration names = settings.keys();

        //For each setting..
        while (names.hasMoreElements()) {
            //Get the name and value..
            String name = (String)names.nextElement();
            Object value = settings.get(name);

            //..and add the setting to the list
            appendSetting(name, value, index);
            index++;
        }
    }

    /**
     * Attaches a setting component to the form, as well
     * as adding it to the list of settings. This method
     * is called for each of the settings when this object
     * is initialised.
     *
     * @param name The name of the setting.
     * @param value The setting's value.
     * @param index The setting's position in the list of settings.
     */
    private void appendSetting(String name, Object value, int index) {
        //The item to contain the setting's info
        Item item = null;

        //Check the type of setting, and display an appopriate type of item
        if (value instanceof String) {
            item = new TextField(name, (String)value, 5, 0);
        }
        else if (value instanceof Integer) {
            item = new TextField(name, Integer.toString(((Integer)value).intValue()), 3,  TextField.NUMERIC);
        }
        else if (value instanceof Boolean) {
            item = new ChoiceGroup(name, Choice.MULTIPLE);
            ((ChoiceGroup)item).append("On", null);

            if (((Boolean)value).booleanValue()) ((ChoiceGroup)item).setSelectedIndex(0, true);
            else ((ChoiceGroup)item).setSelectedIndex(0, false);
        }

        //Add the setting to the list, as well as the form
        items[index] = item;
        this.append(item);
    }

    /**
     * Called when the window is to be closed. This happens when the user selects
     * the "OK" command. The settings are then sent to <code>AbstractGame</code> and
     * checked for changes, and the user is returned to the lobby.
     */
    private void closeWindow() {
        Object value = null;
        Hashtable settings = new Hashtable();

        //For each setting..
        for (int i = 0; i < items.length; i++) {

            //Create the appopriate value object
            if (items[i] instanceof ChoiceGroup) {
                boolean selected[] = new boolean[((ChoiceGroup)items[i]).size()];
                ((ChoiceGroup)items[i]).getSelectedFlags(selected);
                value = selected[0] ? Boolean.TRUE : Boolean.FALSE;
            }
            else if (items[i] instanceof TextField) {
                TextField field = (TextField)items[i];

                String content = field.getString();

                if (field.getConstraints() == TextField.NUMERIC) {
                    value = new Integer(Integer.parseInt(content));
                }
                else value = content;
            }

            //..add the setting to a list..
            settings.put(items[i].getLabel(), value);
        }
        //..and send the list to the game model
        model.fireSettingsUpdated(settings);
    }
}
