package client.framework.views;

import client.framework.models.AbstractGame;
import client.framework.Utils;

import javax.microedition.lcdui.*;
import javax.microedition.rms.RecordEnumeration;
import javax.microedition.rms.RecordStore;
import javax.microedition.rms.RecordStoreException;
import java.util.Hashtable;
import java.util.Enumeration;

/**
 * This method is called when the MIDlet is started.
 * It only contains a form where the user can type his
 * name. After selecting "OK", the user is taken to the
 * <code>LobbyView</code> screen, and the framework is
 * intialised.
 *
 * @author Eivind Sorteberg
 * @author Martin Jarrett
 * @version 2.0
 */
public class StartupWindow extends Form implements CommandListener {
    public static final String UDP = "UDP";
    public static final String TCP = "TCP";

    public static final String NAME = "name";
    public static final String IP = "ip";
    public static final String PORT = "port";
    public static final String COMM = "comm";
    public static final String SIZE = "size";

    //The game model
    private AbstractGame model;

    //The field to type the name
    private TextField namefield;

    private TextField serverIpField;
    private TextField serverPortField;
    private ChoiceGroup transferProtocol;
    private TextField packetSize;

    private Command connect;

    private RecordManager recordManager;

    /**
     * The constructor creates text fields for typing the player name
     * (10 characters), the server address, and the server port. In addition,
     * the user has a choice between using UDP or TCP for the communication.
     *
     * @param model The underlying game model
     */
    public StartupWindow(AbstractGame model) {
        super("");
        this.model = model;

        this.recordManager = new RecordManager();
        Hashtable preferences = recordManager.readPreferences();

        //Initialise the fields
        this.namefield = new TextField("Name:", "", 10, TextField.INITIAL_CAPS_SENTENCE);
        this.namefield.setString((String)preferences.get(NAME));
        this.append(namefield);

        this.serverIpField = new TextField("Server:", (String)preferences.get(IP), 15, TextField.NON_PREDICTIVE);
        this.append(serverIpField);

        this.serverPortField = new TextField("Port:", (String)preferences.get(PORT), 5, TextField.NUMERIC);
        this.append(serverPortField);

        this.transferProtocol = new ChoiceGroup("Transfer Protocol:", ChoiceGroup.EXCLUSIVE);
        this.transferProtocol.append(UDP, null);
        this.transferProtocol.append(TCP, null);
        this.append(transferProtocol);
        this.transferProtocol.setSelectedFlags(new boolean[]{UDP.equals(preferences.get(COMM)), TCP.equals(preferences.get(COMM))});

        this.packetSize = new TextField("Packet Size:", (String)preferences.get(SIZE), 6, TextField.NUMERIC);
        this.append(packetSize);

        //Adds the "OK" command
        this.connect = new Command("Connect", Command.OK, 1);
        this.addCommand(this.connect);
        this.setCommandListener(this);
    }

    /**
     * Called when the user selects the "OK" command. This
     * method initialises the framework, and takes the user
     * to the lobby view.
     *
     * @param command Can only be the "OK" command.
     * @param displayable This <code>Form</code>.
     * @see client.framework.models.AbstractGame#startFramework(String, String[], int)
     */
    public void commandAction(Command command, Displayable displayable) {
        if (command.getCommandType() == Command.OK) {
            this.removeCommand(connect);
            String name = this.namefield.getString();
            String serverIp = this.serverIpField.getString();
            String serverPort = this.serverPortField.getString();
            String transferProtocol = this.transferProtocol.getString(this.transferProtocol.getSelectedIndex());
            String packetSize = this.packetSize.getString();

            String address = "";
            if (TCP.equals(transferProtocol)) {
                address = "socket";
            }
            else if (UDP.equals(transferProtocol)) {
                address = "datagram";
            }
            address = Utils.buildString(new Object[]{address, "://", serverIp, ":", serverPort});

            Hashtable preferences = new Hashtable(5);
            preferences.put(NAME, name);
            preferences.put(IP, serverIp);
            preferences.put(PORT, serverPort);
            preferences.put(COMM, transferProtocol);
            preferences.put(SIZE, packetSize);

            recordManager.writePreferences(preferences);
            model.startFramework(name, new String[]{address}, Integer.parseInt(packetSize));
        }
    }

    /**
     * Class that is used for persistent storage on the mobile phone. By using
     * a <code>RecordStore</code>, the application remembers the values entered
     * by the user last time, so that the user does not have to enter these values
     * every time he connects to a server.
     */
    private class RecordManager {
        // Default values before the user has entered his own values
        public static final String DEFAULT_NAME = "Player";
        public static final String DEFAULT_SERVER = "";
        public static final String DEFAULT_PORT = "15000";     // RTSP port
        public static final String DEFAULT_COMMUNICATION = UDP;
        public static final String DEFAULT_PACKET_SIZE = "128";

        public static final String STORE_NAME = "Preferences";
        public static final char SEPARATOR = ':';
        private RecordStore store;

        /**
         * Initialises the record store.
         */
        public RecordManager() {
            this.store = openRecordStore();
        }

        /**
         * Opens the record store.
         *
         * @return The record store used to save the user's preferences.
         */
        public RecordStore openRecordStore() {
            try {
                return RecordStore.openRecordStore(STORE_NAME, true);
            }
            catch (RecordStoreException e) {
                e.printStackTrace();
                return null;
            }
        }

        /**
         * Reads the previous values from the record store. If no such values
         * are found, the default values are returned.
         *
         * @return A <code>Hashtable</code> containing the user's preferences.
         */
        public Hashtable readPreferences() {
            Hashtable preferences = new Hashtable();
            RecordEnumeration re = null;
            try {
                re = store.enumerateRecords(null, null, false);
                while (re.hasNextElement()) {
                    String record = new String(re.nextRecord());
                    int separatorindex = record.indexOf(SEPARATOR);
                    if (separatorindex > 0) {
                        String name = record.substring(0, separatorindex);
                        String value = record.substring(separatorindex + 1);
                        preferences.put(name, value);
                    }
                }
            } catch (RecordStoreException e) {
                preferences.clear();
            } finally {
                if (re != null) {
                    re.destroy();
                }
            }
            if (!(preferences.size() > 0)) {
                preferences.put(NAME, DEFAULT_NAME);
                preferences.put(IP, DEFAULT_SERVER);
                preferences.put(PORT, DEFAULT_PORT);
                preferences.put(COMM, DEFAULT_COMMUNICATION);
                preferences.put(SIZE, DEFAULT_PACKET_SIZE);
            }
            return preferences;
        }

        /**
         * Writes the user's preferences to the record store when the user
         * has tried connecting to the server. Replaces any existing values
         * in the record store.
         *
         * @param preferences The values used to connect to the server.
         */
        public void writePreferences(Hashtable preferences) {
            try {
                deleteAll();
                Enumeration enumeration = preferences.keys();
                while (enumeration.hasMoreElements()) {
                    String key = (String)enumeration.nextElement();
                    String preference = Utils.buildString(new Object[]{key, new Character(SEPARATOR), preferences.get(key)});
                    byte[] raw = preference.getBytes();
                    store.addRecord(raw, 0, raw.length);
                }
            } catch (RecordStoreException e) {
                /* Do nothing */
            }
        }

        /**
         * Removes all values stored in the record store.
         */
        public void deleteAll() {
            RecordEnumeration re = null;
            try {
                re = store.enumerateRecords(null, null, false);
                while (re.hasNextElement()) {
                    store.deleteRecord(re.nextRecordId());
                }
            } catch (RecordStoreException e) {
                e.printStackTrace();
            } finally {
                if (re != null) {
                    re.destroy();
                }
            }
        }

        /**
         * Closes the record store. Used to clean up and release resources.
         */
        public void close() {
            if (this.store != null) {
                try {
                    this.store.closeRecordStore();
                }
                catch (RecordStoreException e) {
                    /* Do nothing */
                }
            }
        }
    }
}
