package server.brickblock.fwimpl;

import server.framework.threads.EventHandler;
import server.framework.threads.Event;
import server.framework.models.AbstractPlayer;
import server.framework.models.Setting;
import server.framework.network.parser.Action;
import server.brickblock.objects.Powerup;
import server.brickblock.objects.SizeUp;
import server.brickblock.objects.SpeedUp;
import server.brickblock.objects.StrengthUp;

import java.util.*;

/**
 * Thread that is responsible for generating traps and power up
 * objects, as well as player positions on game initiation. In addition, this
 * class is responsible for handling game events.
 *
 * @version 1.0
 * @author Martin Jarrett
 * @author Eivind Sorteberg
 */
public class BBEventHandler implements EventHandler<BBPlayer, BBSession> {
    // The mean time between power ups
    public static final int POWERUP_INTERVAL_SECS = 10;

    // The duration of power ups
    public static final int POWERUP_DURATION_SECS = 10;

    // When the game should stop
    private long stopTime;

    //When should the next power up be created?
    private long nextPowerup;

    // The position of the trap
    private int[] trapPosition;

    // The power ups currently on the game board
    private HashMap<Powerup, Long> powerups;

    // The session to which this thread is assigned
    private BBSession session;

    // Whether a game is currently running
    private boolean running;

    // If the thread has completed its pre-game calculations
    private boolean ready;

    /**
     * Creates the <code>BBEventHandler</code> object. Initialises
     * the power up list and generates the trap position.
     *
     * @param session This thread's session.
     */
    public BBEventHandler(BBSession session) {
        this.ready = false;
        this.session = session;
        this.powerups = new HashMap<Powerup, Long>();
        this.trapPosition = new int[]{
                20 + (int)(Math.random() * (session.getMinimumWidth() - 50)),
                20 + (int)(Math.random() * (session.getMinimumHeight() - 50))
        };
        calculateNextPowerup();
    }

    /**
     * Resets the scores of all players. Called when a new game is started.
     */
    private void resetScores() {
        for (AbstractPlayer player : session.getPlayers()) {
            player.setScore(0);
        }
    }

    /**
     * The loop where power up objects are created, and the players'
     * power up objects are controlled. This loop is started when a
     * game is initiated, and stopped when the game is over.
     */
    public void run() {
        this.running = true;
        this.resetGame();

        // Send the trap position
        this.session.sendAction(
                Action.trap_added,
                new Object[][]{new Object[]{
                        trapPosition[0],
                        trapPosition[1]
                }},
                true);

        // Determine the duration of the game
        int duration = (Integer)this.session.getSettings().getSetting(Setting.time) * 1000;
        this.stopTime = duration > 0 ? System.currentTimeMillis() + duration : Long.MAX_VALUE;
        this.ready = true;

        while (running) {
            this.handlePowerupGeneration();
            this.handlePlayerPowerups();
            this.checkSettings();

            try {
                Thread.sleep(200);
            }
            catch (InterruptedException e) {
                e.printStackTrace();
            }
        }
    }

    /**
     * Resets the scores of all players, and generates positions
     * for all the session's players. Called when a new game is
     * started.
     */
    private void resetGame() {
        this.resetScores();
        this.generatePositions(session.getPlayers());
    }

    /**
     * Loops through the players' power up objects to see if
     * any of the power ups have timed out. If so, a notification
     * of the event is transmitted to the connected clients.
     */
    private void handlePlayerPowerups() {
        for (BBPlayer player : session.getPlayers()) {
            for (Powerup powerup : player.getInactivePowerups()) {
                Object[][] values = new Object[][]{{
                        powerup.getName(),
                        powerup.getPosition()[0],
                        powerup.getPosition()[1],
                        powerup.getAffectedAttribute(player)
                }};
                session.sendAction(Action.powerup_inactive, values, player, false, true);
            }
        }
    }

    /**
     * Generates a new random power up object if the time passed
     * since the last power up generation is long enough. Also checks
     * if any of the power ups on the game board have timed out
     * and should be removed.
     */
    private void handlePowerupGeneration() {
        if (nextPowerup < System.currentTimeMillis()) {
            //adds a random powerup to the board
            this.addRandomPowerup();

            //determines the time for the next power up
            this.calculateNextPowerup();

        }
        // Checks if any of the power ups have been on the board for too long
        this.getTimedOutPowerups();
    }

    /* METHODS FOR CHECKING SETTINGS */

    /**
     * Checks if the game has timed out, or if the number
     * of players connected to the session is too large. If
     * so, the game is stopped.
     */
    public void checkSettings() {
        this.checkTime();
        this.checkNofPlayers();
    }

    /**
     * Checks the number of players.
     */
    private void checkNofPlayers() {
        int maxPlayers = (Integer)this.session.getSettings().getSetting(Setting.nof_players);
        if (this.session.getPlayers().size() > maxPlayers) {
            this.session.stopGame("Too many players: " + this.session.getPlayers().size());
        }
    }

    /**
     * Checks if the game has timed out.
     */
    private void checkTime() {
        long currentTime = System.currentTimeMillis();
        if (currentTime > stopTime) {
            this.session.stopGame("Time is up!");
        }
    }

    /**
     * Checks if any of the players have reached the score limit. If so,
     * the game is ended. This method is called whenever a player has
     * scored.
     *
     * @param player The player that has scored.
     */
    private void checkScore(BBPlayer player) {
        int scoreLimit = (Integer)this.session.getSettings().getSetting(Setting.score);
        if (Math.abs(player.getScore()) >= scoreLimit) {
            this.session.stopGame(player.getName() + " lost the game!");
        }
    }

    /* ******************************************/

    /**
     * Loops through the power ups on the game board to see
     * if any have timed out and should be removed. If so, a
     * notification is sent to the clients.
     */
    private void getTimedOutPowerups() {
        long currentTime = System.currentTimeMillis();
        Set<Powerup> toRemove = new HashSet<Powerup>();
        for (Powerup powerup : this.powerups.keySet()) {
            long timeout = this.powerups.get(powerup);
            if (timeout <= currentTime) {
                toRemove.add(powerup);
            }
        }
        for (Powerup powerup : toRemove) {
            this.powerups.remove(powerup);
            this.firePowerupRemoved(powerup);
        }
    }

    /**
     * Not implemented.
     *
     * @param player Not used.
     */
    public void fireEventOccured(AbstractPlayer player) {
    }

    /**
     * Called when a power up object has been generated. Sends
     * a notification to the connected clients.
     *
     * @param powerup The new power up object.
     */
    public void firePowerupAdded(Powerup powerup) {
        this.session.sendAction(Action.powerup_added, new Object[][]{powerup.getAttributes()}, false);
    }

    /**
     * Called when a power up object has been removed. Sends
     * a notification to the connected clients.
     *
     * @param powerup The power up that is to be removed.
     */
    public void firePowerupRemoved(Powerup powerup) {
        this.session.sendAction(Action.powerup_removed, new Object[][]{powerup.getAttributes()}, false);
    }

    /**
     * Sets this thread's session.
     *
     * @param session This thread's session.
     */
    public void setSession(BBSession session) {
        this.session = session;
    }

    /**
     * Generates a random powerup and adds it to the board.
     */
    private void addRandomPowerup() {
        Powerup powerUp;
        int[] position = generatePosition(Powerup.DEFAULT_SIZE);
        int x = position[0];
        int y = position[1];

        switch ((int) (Math.random() * 3)) {
            case 0:
                powerUp = new SizeUp(x, y, Powerup.DEFAULT_DURATION_SECS);
                break;
            case 1:
                powerUp = new StrengthUp(x, y, Powerup.DEFAULT_DURATION_SECS);
                break;
            case 2:
                powerUp = new SpeedUp(x, y, Powerup.DEFAULT_DURATION_SECS);
                break;
            default:
                powerUp = null;
                break;
        }

        this.addPowerup(powerUp);
        this.firePowerupAdded(powerUp);
    }

    /**
     * Calculates when a new power up object shall be generated.
     */
    private void calculateNextPowerup() {
        nextPowerup =   System.currentTimeMillis() +                          // now
                        POWERUP_INTERVAL_SECS * 1000 / 2 +                    // at least half interval time
                        (long)(Math.random() * POWERUP_INTERVAL_SECS * 1000); // result: [-interval, interval] from now
    }

    /**
     * Adds a new power up object to the list of power ups.
     *
     * @param powerup The new power up.
     */
    private void addPowerup(Powerup powerup) {
        this.powerups.put(powerup, System.currentTimeMillis() + POWERUP_DURATION_SECS * 1000);
    }

    /**
     * Generates a random position within the game board.
     *
     * @param size The size of the player object.
     * @return A random position.
     */
    private int[] generatePosition(int size) {
        int x = (int) (Math.random() * (session.getMinimumWidth() - size));
        int y = (int) (Math.random() * (session.getMinimumHeight() - size));

        return new int[]{x, y};
    }

    /**
     * Stops the thread. Called when the game is over.
     */
    public void stop() {
        this.running = false;
    }

    /**
     * Called when a game related event has occurred. The action that is taken
     * depends on the event that has occurred.
     *
     * @param player The player that caused the event.
     * @param event The event.
     * @param values Values related to the event.
     */
    public void fireEventOccured(BBPlayer player, Event event, Object[] values) {
        switch (event) {
            case score:
                notifyAboutPlayerScored(player);
                break;
            case join:
                Object message = session.getServer().getParser().createMessage(
                    Action.trap_added,
                    new Object[][]{{
                        trapPosition[0],
                        trapPosition[1]
                    }},
                    null);
                player.sendMessage(message, true);
                break;
            case powerup_activated:
                String name = values[0].toString();
                int x = Integer.parseInt(values[1].toString());
                int y = Integer.parseInt(values[2].toString());
                Powerup powerup = findPowerup(name, x, y);
                if (powerup != null) {
                    this.session.sendAction(
                            Action.powerup_removed,
                            new Object[][]{powerup.getAttributes()},
                            player,
                            false,
                            true
                    );
                    player.pickUpPowerup(powerup);
                    this.powerups.remove(powerup);
                }
                break;
        }
    }

    /**
     * Called when a player has scored. Reduces the player's score with one point,
     * and generates a new position for the player. The connected clients are
     * then notified of the event.
     *
     * @param player The player that scored.
     */
    private void notifyAboutPlayerScored(BBPlayer player) {
        player.setScore(player.getScore() - 1);
        int[] randomCorner = findAvailableCorner(player, session.getPlayers(), session.getMinimumWidth(), session.getMinimumHeight());
        player.setPosition(randomCorner);
        this.session.firePlayerListChanged();
        this.checkScore(player);

        // Send the player's new score to all participants
        session.sendAction(
                Action.score,
                new Object[][]{{player.getScore(), player.getPosition()[0], player.getPosition()[1]}},
                player,
                false,
                true
        );
    }

    /**
     * Loops through the list of power ups and returns the power up with
     * the specified attributes. Called when a player tries to pick up a
     * power up.
     *
     * @param name The type of power up.
     * @param x The power up's x-position.
     * @param y The power up's y-position.
     * @return The power up with the specified attributes, if it is still available.
     */
    private Powerup findPowerup(String name, int x, int y) {
        for (Powerup powerup : this.powerups.keySet()) {
            if (name.equals(powerup.getName()) && x == powerup.getPosition()[0] && y == powerup.getPosition()[1]) {
                return powerup;
            }
        }
        return null;
    }

/* POSITION GENERATION METHODS */

    /**
     * Generates random positions for the players given in the list.
     * Called when a new game is started, and the players are to be
     * placed on the game board.
     *
     * @param players The list of players.
     */
    public void generatePositions(final List<BBPlayer> players) {
        for (int i = 0; i < players.size(); i++) {
            int[] position = BBEventHandler.generateRandomPosition(this.session.getMinimumWidth(), this.session.getMinimumHeight());
            if (!positionIsOccupied(position, players, i)) {
                BBPlayer player = players.get(i);
                this.session.setPlayerPosition(player, position);
            }
        }
    }

    /**
     * Checks if a given position is occupied by any players. Loops through
     * the first players in the list, but not all, since positions will be
     * generated for one and one player further down the list.
     *
     * @param position The position to check.
     * @param players The list of relevant players.
     * @param playerToCheck The last position in the list to check.
     * @return True if the position is available, false if not.
     */
    public static boolean positionIsOccupied(int[] position, final List<BBPlayer> players, int playerToCheck) {
        for (int i = 0; i < playerToCheck; i++) {
            AbstractPlayer player = players.get(i);
            int[] playerPosition = player.getPosition();
            int size = player.getSize();
            if (    playerPosition[0] < position[0] &&
                    playerPosition[0] + size >= position[0] &&
                    playerPosition[1] < position[1] &&
                    playerPosition[1] + size >= position[1]) {
                return true;
            }
        }
        return false;
    }

    /**
     * Generates a random position within the specified bounds.
     *
     * @param screenWidth The maximum x-position.
     * @param screenHeight The maximum y-position.
     * @return The random position.
     */
    public static int[] generateRandomPosition(int screenWidth, int screenHeight) {
        return new int[]{
            (int)(Math.random() * screenWidth - 10),
            (int)(Math.random() * screenHeight - 10)
        };
    }

    /**
     * Checks if one of the game board's four corners is available. If so, the available
     * corner is returned. If not, another random position is returned.
     *
     * @param player The player that needs to find a corner.
     * @param players The other connected players.
     * @param screenWidth The width of the game board.
     * @param screenHeight The height of the game board.
     * @return An available position.
     */
    private static int[] findAvailableCorner(BBPlayer player, List<BBPlayer> players, int screenWidth, int screenHeight) {
        boolean[] testedCorners = new boolean[4];
        int nofTestedCorners = 0;
        while (nofTestedCorners < 4) {
            // Generate a random corner position
            int randomCorner = (int)(Math.random() * 4);
            // If the corner has not been tested..
            if (!testedCorners[randomCorner]) {
                // Get the position of the corner
                int[] position = new int[]{
                        (randomCorner % 2) * (screenWidth - player.getSize()),
                        (randomCorner / 2) * (screenHeight - player.getSize())
                };
                // If the position is available, return the position
                if (!positionIsOccupied(position, players)) {
                    return position;
                }
                // Or else, test a new corner
                else {
                    testedCorners[randomCorner] = true;
                    nofTestedCorners++;
                }
            }
        }
        // If all corners have been tested, and none were available, return another available random position
        return findAvailablePosition(players, screenWidth, screenHeight);
    }

    /**
     * Loops through the list of players to see if the specified position is available.
     *
     * @param position The position to check.
     * @param players The list of players.
     * @return True if the position is avaible, false if not.
     */
    private static boolean positionIsOccupied(int[] position, List<BBPlayer> players) {
        return positionIsOccupied(position, players, players.size());
    }

    /**
     * Finds a position unoccupied by any players.
     *
     * @param players The list of players.
     * @param screenWidth The width of the game board.
     * @param screenHeight The height of the game board.
     * @return An unoccupied position.
     */
    private static int[] findAvailablePosition(List<BBPlayer> players, int screenWidth, int screenHeight) {
        while (true) {
            int[] randomPosition = generateRandomPosition(screenWidth, screenHeight);
            if (!positionIsOccupied(randomPosition, players)) {
                return randomPosition;
            }
        }
    }

    /**
     * Checks whether the thread has completed its pre-game calculations and
     * is ready to start a game.
     *
     * @return True if the thread is ready, false if not.
     */
    public boolean isReady() {
        return this.ready;
    }

    /* *******************************************************************************/

}
