package server.brickblock.fwimpl;

import server.framework.models.AbstractPlayer;
import server.framework.models.AbstractServer;
import server.framework.network.SocketWrapper;
import server.testmodules.TestModule;
import server.brickblock.objects.Powerup;

import java.util.List;
import java.util.ArrayList;

/**
 * BrickBlock's <code>AbstractPlayer</code> implemention.
 * In addition to the attributes and functionality found
 * in <code>AbstractPlayer</code>, this class contains
 * information about the player's strength. In addition,
 * this class offers
 * functionality for handling a player's bonuses in terms
 * of active <code>Powerup</code> objects.
 *
 * @author Martin Jarrett
 * @author Eivind Sorteberg
 * @version 1.0
 */
public class BBPlayer extends AbstractPlayer {
    // The player's active power ups
    private List<Powerup> powerups;

    // The player's strength
    private int strength;

    /**
     * Creates a <code>BBPlayer</code> object with the specified
     * attributes. Calls the parent constructor and initialises
     * the power up list.
     *
     * @param socket The player's <code>SocketWrapper</code> object.
     * @param server The server model.
     * @param testtype The test type, if any.
     */
    public BBPlayer(SocketWrapper socket, AbstractServer server, TestModule.TestType testtype) {
        super(socket, server, testtype);
        this.powerups = new ArrayList<Powerup>();
    }

    /**
     * Returns the player's strength.
     *
     * @return The player's strength.
     */
    public int getStrength() {
        return strength;
    }

    /**
     * Sets the player's strength.
     *
     * @param strength The player's strength.
     */
    public void setStrength(int strength) {
        this.strength = strength;
    }

    /**
     * Lets the player pick up a power up object, and sets
     * the time for when this power up will be deactivated.
     *
     * @param powerup The picked up power up.
     */
    public void pickUpPowerup(Powerup powerup) {
        powerup.setTimeout(System.currentTimeMillis() + powerup.getDuration() * 1000);
        powerup.activate(this);
        this.powerups.add(powerup);
    }

    /**
     * Loops through the player's power up objects, and checks if
     * any of the power ups have timed out. If so, those power ups
     * are deactivated.
     *
     * @return The power ups that are deactivated.
     */
    public List<Powerup> getInactivePowerups() {
        //Gets the current time
        long currentTime = System.currentTimeMillis();
        List<Powerup> expired = new ArrayList<Powerup>();

        //For all the power ups..
        for (Powerup powerup : powerups) {
            //Get the timeout time
            long timeout = powerup.getTimeout();

            //If the time has passed..
            if (currentTime > timeout) {
                //Deactivate the power up
                powerup.deactivate(this);
                expired.add(powerup);
            }
        }
        this.powerups.removeAll(expired);

        //Return the list of timed out power ups
        return expired;
    }
}
