package server.brickblock.fwimpl;

import server.framework.models.AbstractServer;
import server.framework.Utils;
import server.framework.network.parser.ActionPair;
import server.framework.network.SocketWrapper;
import server.framework.threads.Event;
import server.testmodules.TestModule;

/**
 * The BrickBlock implementation of the <code>AbstractServer</code> model.
 * Handles special actions that are not handled by the parent class.
 *
 * @version 1.0
 * @author Martin Jarrett
 * @author Eivind Sorteberg
 */
public class BBServer extends AbstractServer<BBSession, BBPlayer> {
    /**
     * Creates a <code>BBServer</code> object by calling the parent class'
     * constructor. Does not add any extra content to the object.
     *
     * @param port The server port.
     * @param commType The transport protocol in use.
     * @param testType The test type, if any.
     */
    public BBServer(int port, CommType commType, TestModule.TestType testType) {
        super(port, commType, testType);
    }

    /**
     * Called when special actions are received that could not be handled by the
     * parent class. The effects of these actions depends on the type of action.
     *
     * @param actionPair Wrapper containing the action and values.
     * @param sender The sender of the message.
     * @param session The session to which the action applies.
     * @return Whether the message should be forwarded to the connected clients.
     */
    public boolean notifyAboutSpecialMessageReceived(ActionPair actionPair, BBPlayer sender, BBSession session) {
        switch (actionPair.getAction()) {
            case powerup_removed:
                if (sender != null && session != null) session.notifyAboutEventOccured(Event.powerup_activated, sender, actionPair.getValues()[0]);
                return false;
            case force:
                if (session != null) session.notifyAboutEventOccured(Event.push, sender, actionPair.getValues()[0]);
            case powerup_added:
            case powerup_inactive:
            case trap_added:
                // All these actions should be forwarded
                return true;
            default:
                Utils.printTimeStampedInfo("Invalid message: " + actionPair, true);
                return false;
        }
    }

    /**
     * Creates a <code>BBPlayer</code> instance.
     *
     * @param wrapper The player's <code>SocketWrapper</code> object.
     * @param testType The test type, if any.
     * @return A <code>BBPlayer</code> instance.
     */
    protected BBPlayer createPlayerInstance(SocketWrapper wrapper, TestModule.TestType testType) {
        return new BBPlayer(wrapper, this, testType);
    }

    /**
     * Creates a <code>BBSession</code> instance.
     *
     * @param testtype The test type, if any.
     * @return A <code>BBSession</code> instance.
     */
    protected BBSession createSessionInstance(TestModule.TestType testtype) {
        return new BBSession(this, testtype);
    }
}
