package server.brickblock.objects;

import server.brickblock.fwimpl.BBPlayer;

/**
 * Abstract class used for power up objects. All
 * power up objects have an increment value that determines
 * the effect of the power up,
 * and can be activated and deactivated.
 *
 * @author Martin Jarrett
 * @author Eivind Sorteberg
 * @version 2.0
 */
public abstract class Powerup {
    public static final int DEFAULT_SIZE = 10;
    public static final int DEFAULT_DURATION_SECS = 10;

    private int xPos;
    private int yPos;
    private int duration;
    private long timeout;

    /**
     * Constructor that assigns a position to the power up
     * object, as well as defining the duration.
     *
     * @param xPos     The object's x-position on the screen.
     * @param yPos     The object's y-position on the screen.
     * @param duration How long the power up remains active.
     */
    protected Powerup(int xPos, int yPos, int duration) {
        this.setPosition(xPos, yPos);
        this.duration = duration;
    }

    /**
     * Sets the position of the power up.
     *
     * @param xPos The object's x-position.
     * @param yPos The object's y-position.
     */
    public void setPosition(int xPos, int yPos) {
        this.xPos = xPos;
        this.yPos = yPos;
    }

    /**
     * @return The object's position.
     */
    public int[] getPosition() {
        return new int[]{xPos, yPos};
    }

    /**
     * Returns the duration of the power up. This is
     * how long the power up remains active after it
     * has been picked up, and is not affected by how
     * long it has been on the table before it is picked
     * up.
     *
     * @return The power up's duration.
     */
    public int getDuration() {
        return duration;
    }

    /**
     * Sets the duration of the power up to the specified
     * value. This is how long the power up remains active
     * after it has been picked up, and is not affected
     * by how long it has been on the table before it is
     * picked up.
     *
     * @param duration The power up's duration.
     */
    public void setDuration(int duration) {
        this.duration = duration;
    }

    /**
     * "Turns on" the power up by increasing one of the player's
     * attributes. How this power up affects the player is
     * dependent on how this method is implemented.
     *
     * @param player The player that is affected by the power up.
     */
    public abstract void activate(BBPlayer player);

    /**
     * "Turns off" the power up by restoring the player's
     * attribute to the value it would have had if the power
     * up had not been picked up.
     *
     * @param player The player that is affected by the power up.
     */
    public abstract void deactivate(BBPlayer player);

    /**
     * Returns the player attribute that is affected by the power up
     * object, i.e. speed, size, or strength.
     *
     * @param player A <code>BBPlayer</code> object.
     * @return The affected attribute.
     */
    public abstract Object getAffectedAttribute(BBPlayer player);

    /**
     * Returns the power up's attribues, i.e. the power up's
     * type, position, and increment value.
     *
     * @return The power up's attributes.
     */
    public String[] getAttributes() {
        return new String[]{
                this.getName(),
                Integer.toString(this.getPosition()[0]),
                Integer.toString(this.getPosition()[1]),
                Integer.toString(this.getIncrement())
        };
    }

    /**
     * Returns a three letter String representation of the power up object.
     *
     * @return The name of the power up object
     */
    public abstract String getName();

    /**
     * Sets the time for when this power up will no longer remain
     * active for the player that has picked it up.
     *
     * @param timeout When the power up is deactivated.
     */
    public void setTimeout(long timeout) {
        this.timeout = timeout;
    }

    /**
     * @return When the power up is deactivated.
     */
    public long getTimeout() {
        return this.timeout;
    }

    /**
     * Returns the power up's increment value, i.e. how much the
     * player's affected attribute is incremented when picking up the power up.
     *
     * @return The power up's increment value.
     */
    public abstract int getIncrement();
}
