package server.framework.gui;

import server.framework.models.AbstractPlayer;
import server.framework.models.AbstractSession;
import server.framework.models.Setting;
import server.testmodules.TestModule;

import javax.swing.*;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.*;
import java.util.List;
import java.util.Collections;

/**
 * The <code>PlayerPanel</code> displays information about
 * the sessions players. The player's color, score, and
 * possible team is displayed. The player information is
 * updated when a player connects, disconnects, or dies,
 * or when the settings are changed. The start-test button
 * is added to this panel.
 *
 * @author Martin Jarrett
 * @author Eivind Sorteberg
 * @version 1.0
 */
public class PlayerPanel<P extends AbstractPlayer> extends JPanel implements ActionListener {

    // The predefined colors array representing the two teams.
    private Color[] teamColors = {Color.RED, Color.BLUE};

    // The session to which the players are connected.
    private AbstractSession<P, ?> session;

    // Button used to start a test when tests are performed
    private JButton button;

    /**
     * Creates a <code>PlayerPanel</code> with the specified attribute. Sets the
     * session and initializes the <code>JPanel</code>.
     *
     * @param session The session the players are connected to.
     */
    public PlayerPanel(AbstractSession<P, ?> session) {
        this.session = session;
        this.initialize();
    }

    /**
     * When a new player connects to the session, a player disconnects,
     * a change in score occurs, or teams are enabled/disabled (i.e.
     * whenever the playerlist is updated), the <code>PlayerPanel</code>
     * needs to update to display the correct information. This is done
     * by calling the repaint-method of the parent class.
     */
    public void firePlayerListUpdated() {
        this.repaint();
    }

    /**
     * If the button is pressed, the test must start and the button is
     * disabled since it has fullfilled its responsibility.
     *
     * @param e The button is pressed.
     */
    public void actionPerformed(ActionEvent e){
        this.session.runTest();
        this.button.setEnabled(false);
    }

    /**
     * This method uses the parent class' method to paint the panel. Players
     * are sorted and listed after their score. If teams are enabled, the players
     * are divided in their respective teams. The cummulative player scores for each
     * team is shown as a team score. 
     *
     * @param g The Graphics object used to paint the panel.
     */
    protected void paintComponent(Graphics g) {
        super.paintComponent(g);
        Graphics2D g2d = (Graphics2D)g;
        int posX = 15;
        int posY = 10;
        int lineHeight = 15;
        int[] teamScore = new int[teamColors.length];
        boolean teamsEnabled = (Boolean)session.getSettings().getSetting(Setting.teams);

        List<P> players = session.getPlayers();
        Collections.sort(players);
        int nofPlayers = players.size();
        for (int i = nofPlayers - 1; i >= 0; i--) {
            AbstractPlayer player = players.get(i);
            int[] color = player.getColor();
            int score = player.getScore();
            if (teamsEnabled) {
                int team = player.getTeam();
                teamScore[team] += player.getScore();
                g2d.setColor(teamColors[team]);
                g2d.fillRect(posX, posY - lineHeight + 5, lineHeight - 4, lineHeight - 4);
                g2d.setColor(Color.BLACK);
                g2d.drawRect(posX, posY - lineHeight + 5, lineHeight - 4, lineHeight - 4);
                //noinspection SuspiciousNameCombination
                posX += lineHeight;
            }
            g2d.setColor(new Color(color[0], color[1], color[2]));
            g2d.drawString(player.toString(), posX, posY);
            posX = 180;
            g2d.drawString("" + score, posX, posY);
            posY += lineHeight;
            posX = 15;
        }

        if (teamsEnabled) {
            posY += lineHeight;
            for (int team_no = 0; team_no < teamColors.length; team_no++) {
                g2d.setColor(teamColors[team_no]);
                g2d.fillRect(0, posY - lineHeight + 3, 200, lineHeight);
                g2d.setColor(Color.WHITE);
                g2d.drawString("Team #" + team_no, posX, posY);
                posX = 180;
                g2d.drawString("" + teamScore[team_no], posX, posY);
                posY += lineHeight;
                posX = 15;
            }
        }
    }

    /**
     * Used to set the size of the <code>PlayerPanel</code>. It also retrieve the
     * testtype and if there is a test to be run, a button is added to the panel
     * with its command and listener.
     */
    private void initialize() {
        this.setPreferredSize(new Dimension(200, 200));
        TestModule.TestType testtype = session.getTestType();
        if (testtype != null) {
            this.button = new JButton("Start " + testtype.name().substring(0, 1).toUpperCase() + testtype.name().substring(1).toLowerCase() + "Test");
            this.button.setActionCommand(testtype.name().toLowerCase() + "start");
            this.button.addActionListener(this);
            this.add(button);
        }
    }
}
