package server.framework.gui;

import server.framework.models.AbstractPlayer;
import server.framework.models.AbstractSession;

import javax.swing.*;
import java.util.HashMap;
import java.util.ArrayList;

/**
 * The server's gui which displays information about the server's sessions
 * and these sessions' players and settings. If tests are performed, they
 * are started when a button in this gui is pressed.
 *
 * @author Martin Jarrett
 * @author Eivind Sorteberg
 * @version 1.0
 */
public class ServerGUI<S extends AbstractSession<P, ?>, P extends AbstractPlayer> extends JFrame {

    //The numbers of panels to occur under each other.
    public static final int NOF_PANELS_VERTICAL = 4;

    // HashMap of the current sessions on the server.
    private HashMap<Integer, S> sessions;

    // The placements of the sessions' panels in the ServerGUI.
    private HashMap<SessionPanel<P>, Box> placements;

    // The boxes to which the panels can be added.
    private ArrayList<Box> boxes;

    // The background Box.
    private Box bgBox;

    /**
     * Creates a <code>ServerGUI</code>-frame and initialize the various
     * attributes.
     */
    public ServerGUI() {
        super("Server GUI");
        this.sessions = new HashMap<Integer, S>();
        this.placements = new HashMap<SessionPanel<P>, Box>();
        this.boxes = new ArrayList<Box>();
        this.bgBox = Box.createHorizontalBox();
        this.add(bgBox);
        this.setResizable(false);
        this.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
        this.setVisible(true);
    }

    /**
     * Adds the session's <code>SessionPanel</code> to the <code>ServerGUI</code>
     * after the panel is created. The session is added to the <code>ServerGUI</code>'s
     * session-ArrayList with its id as key. The <code>SessionPanel</code>
     * placement on the frame is determined based on the number of
     * <code>SessionPanels</code> already in the ServerGUI.
     *
     * @param session The session which is to be added.
     */
    public void addSession(S session) {
        if (sessions.size() % NOF_PANELS_VERTICAL == 0) {
            Box newBox = Box.createVerticalBox();
            this.boxes.add(newBox);
            this.bgBox.add(newBox);
        }

        SessionPanel<P> panel = new SessionPanel<P>(session);
        session.setSessionPanel(panel);
        this.sessions.put(session.getId(), session);

        Box box = this.boxes.get(sessions.size() / NOF_PANELS_VERTICAL);
        box.add(panel);
        this.placements.put(panel, box);

        this.pack();
        this.repaint();
    }

    /**
     * Removes the session from the <code>ServerGUI</code>. The id is used to identify
     * the session in question and the session's SessionPanel is removed
     * from the gui.
     * 
     * @param sessionId The id of the session to be removed.
     */
    public void removeSession(int sessionId) {
        S session = this.sessions.remove(sessionId);
        this.placements.get(session.getSessionPanel()).remove(session.getSessionPanel());
        this.placements.remove(session.getSessionPanel());
        this.pack();
        this.repaint();
    }
}
