package server.framework.gui;

import server.framework.models.AbstractPlayer;
import server.framework.models.AbstractSession;

import javax.swing.*;
import javax.swing.border.BevelBorder;
import java.awt.*;

/**
 * The <code>SessionPanel</code> displays information about the session
 * by using two other panels containing important information. This class
 * determines the layout of the gui and when the information should be updated. 
 *
 * @author Martin Jarrett
 * @author Eivind Sorteberg
 * @version 1.0
 */
public class SessionPanel<P extends AbstractPlayer> extends JPanel {

    // The session's PlayerPanel.
    private PlayerPanel playerPanel;

    // The session's SettingsPanel.
    private SettingsPanel settingsPanel;

    /**
     * Creates a <code>SessionPanel</code> object with the specified attributes.
     * Also, the constructors for the <code>PlayerPanel</code> and
     * <code>SettingsPanel</code> objects are called. This are added to the
     * <code>SessionPanel</code> in a specific layout.
     *
     * @param session The session which is to be displayed.
     */
    public SessionPanel(AbstractSession<P, ?> session) {
        this.playerPanel = new PlayerPanel<P>(session);
        this.settingsPanel = new SettingsPanel(session);

        JPanel bottomPanel = new JPanel(new BorderLayout());
        bottomPanel.add(createTitledPanel("Players", 3, playerPanel), BorderLayout.WEST);
        bottomPanel.add(createTitledPanel("Settings", 3, settingsPanel), BorderLayout.EAST);

        this.setBorder(BorderFactory.createEmptyBorder(0, 4, 0, 4));
        this.setLayout(new BorderLayout());
        this.add(new JLabel("<html><h2>" + session.getName() + "</h2></html>"), BorderLayout.NORTH);
        this.add(bottomPanel, BorderLayout.CENTER);
    }

    /**
     * Used to create titles/headers for the <code>SessionPanel</code>'s content panels.
     *
     * @param header The header for the JPanel created.
     * @param headerSize The size of the header.
     * @param content The JPanel to be added to the created panel.
     * @return The created panel.
     */
    private JPanel createTitledPanel(String header, int headerSize, JPanel content) {
        JPanel panel = new JPanel(new BorderLayout());
        panel.setBorder(BorderFactory.createCompoundBorder(
                BorderFactory.createBevelBorder(BevelBorder.LOWERED), // Outside
                BorderFactory.createEmptyBorder(0, 4, 0, 4))          // Inside
        );
        panel.add(new JLabel("<html><h" + headerSize  + ">" + header + "</h" + headerSize + "></html>"), BorderLayout.NORTH);
        panel.add(content, BorderLayout.CENTER);
        return panel;
    }

    /**
     * The settings list has changed and the <code>SettingsPanel</code>
     * is notified so that the updated information is displayed. The
     * <code>PlayerPanel</code> is also notified since the teams setting
     * may have been change which would lead to the player list to change.
     */
    public void fireSettingsListChanged() {
        this.settingsPanel.fireSettingsListChanged();
        this.playerPanel.firePlayerListUpdated();
    }

    /**
     * The player list has changed and the <code>PlayerPanel</code>
     * is notified so that the updated information is displayed.
     */
    public void firePlayerListUpdated() {
        this.playerPanel.firePlayerListUpdated();
    }
}
