package server.framework.models;

import server.framework.network.Communicator;
import server.framework.network.ServerSocketCommunicator;
import server.framework.network.SocketWrapper;
import server.testmodules.TestModule;
import java.net.SocketAddress;

/**
 * An <code>AbstractPlayer</code> object is a representation of
 * a participant in a framework game.
 *
 * The player has different attributes such as a name, id,
 * color, score, and size.
 *
 * In addition, each <code>AbstractPlayer</code> object connected to the
 * server has a uniquely assigned <code>SocketWrapper</code> object, that
 * is used for communication with the assosiated client.
 *
 * @author Eivind Sorteberg
 * @author Martin Jarrett
 * @version 2.0
 */
public abstract class AbstractPlayer implements Comparable<AbstractPlayer> {
    private static int nextId = 0;

    // ATTRIBUTES
    private int id;
    private String name;
    private int score;
    private int[] color;
    private int team;
    private boolean ready;
    private int[] screenResolution;
    private int[] position;

    private int sessionId;

    // COMMUNICATION
    private SocketWrapper socket;
    private Communicator communicator;
    private SocketAddress address;

    //POWER UP ATTRIBUTES
    private int size;
    private int speed;

    // Alive requests
    private int lastAliveReceived;

    /**
     * Constructor for AbstractPlayer. All implementing
     * subclasses should call this contructor through a
     * <code>super()</code>-call. The constructor creates an
     * <code>AbstractPlayer</code>-instance, and gives default values
     * to all attributes not specified as input parameters.
     *
     * @param server The server to which the player is connected.
     * @param socket The <code>SocketWrapper</code> used for communication.
     * @param testtype The test type, if any.
     */
    public AbstractPlayer(SocketWrapper socket, AbstractServer server, TestModule.TestType testtype) {
        if (testtype == null) {
            this.communicator = new ServerSocketCommunicator(socket, server);
        }
        else {
            System.out.println("Running with testtype: " + testtype);
            this.communicator = testtype.createTestModule(testtype, socket, server);
        }

        this.id = nextId;
        this.name = "Unnamed " + id;
        this.score = 0;
        this.color = new int[3];
        this.team = -1;
        this.ready = false;
        this.screenResolution = new int[2];
        this.position = new int[2];
        this.sessionId = -1;

        this.size = 10;

        // Increment the id for the next player
        nextId++;

        this.socket = socket;
        this.address = socket.getRemoteAddress();
        this.socket.setPlayer(this);
    }

    /* PLAYER ATTRIBUTES */

    /**
     * @return The player's id.
     */
    public int getId() {
        return this.id;
    }

    /**
     * Sets the name of the player.
     *
     * @param name The player's name.
     */
    public void setName(String name) {
        this.name = name;
    }

    /**
     * @return The player's name.
     */
    public String getName() {
        return this.name;
    }

    /**
     * Increases the player's score with the provided value.
     *
     * @param inc The score increase.
     */
    public void increaseScore(int inc) {
        this.score += inc;
    }

    /**
     * @return The player's score.
     */
    public int getScore() {
        return this.score;
    }

    /**
     * Returns a string representation of the player, consisting of the player's name and id.
     *
     * @return A string representation of the player.
     */
    public String toString() {
        return this.getName() + (" (" + this.getId() + ")");
    }

    /**
     * Returns the address of the client that is represented by this
     * <code>AbstractPlayer</code> object.
     *
     * @return The address of the player's associated client.
     */
    public SocketAddress getAddress() {
        return address;
    }

    /**
     * Called when a player's attributes has been received. Parses a String
     * containing attribute values and applies the correct values to the
     * different settings. The attribute String is always created by the
     * <code>getAttributeValues()</code> method, and so the rules from that
     * method are used for parsing.
     *
     * @param values The player's attributes
     * @see AbstractPlayer#getAttributeValues()
     */
    public void setAttributeValues(String[] values) {
        //Get the attribute(s)
        this.setName(values[0]);
        this.color[0] = Integer.parseInt(values[1]);
        this.color[1] = Integer.parseInt(values[2]);
        this.color[2] = Integer.parseInt(values[3]);
        this.increaseScore(Integer.parseInt(values[4]));
//        this.team = Integer.parseInt(values[5]);
        this.setReady(values[6]);
        this.position[0] = Integer.parseInt(values[7]);
        this.position[1] = Integer.parseInt(values[8]);
        this.screenResolution[0] = Integer.parseInt(values[9]);
        this.screenResolution[1] = Integer.parseInt(values[10]);
    }

    /**
     * Returns a String containing the most important player attributes.
     * From this array, a String is created and transmitted to the server
     * when joining connecting to a server, and the player's representation
     * on the other mobile phones is created from this attribute String.
     *
     * @return A String array containing the player's attributes.
     */
    public String[] getAttributeValues() {
        return new String[]{
                this.name,
                Integer.toString(this.color[0]),
                Integer.toString(this.color[1]),
                Integer.toString(this.color[2]),
                Integer.toString(this.score),
                Integer.toString(this.team),
                Boolean.toString(this.ready),
                Integer.toString(this.position[0]),
                Integer.toString(this.position[1]),
                Integer.toString(this.screenResolution[0]),
                Integer.toString(this.screenResolution[1])
        };
    }

    /* COMMUNICATION SPECIFIC METHODS */


    /**
     * Sends a message using the player's assigned <code>SocketWrapper</code>.
     *
     * @param message The message to send.
     * @param confirm Whether the receiver should confirm that the message has been received.
     */
    public void sendMessage(Object message, boolean confirm) {
        this.communicator.sendMessage(message, confirm);
    }

    /**
     * @return The player's associated <code>SocketWrapper</code> object.
     */
    public SocketWrapper getSocket() {
        return this.socket;
    }

    /**
     * Called whenever the player has disconnected from the server. Makes
     * sure the connection is closed, and frees all resources occupied by the
     * player object.
     */
    public void notifyAboutPlayerDisconnected() {
        this.communicator.disconnect(null);
        this.communicator = null;
        this.socket = null;
    }

    /**
     * Sets the player's ready status to the given value.
     *
     * @param ready Whether the player is ready or not.
     */
    public void setReady(boolean ready) {
        this.ready = ready;
    }

    /**
     * Sets the player's ready status to the given value,
     * converted to a boolean value.
     *
     * @param ready Whether the player is ready or not.
     */
    public void setReady(String ready) {
        this.ready = "true".equals(ready);
    }

    /**
     * Sets the player's size to the given value.
     *
     * @param size The player's size.
     */
    public void setSize(int size) {
        this.size = size;
    }

    /**
     * @return The player's size.
     */
    public int getSize() {
        return this.size;
    }

    /**
     * Sets the player's speed to the given value.
     *
     * @param speed The player's speed.
     */
    public void setSpeed(int speed) {
        this.speed = speed;
    }

    /**
     * @return The player's speed.
     */
    public int getSpeed() {
        return this.speed;
    }

    /**
     * @return The player's associated client's screen resolution.
     */
    public int[] getScreenResolution() {
        return this.screenResolution;
    }

    /**
     * @return The player's color.
     */
    public int[] getColor() {
        return this.color;
    }

    /**
     * Runs the test specified by the player's test type
     * parameter.
     */
    public void runTest() {
        if(communicator instanceof TestModule){
            ((TestModule)communicator).runTest();
        }
    }

    /**
     * @return The id of the session to which the player is connected.
     */
    public int getSessionId() {
        return this.sessionId;
    }

    /**
     * Sets the id of the session to which the player is connected.
     * @param sessionId The id of the session to which the player is connected.
     */
    public void setSessionId(int sessionId) {
        this.sessionId = sessionId;
    }

    /**
     * Sets the player's position.
     *
     * @param position The player's position represented as a String array.
     */
    public void setPosition(String[] position) {
        this.position[0] = Integer.parseInt(position[0]);
        this.position[1] = Integer.parseInt(position[1]);
    }

    /**
     * Sets the player's position.
     *
     * @param position The player's position.
     */
    public void setPosition(int[] position) {
        this.position = position;
    }

    /**
     * Returns an array containing the player's id and position.
     * This method is used by the <code>GameThread</code> to create
     * the bundled position messages.
     *
     * @return An array containing the player's id and position.
     */
    public Object[] getPositionArray() {
        return new Object[]{
                this.getId(),
                this.position[0],
                this.position[1]
        };
    }

    /**
     * @return The player's position.
     */
    public int[] getPosition() {
        return this.position;
    }

    /**
     * Sets the player's score to the specified value.
     *
     * @param score The player's score.
     */
    public void setScore(int score) {
        this.score = score;
    }

    /**
     * Sets the player's team to the specified value.
     *
     * @param team The player's team.
     */
    public void setTeam(int team) {
        this.team = team;
    }

    /**
     * @return The player's team.
     */
    public int getTeam() {
        return this.team;
    }

    /**
     * Compares two player objects. If this player's score is higher
     * than the other's a positive value is returned. If this player's score
     * is lower than the other player's, a negative value is returned. Finally,
     * if the players have equal score, 0 is returned.
     *
     * @param p2 The player with which to compare the score.
     * @return The difference in player scores.
     */
    public int compareTo(AbstractPlayer p2) {
        return this.getScore() - p2.getScore();
    }

    /**
     * Sets the value of the last received alive confirmation
     * when this is received. Only higher requests ids are stored.
     *
     * @param aliveConfirmationId The received alive confirmation.
     */
    public void notifyAboutAlive(int aliveConfirmationId) {
        if (aliveConfirmationId > this.lastAliveReceived) {
            this.lastAliveReceived = aliveConfirmationId;
        }
    }

    /**
     * @return The last received alive confirmation.
     */
    public int lastAliveConfirmation() {
        return this.lastAliveReceived;
    }
}
