package server.framework.models;

import server.framework.gui.ServerGUI;
import server.framework.network.parser.Action;
import server.framework.network.parser.ActionPair;
import server.framework.network.parser.MessageParser;
import server.framework.network.parser.StringParser;
import server.framework.network.SocketWrapper;
import server.framework.Utils;
import server.framework.threads.Event;
import server.testmodules.TestModule;

import java.io.IOException;
import java.net.*;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * This class is the top-level model for the server application. This model contains a list
 * of all the players connected to the server, as well as a list of the currently running sessions.
 *
 * In addition, the server model is responsible for receiving notifications from the communication
 * module when messages are received.
 */
public abstract class AbstractServer<S extends AbstractSession<P, ?>, P extends AbstractPlayer> {
    /**
     * This enum contains the support for tcp and udp communication
     * specified in the arguments when starting the server application.
     */
    public enum CommType {
        tcp,udp;

        public static CommType getCommType(String commType) {
            for (CommType type : CommType.values()) {
                if (type.name().equals(commType.toLowerCase())) {
                    return type;
                }
            }
            return null;
        }

    }

    // The communication type (TCP or UDP)
    private CommType commType;

    // Used for TCP communication
    private ServerSocket serverSocket;

    // Used for UDP communication
    private DatagramSocket datagramSocket;

    // Player list
    private List<P> players;

    // Session list
    private Map<Integer, S> sessions;

    // The message parser used to create messages
    private MessageParser parser;

    // The test type, if any
    private TestModule.TestType testType;

    // The server GUI
    private ServerGUI<S, P> gui;

    /**
     * Creates an <code>AbstractServer</code> object with the specified
     * communication protocol and port number. The communication
     * channel is opened, and the server GUI is created.
     *
     * In addition, the player and session lists are initialised.
     *
     * @param port The port number on which the server runs.
     * @param commType The transport protocol to use.
     * @param testType The test type, if any.
     */
    public AbstractServer(int port, CommType commType, TestModule.TestType testType) {
        this.testType = testType;
        this.parser = new StringParser();
        this.sessions = new HashMap<Integer, S>();
        this.players = new ArrayList<P>();
        this.gui = new ServerGUI<S, P>();
        this.commType = commType;
        try {
            Utils.printTimeStampedInfo("Running " + commType + " server on " + InetAddress.getLocalHost());
        }
        catch (UnknownHostException e) {
            e.printStackTrace();
        }

        try {
            switch(commType) {
                case tcp: // Open TCP socket
                    this.serverSocket = new ServerSocket(port);
                    this.listenOnTcp();
                    break;
                case udp: // Open UDP socket
                    this.datagramSocket = new DatagramSocket(port);
                    this.receiveInitialPacket();
            }
        }
        catch (IOException e) {
            e.printStackTrace();
        }
    }

    /**
     * Creates a new session, and adds the session to the server GUI.
     *
     * @param testtype The test type, if any.
     * @return The new session instance.
     */
    private S createSession(TestModule.TestType testtype) {
        S newSession = createSessionInstance(testtype);
        this.sessions.put(newSession.getId(), newSession);
        this.gui.addSession(newSession);
        return newSession;
    }

    /**
     * Creates a new session instance of the correct type.
     *
     * @param testtype The test type, if any.
     * @return The new session instance.
     */
    protected abstract S createSessionInstance(TestModule.TestType testtype);

    /**
     * @return The currently used message parser implementation.
     */
    public MessageParser getParser() {
        return this.parser;
    }

    /**
     * @return The server GUI.
     */
    public ServerGUI<S, P> getGUI() {
        return gui;
    }

    /**
     * @return The transport protocol (TCP or UDP).
     */
    public CommType getCommType() {
        return this.commType;
    }

    /**
     * Listens for connections on the TCP socket. Is only used when TCP is
     * used as transport protocol. Each time a connection is detected, a
     * player object is created, and the server returns to waiting for more
     * connections.
     */
    private void listenOnTcp() {
        new Thread() {
            boolean running = true;
            public void run() {
                try {
                    while (running) {
                        SocketWrapper wrapper = new SocketWrapper(serverSocket.accept());
                        createPlayer(wrapper);
//                        System.out.println("Connection received");
                    }
                }
                catch (IOException e) {
                    e.printStackTrace();
                }
            }
        }.start();
    }

    /**
     * Listens for the first connection on the UDP socket. When this is received,
     * a player object is created, and the server returns to listening for more
     * connections.
     *
     * @throws IOException Should not happen.
     */
    private void receiveInitialPacket() throws IOException {
        byte[] rec = new byte[64];
        DatagramPacket initialpacket = new DatagramPacket(rec, rec.length);
        this.datagramSocket.receive(initialpacket);
        StringBuilder message = new StringBuilder();
        for (byte b : rec) {
            char c = (char)b;
            if (c == '|') break;
            message.append(c);
        }
        SocketWrapper socket = new SocketWrapper(datagramSocket);
        socket.setRemoteAddress(initialpacket.getSocketAddress());
        this.notifyAboutMessageReceived(message.toString(), socket);
    }

    /**
     * Method that is called whenever one of the communication objects receives a
     * message from a client. The <code>MessageParser</code> instance is used to
     * parse the message, and actions are taken dependent of the contents of the
     * message.
     *
     * @param word The message that was received.
     * @param socket The <code>SocketWrapper</code> object that received the message.
     */
    @SuppressWarnings(value = {"unchecked"})
    public void notifyAboutMessageReceived(Object word, SocketWrapper socket) {
        boolean newPlayer = false;  // Used for info print
        P sender = findPlayer(socket);
        if (sender == null) {
            sender = createPlayer(socket);
            newPlayer = true;
        }

        boolean confirm = false;
        S session = sessions.get(sender.getSessionId());
        ActionPair actionPair = parser.parseMessage(word);
        switch (actionPair.getAction()) {
            case player:        // Used when connecting to the server, and when refreshing the session list
                sender.setAttributeValues(actionPair.getValues()[0]);
                if (newPlayer) Utils.printTimeStampedInfo(sender.getName() + " connected to server");
                Object sessions = parser.createMessage(Action.session_list, getSessionAttributes(), null);
                sender.sendMessage(sessions, true);
                return;
            case session_selected:
                session = findSession(Integer.parseInt(actionPair.getValues()[0][0]));
                if(session != null){
                    session.addPlayer(sender);
                }
                return;
            case ready:
                sender.setReady("true".equals(actionPair.getValues()[0][0]));
                confirm = true;
                break;
            case disconnect:
                handleDisconnect(sender);
                return;
            case start:
                boolean gameStarted = session.isGameStarted();
                if (session != null) {
                    if (gameStarted) {
                        session.notifyAboutEventOccured(Event.join, sender, null);
                    }
                    else {
                        session.startGame();
                    }
                }
                return;
            case position:
                // A position update may be received after a player has been disconnected (especially with UDP)
                // If the disconnected player was the last in the session, the session may also be null when the message is received
                if (session != null){
                    session.firePositionReceived(sender, actionPair.getValues()[0]);
                }
                return;
            case score:
                if (session != null) session.notifyAboutEventOccured(Event.score, sender, null);
                return;
            case setting_changed:
                if (session != null) session.fireSettingChanged(actionPair.getValues());
                break;
            case alive_request:
                sender.notifyAboutAlive(Integer.parseInt(actionPair.getValues()[0][0]));
                return;
                // Occurs when running server and client on local host; should not have any effect
            case empty_message:
                return;
            default:
                // Check if the server implementation contains these actions
                boolean forward = notifyAboutSpecialMessageReceived(actionPair, sender, session);
                if (!forward) return;
        }

        // Forward the action to all other participants if so is desired
        if (session != null) {
            // To handle disconnection before the player has connected to a session
            session.sendAction(actionPair.getAction(), actionPair.getValues(), sender, true, confirm);
        }
    }

    /**
     * Method that must be implemented in classes extendint this class. This method handles
     * actions that are not specified among the default framework actions.
     *
     * @param actionPair The contents of the received message.
     * @param sender The sender of the message.
     * @param session The session to which the message applies.
     * @return True if the message should be forwarded to all clients.
     */
    public abstract boolean notifyAboutSpecialMessageReceived(ActionPair actionPair, P sender, S session);

    private S findSession(int id) {
        if (id >= 0) {
            return this.sessions.get(id);
        }
        else {
            return createSession(this.testType);
        }
    }

    /**
     * Returns the player owning the specified <code>SocketWrapper</code> object.
     * Used to determine the sender of a message.
     *
     * @param socket The socket where the message was received.
     * @return The player "owning" the socket.
     */
    public P findPlayer(SocketWrapper socket) {
        for (P player : this.players) {
            if (player.getSocket().equals(socket)) {
                return player;
            }
        }
        return null;
    }

    /**
     * Method that is called whenever a disconnection notification is received.
     * Removes the player from the player lists and notifies the relevant session
     * of the disconnection.
     *
     * @param sender The player that disconnected.
     */
    public void handleDisconnect(P sender) {
        this.players.remove(sender);
        Utils.printTimeStampedInfo(sender.getName() + " disconnected");
        S session = this.sessions.get(sender.getSessionId());
        if (session != null) {
            session.handleDisconnect(sender);
        }
    }

    /**
     * Creates a player object owning the specified <code>SocketWrapper</code> object.
     *
     * @param socket The player's socket.
     * @return A new player instance.
     */
    public P createPlayer(SocketWrapper socket) {
        SocketWrapper wrapper = new SocketWrapper(socket.getSocket());
        wrapper.setRemoteAddress(socket.getRemoteAddress());
        P player = createPlayerInstance(wrapper, this.testType);
        this.players.add(player);
        return player;
    }

    /**
     * Creates a player object owning the specified <code>SocketWrapper</code> object.
     * Is used to create the correct kind of player object for the framework implementation.
     *
     * @param wrapper The player's socket wrapper.
     * @param testType The test type, if any.
     * @return A new player instance.
     */
    protected abstract P createPlayerInstance(SocketWrapper wrapper, TestModule.TestType testType);

    /**
     * Returns the attributes of the currently active sessions. This
     * method is called when a new player connects to the server, and
     * a session list needs to be sent.
     *
     * @return The currently active sessions.
     */
    private Object[][] getSessionAttributes() {
        Object[][] returnValue = new Object[this.sessions.size()][];
        int position = 0;
        for (int key : this.sessions.keySet()) {
            returnValue[position] = this.sessions.get(key).getAttributes();
            position++;
        }
        return returnValue;
    }

    /**
     * Called when the player list has changed. Notifies the relevant session
     * of the change.
     *
     * @param session The session affected by the event.
     */
    public void firePlayerListChanged(AbstractSession session) {
        session.firePlayerListChanged();
    }

    /**
     * @return The test type currently used.
     */
    public TestModule.TestType getTestType(){
        return testType;
    }

    /**
     * Starts a test run of the type specified by the arguments
     * when starting the server.
     */
    public void runTest() {
        for (int key : this.sessions.keySet()) {
            this.sessions.get(key).runTest();
        }
    }

    /**
     * Called when a session has been closed. Removes the session
     * from the session list and the server GUI.
     *
     * @param sessionId The session that is to be removed.
     */
    public void fireSessionClosed(int sessionId) {
        this.sessions.remove(sessionId);
        this.gui.removeSession(sessionId);
    }
}