package server.framework.models;

import server.framework.Utils;
import server.framework.gui.SessionPanel;
import server.framework.network.parser.Action;
import server.framework.threads.Event;
import server.framework.threads.EventHandler;
import server.framework.threads.GameThread;
import server.testmodules.TestModule;

import java.util.ArrayList;
import java.util.List;

/**
 * This class represents a session. It contains a list of the players connected
 * to the session, and keeps track of the settings applying to the session.
 *
 * Also, this model is responsible for
 * controlling the currently active game threads when running a game, and notifying
 * the <code>EventHandler</code> implementation when a message is received that
 * is likely to cause a game event to occur.
 *
 * @version 1.0
 * @author Martin Jarrett
 * @author Eivind Sorteberg
 */
public abstract class AbstractSession<P extends AbstractPlayer, EH extends EventHandler> {
    // The interval between each alive request
    public static final long ALIVE_REQUEST_INTERVAL = 2000;

    private static int sessionCounter = 0;

    private int id;
    private List<P> players;
    protected AbstractServer<?, P> server;
    private int[] screenResolution;
    private SettingsList settingsList;
    private String name;
    protected EH eventHandler;
    private GameThread gameThread;
    private int[] teams;
    private ConnectionChecker connectionChecker;
    private int aliveRequestId = 0;
    private TestModule.TestType testType;
    private SessionPanel<P> sessionPanel;

    /**
     * Creates a session, and initialises the variables.
     *
     * @param server The server to which this session belongs.
     * @param testtype The test type, if any.
     */
    public AbstractSession(AbstractServer<?, P> server, TestModule.TestType testtype) {
        this.server = server;
        this.players = new ArrayList<P>();
        this.id = ++sessionCounter;
        this.name = "Session #" + id;
        this.teams = new int[2];
        this.testType = testtype;

        this.screenResolution = new int[]{Integer.MAX_VALUE, Integer.MAX_VALUE};
        this.settingsList = new SettingsList();
//        this.sessionPanel = new SessionPanel<P>(this, testtype);
        if (server.getTestType() == null) {
            this.connectionChecker = new ConnectionChecker();
            new Thread(this.connectionChecker).start();
        }
    }

    /**
     * Sets this session's <code>SessionPanel</code> in the server GUI.
     *
     * @param sessionPanel This session's <code>SessionPanel</code>
     */
    public void setSessionPanel(SessionPanel<P> sessionPanel) {
        this.sessionPanel = sessionPanel;
    }

    /**
     * Adds a new player to the session. If teams are enabled, the player
     * is told which team he belongs to. Also, the new player receives
     * player information on the other player's connected to the session.
     * Finally, the other player's connected to the session are notified
     * that a new player has entered.
     *
     * @param newPlayer The new player.
     */
    public void addPlayer(P newPlayer) {
        Utils.printTimeStampedInfo(newPlayer.getName() + " joined " + this.getName());
        newPlayer.setSessionId(this.getId());
        if (server.getTestType() == null) {
            newPlayer.notifyAboutAlive(this.aliveRequestId);
            newPlayer.setTeam(this.getNextTeam());
            Object teamMessage = server.getParser().createMessage(Action.team, new Object[][]{{newPlayer.getTeam()}}, null);
            newPlayer.sendMessage(teamMessage, false);
            Object settingsMessage = server.getParser().createMessage(Action.setting_list, this.settingsList.getSettingsList(), null);
            newPlayer.sendMessage(settingsMessage, false);

            for (AbstractPlayer registered : this.players) {
                Object regPlayerInfo = server.getParser().createMessage(Action.player, new Object[][]{registered.getAttributeValues()}, registered);
                newPlayer.sendMessage(regPlayerInfo, true);
                Object newPlayerInfo = server.getParser().createMessage(Action.player, new Object[][]{newPlayer.getAttributeValues()}, newPlayer);
                registered.sendMessage(newPlayerInfo, true);
            }
        }
        this.players.add(newPlayer);
        this.checkResolution();
        this.firePlayerListChanged();
    }

    /**
     * @return The current test type, if any.
     */
    public TestModule.TestType getTestType() {
        return testType;
    }

    /**
     * @return Which team should get the next player, to make the distribution as even as possible.
     */
    private int getNextTeam() {
        if (teams[0] < teams[1]) {
            teams[0]++;
            return 0;
        }
        else if (teams[1] < teams[0]) {
            teams[1]++;
            return 1;
        }
        else {
            int team = (int)(Math.random() * 2);
            teams[team]++;
            return team; 
        }
    }

    /**
     * Called when a player disconnects from the session. Notifies
     * the player himself that he has been removed, and also notifies the
     * other connected clients. If the player was the last player in the
     * session, the session is closed, and the resources are released.
     *
     * @param player The disconnected player.
     */
    public void handleDisconnect(P player) {
        player.notifyAboutPlayerDisconnected();
        this.players.remove(player);
        this.sendAction(Action.disconnect, null, player, true, true);
        if (this.players.size() == 0) {
            this.stopGame("No more players left in the game.");
            this.connectionChecker.stop();
            this.server.fireSessionClosed(this.getId());
        }
        else {
            this.firePlayerListChanged();
        }
    }

    /**
     * Method that is called when the game has been stopped for some
     * reason. All connected players are notified about the event, and
     * the game threads are stopped.
     *
     * @param reason The reason why the game was stopped.
     */
    public void stopGame(String reason) {
        this.sendAction(Action.game_over, new String[][]{new String[]{reason}}, true);
        Utils.printTimeStampedInfo("Stopped game (" + reason + ") [" + this.getName() + "]");
        if (this.eventHandler != null) {
            this.eventHandler.stop();
            this.eventHandler = null;
        }
        if (this.gameThread != null) {
            this.gameThread.stop();
            this.gameThread = null;
        }
    }

    /**
     * Called when one or more setting values have been changed.
     * Updates the settings list and the settings panel.
     *
     * @param values The new setting values.
     */
    public void fireSettingChanged(String[][] values) {
        this.settingsList.fireSettingChanged(values);
        this.sessionPanel.fireSettingsListChanged();
    }

    /**
     * Called when the player list has changed, e.g. because a player
     * has scored. Updates the player panel.
     */
    public void firePlayerListChanged() {
        this.sessionPanel.firePlayerListUpdated();
    }

    /**
     * @return The width of the slimmest client screen connected to the session.
     */
    public int getMinimumWidth() {
        return this.screenResolution[0];
    }

    /**
     * @return The height of the lowest client screen connected to the session.
     */
    public int getMinimumHeight() {
        return this.screenResolution[1];
    }

    /**
     * @return The width and height that shall be used in the game.
     */
    public String[] getScreenResolution() {
        return new String[]{
                Integer.toString(getMinimumWidth()),
                Integer.toString(getMinimumHeight())
        };
    }

    /**
     * Called when the session is closed. Stops the thread responsible for
     * sending alive requests.
     */
    public void exit() {
        this.connectionChecker.stop();
    }

    /**
     * @return The players connected to the session.
     */
    public List<P> getPlayers() {
        return this.players;
    }

    /**
     * @return The settings applying to the session.
     */
    public SettingsList getSettings() {
        return this.settingsList;
    }

    /**
     * Loops through all player's resolution, and finds the maximum
     * resolution that will fit with all connected client phones.
     */
    public void checkResolution() {
        for (AbstractPlayer player : players) {
            int[] resolution = player.getScreenResolution();

            if (this.screenResolution[0] > resolution[0]) {
                this.screenResolution[0] = resolution[0];
            }
            if (this.screenResolution[1] > resolution[1]) {
                this.screenResolution[1] = resolution[1];
            }
        }
    }

    /**
     * Sends the specified action to all connected players.
     *
     * @param action The action to send.
     * @param values The values related to the action.
     * @param confirm Whether the receiver(s) should confirm the receival.
     */
    public void sendAction(Action action, Object[][] values, boolean confirm) {
        this.sendAction(action, values, null, false, confirm);
    }

    /**
     * Sends the specified action to all connected players, except the sender of the action,
     * if he does not need to receive the message.
     *
     * @param action The action to send.
     * @param values The values related to the action.
     * @param sender The sender of the action.
     * @param skipSender Whether the sender should receive the message or not.
     * @param confirm Whether the receiver(s) should confirm the receival.
     */
    public void sendAction(Action action, Object[][] values, AbstractPlayer sender, boolean skipSender, boolean confirm) {
        Object message = server.getParser().createMessage(action, values, sender);
        for (AbstractPlayer player : this.players) {
            if (player == sender && skipSender) {
                // Do not send the message to the player involved
                continue;
            }
            player.sendMessage(message, confirm);
        }
    }

    /**
     * @return The session's id.
     */
    public Integer getId() {
        return this.id;
    }

    /**
     * Returns the session's id, name, current number of players and
     * maximum number of players.
     *
     * @return The session's attributes.
     */
    public Object[] getAttributes() {
        return new Object[]{
                this.getId(),
                this.getName(),
                this.players.size(),
                this.settingsList.getSetting(Setting.nof_players)
        };
    }

    /**
     * @return The session's name.
     */
    public String getName() {
        return this.name;
    }

    /**
     * Starts a game. Starts the game thread and event thread. When the event
     * thread has performed its pre-game calculations, a start command is
     * transmitted to all connected clients.
     */
    public void startGame() {
        System.out.println(Utils.getCurrentTime() + ": Started game [" + this.getName() + "]");
        this.eventHandler = createEventHandler();
        this.gameThread = new GameThread(this);
        new Thread(eventHandler).start();
        while (!eventHandler.isReady()) {
            // Wait for the initiation of the eventhandler
            try {
                Thread.sleep(50);
            }
            catch (InterruptedException e) {
                e.printStackTrace();
            }
        }
        new Thread(gameThread).start();
        for (AbstractPlayer player : players) {
            Object message = server.getParser().createMessage(
                    Action.start,
                    new Object[][]{new Object[]{
                            screenResolution[0],
                            screenResolution[1],
                            player.getPosition()[0],
                            player.getPosition()[1]
                    }},
                    null);
            player.sendMessage(message, true);
        }
    }

    /**
     * Creates an <code>EventHandler</code> instance of the correct type.
     *
     * @return An <code>EventHandler</code> instance.
     */
    protected abstract EH createEventHandler();

    /**
     * @return The server to which this session belongs.
     */
    public AbstractServer getServer() {
        return server;
    }

    /**
     * Called when a position is received from one of the session's players.
     * Sends this position to the game thread, who will then forward the
     * position next time it sends a position bundle.
     *
     * @param sender The sender of the position.
     * @param position The position that was received.
     */
    public void firePositionReceived(AbstractPlayer sender, String[] position) {
        if (this.gameThread != null) {
            this.gameThread.addPosition(sender, position);
            sender.setPosition(position);
        }
    }

    /**
     * @return True if a game is in progress, false if not.
     */
    public boolean isGameStarted() {
        return !(this.eventHandler == null && this.gameThread == null);
    }

    /**
     * Sets the position of the specified player.
     *
     * @param player The player whose position to update.
     * @param position The position of the player.
     */
    public void setPlayerPosition(P player, int[] position) {
        player.setPosition(position);
        this.gameThread.addPosition(player, new Integer[]{position[0], position[1]});
    }

    /**
     * @return The session's session panel.
     */
    public SessionPanel getSessionPanel() {
        return this.sessionPanel;
    }

    /**
     * Runs the test when a test has been specified.
     */
    public void runTest() {
        for (AbstractPlayer player : players) {
            player.runTest();
        }
    }

    /**
     * Called when an event has occurred that the <code>EventHandler</code> thread should
     * handle. If the session needs to perform actions because of the event, these actions
     * are specified in the implementation of this method.
     *
     * @param event The event that has occured.
     * @param sender The player that caused the event.
     * @param values The values related to the event.
     */
    public abstract void notifyAboutEventOccured(Event event, P sender, Object[] values);

    /**
     * The thread that is responsible for sending alive requests with
     * regular intervals. The time between each request is specified by
     * the <code>ALIVE_REQUEST_INTERVAL</code> variable.
     */
    private class ConnectionChecker implements Runnable {
        // The number of allowed packetlosses before the client is disconnected.
        public static final int ALLOWED_PACKETLOSS = 10;

        private boolean running;

        /**
         * Runs continously when a client has connected to a session. Whenever the client
         * responds to an alive request, this is registered. If a client fails to respond to
         * more than the number of requests specified in <code>ALLOWED_PACKETLOSS</code>,
         * the client is disconnected.
         */
        public void run() {
            this.running = true;
            long nextAliveRequest = System.currentTimeMillis() + ALIVE_REQUEST_INTERVAL;

            while (this.running) {
                if (System.currentTimeMillis() > nextAliveRequest) {
                    checkAlive(aliveRequestId);
                    sendAction(Action.alive_request, new Object[][]{{aliveRequestId}}, false);
                    aliveRequestId++;
                    nextAliveRequest = System.currentTimeMillis() + ALIVE_REQUEST_INTERVAL;
                }
            }
        }

        /**
         * Loops through the list of connected players and checks if any have passed
         * the limit for allowed loss of alive requests. If so, the players are removed
         * from the session, and all clients are notified of the disconnection.
         *
         * @param aliveRequestId The id of the last alive request that was sent.
         */
        public void checkAlive(int aliveRequestId) {
            List<P> timedOut = new ArrayList<P>();
            for (P player : players) {
                if (player.lastAliveConfirmation() < aliveRequestId - ALLOWED_PACKETLOSS) {
                    Utils.printTimeStampedInfo(player.getName() + " timed out and was removed from " + getName(), true);
                    timedOut.add(player);
                }
            }
            for (P player : timedOut) {
                Object toSend = server.getParser().createMessage(Action.disconnect, null, null);
                player.sendMessage(toSend, false);
                server.handleDisconnect(player);
            }
        }

        public void stop() {
            this.running = false;
        }
    }
}