package server.framework.models;

import java.util.ArrayList;
import java.util.HashMap;

/**
 * This class contains a list of the settings applying to a session, and is used by the
 * session's <code>GameThread</code> to ensure the validity of the game's current state.
 *
 * When settings are added to a game, this class needs to have the settings added.
 *
 * @version 1.0
 * @author Eivind Sorteberg
 * @author Martin Jarrett
 */
public class SettingsList {
    // The list of settings
    private HashMap<Setting, Object> settings;

    /**
     * Initialises the list of settings.
     */
    public SettingsList() {
        this.settings = new HashMap<Setting, Object>();
        intialiseSettings();
    }

    /**
     * Adds settings to the list of settings, along
     * with the settings' default values.
     */
    private void intialiseSettings() {
        this.settings.put(Setting.score, 5);
        this.settings.put(Setting.time, 0);
        this.settings.put(Setting.teams, false);
        this.settings.put(Setting.nof_players, 8);
    }

    /**
     * @return An array containing the settings' names and values.
     */
    public Object[][] getSettingsList() {
        ArrayList<Object[]> values = new ArrayList<Object[]>();
        for (Setting setting : this.settings.keySet()) {
            values.add(new Object[]{setting.getText(), settings.get(setting)});
        }

        return values.toArray(new Object[values.size()][]);
    }

    /**
     * @return The settings list.
     */
    public HashMap<Setting, Object> getSettings() {
        return this.settings;
    }

    /**
     * Method used to get a setting's value.
     *
     * @param setting The setting.
     * @return The specified setting's value.
     */
    public Object getSetting(Setting setting) {
        return this.settings.get(setting);
    }

    /**
     * Method that is called whenever a player has changed the
     * value of one or more settings. Updates the relevant setting's values.
     *
     * @param values The names and values of the changed settings.
     */
    public void fireSettingChanged(String[][] values) {
        for (String[] setting : values) {
            Setting settingId = Setting.getSetting(setting[0]);
            Object value = getSettingType(setting[1], settingId);
            settings.put(settingId, value);
        }
    }

    private Object getSettingType(String value, Setting settingId) {
        Object oldValue = settings.get(settingId);
        if (oldValue instanceof Integer) {
            return Integer.parseInt(value);
        }
        else if (oldValue instanceof String) {
            return value;
        }
        else if (oldValue instanceof Boolean) {
            return "true".equals(value);
        }
        else {
            return null;
        }
    }
}
