package server.framework.network;

import server.framework.models.AbstractServer;

import java.util.Deque;
import java.util.LinkedList;

/**
 * The server implementation of the <code>Communicator</code> interface. Each
 * player object has an assigned <code>ServerSocketCommunicator</code> object,
 * that is used to communicate with the associated client. This class uses
 * the player's <code>SocketWrapper</code> object to determine how messages
 * are sent and received.
 *
 * @version 1.0
 * @author Martin Jarrett
 * @author Eivind Sorteberg
 */
public class ServerSocketCommunicator implements Communicator {
    protected final SenderThread senderThread;
    protected ReceiverThread receiverThread;
    protected SocketWrapper socket;
    protected AbstractServer server;

    /**
     * Creates a new <code>ServerSocketCommunicator</code> instance, and
     * initialises the sender and receiver threads.
     *
     * @param socket The <code>SocketWrapper</code> to use for the communication.
     * @param server The server model to notify when messages are received.
     */
    public ServerSocketCommunicator(SocketWrapper socket, AbstractServer server) {
        this.server = server;
        this.socket = socket;
        this.senderThread = new SenderThread();
        this.receiverThread = new ReceiverThread();

        new Thread(this.senderThread).start();
        new Thread(this.receiverThread).start();
    }

    /**
     * Not implemented, since this is the server implementation.
     *
     * @param addresses Not used.
     * @return Not used.
     */
    public boolean connect(String[] addresses) {
        return false;
    }

    /**
     * Not implemented.
     *
     * @param adresses Not used.
     */
    public void disconnect(String[] adresses) {
    }

    /**
     * Notifies the sender thread that a new message is to be sent.
     *
     * @param message The message to send.
     * @param confirm Not used in this implementation.
     */
    public void sendMessage(Object message, boolean confirm) {
        this.senderThread.sendMessage(message.toString());
    }

    /**
     * Method that is called when the receiver thread receives a new message.
     * When this happens, the server model is notified about the event.
     *
     * @param message The message that has been received.
     */
    public void notifyAboutMessageReceived(Object message) {
        this.server.notifyAboutMessageReceived(message, this.socket);
    }

    /**
     * Closes the connection.
     */
    public void close() {
        this.socket.close();
    }

    /**
     * Not implemented.
     */
    public void searchForNodes() {
    }

    /**
     * @return The <code>SocketWrapper</code> object associated with this communicator.
     */
    public SocketWrapper getSocket() {
        return this.socket;
    }

    /**
     * Thread that is responsible for sending messages to the socket's associated
     * client.
     */
    protected class SenderThread implements Runnable {
        // How long the thread sleeps between each send
        private static final int SEND_SLEEP = 5;

        // The queue of messages to send
        private Deque<String> queue;
        private boolean running = true;

        /**
         * Initialises the message queue.
         */
        public SenderThread() {
            this.queue = new LinkedList<String>();
        }

        /**
         * Sends messages that are waiting to be sent. When the send queue is empty,
         * this thread sleeps until new messages to be sent arrive.
         *
         * The thread uses the associated <code>SocketWrapper</code> object to
         * send the messages, and the implementation of the sending is therefore
         * abstracted from this method.
         */
        public void run() {
            this.running = true;
            while (running) {
                synchronized (this) {
                    try {
                        while (!queue.isEmpty()) {
                            String message = queue.getFirst();
                            socket.sendMessage(message);
                            queue.removeFirst();
    //                        System.out.println("Sent:" + message);
                            Thread.sleep(SEND_SLEEP);
                        }
                        this.wait();
                    }
                    catch (InterruptedException e) {
                        e.printStackTrace();
                    }
                }
            }
        }

        /**
         * Called when a new message is to be sent. Adds the message to the
         * message queue, and notifies the sender thread that messages are
         * waiting to be sent.
         *
         * @param message The message to be sent.
         */
        public void sendMessage(String message) {
            this.queue.add(message);
            synchronized (this) {
                this.notify();
            }
        }

        public void stop() {
            this.running = false;
        }
    }

    /**
     * Thread responsible for listening to the connection, and detecting
     * when new messages arrive. When this happens, the thread calls the
     * <code>notifyAboutMessageReceived()</code>.
     *
     * Between each loop in the listen thread, the thread sleeps for an
     * amount of time specified by the <code>SLEEP_TIME</code>
     * variable, to avoid occupying too much resources.
     */
    protected class ReceiverThread implements Runnable {
        public static final int SLEEP_TIME = 10;

        private boolean running = true;

        public ReceiverThread() {
        }

        /**
         * The loop listening for incoming messages on the connection.
         *
         * The thread uses the associated <code>SocketWrapper</code> object to
         * receive the messages, and the implementation of the receival is therefore
         * abstracted from this method.
         */
        public void run() {
            this.running = true;
            while (socket == null) { /* Wait for socket */}

            while (this.running) {
                String message = socket.readMessage(SLEEP_TIME);
                notifyAboutMessageReceived(message);
            }
        }

        public void stop() {
            this.running = false;
        }

        /**
         * Sets the message size used for the UDP communication.
         *
         * @param messageSize The message size.
         */
        public void setMessageSize(int messageSize) {
            socket.setMessageSize(messageSize);
        }
    }
}

