package server.framework.network;

import server.framework.models.AbstractPlayer;
import server.framework.network.parser.StringParser;

import java.io.DataInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.net.Socket;
import java.net.SocketAddress;

/**
 * A wrapper class that is used to provide common methods for sending and receiving
 * messages independent of whether TCP or UDP is used for the message transmission.
 *
 * Each connected client has a unique <code>SocketWrapper</code> instance, that is
 * identified by its <code>remoteAddress</code> variable.
 *
 * @version 1.0
 * @author Martin Jarrett
 * @author Eivind Sorteberg
 */
public class SocketWrapper {
    // The socket object; contents depend on whether TCP or UDP is used.
    private Object socket;

    // Used for Socket
    private OutputStreamWriter writer;
    private InputStreamReader reader;

    // Used for DatagramSocket
    private SocketAddress remoteAddress;
    private AbstractPlayer player;
    private int messageSize;

    /**
     * Creates a <code>SocketWrapper</code> instance. If the socket parameter
     * is a <code>Socket</code> object, this object will use TCP for its communication.
     * If the socket is a <code>DatagramSocket</code>, it will use UDP.
     *
     * @param socket The socket determining what kind of transport protocol to use.
     */
    public SocketWrapper(Object socket) {
        this.socket = socket;
        this.initialize();
    }

    /**
     * Sets the address of the associated client.
     *
     * @param remoteAddress The address of the associated client.
     */
    public void setRemoteAddress(SocketAddress remoteAddress) {
        this.remoteAddress = remoteAddress;
    }

    /**
     * @return The address of the associated client.
     */
    public SocketAddress getRemoteAddress() {
        return remoteAddress;
    }

    /**
     * Initialises the necessary writer and reader objects if TCP
     * is to be used for the transmission.
     */
    private void initialize() {
        this.messageSize = StringParser.MAXIMUM_MESSAGE_SIZE;
        if (this.socket instanceof Socket) {
            try {
                this.writer = new OutputStreamWriter(((Socket) this.socket).getOutputStream());

                DataInputStream dis = new DataInputStream(((Socket) this.socket).getInputStream());
                this.reader = new InputStreamReader(dis);
            }
            catch (IOException e) {
                e.printStackTrace();
            }
        }
    }

    /**
     * Checks whether two <code>SocketWrapper</code> objects are equal. This method
     * is used to determine the sender of received messages.
     *
     * @param other The <code>SocketWrapper</code> that received the message.
     * @return Whether this object's player was the sender of the message or not.
     */
    public boolean equals(SocketWrapper other) {
        if (this.socket instanceof Socket) {
            return ((Socket)this.getSocket()).getRemoteSocketAddress().equals(((Socket)other.getSocket()).getRemoteSocketAddress());
        }
        else if (this.socket instanceof DatagramSocket) {
            return this.player.getAddress().equals(other.getRemoteAddress());
        }
        else {
            System.err.println("Illegal socket value");
            return false;
        }
    }

    /**
     * @return The socket object.
     */
    public Object getSocket() {
        return this.socket;
    }

    /**
     * Sends a message in the specified format. The method for sending the message
     * depends on what kind of socket the socket object is. However, since both
     * methods are contained in this message, the differences in implementation
     * is abstracted from the users of this method.
     *
     * @param message The message to send
     */
    public void sendMessage(String message) {
        if (this.socket instanceof Socket) {
            try {
                this.writer.write(message);
                this.writer.flush();
            }
            catch (IOException e) {
                //e.printStackTrace();
            }
        }
        else if (this.socket instanceof DatagramSocket) {
            try {
                DatagramPacket packet = new DatagramPacket(message.getBytes(), message.length(), this.player.getAddress());
                ((DatagramSocket) this.socket).send(packet);
            }
            catch (IOException e) {
                e.printStackTrace();
            }
        }
    }

    /**
     * Listens for incoming messages. As with the previous method, how this is performed
     * depends on the transport protocol in use. But since both methods are implemented
     * in this method, this is abstracted from the users of this method.
     *
     * @param sleeptime How long the tread sleeps between each time it checks the input stream (TCP only).
     * @return A received message.
     */
    public String readMessage(int sleeptime) {
        if (this.socket instanceof Socket) {
            try {
                StringBuilder message = new StringBuilder();
                while (!reader.ready()) {
                    try {
                        Thread.sleep(sleeptime);
                    }
                    catch (InterruptedException e) {
                        e.printStackTrace();
                    }
                }
                while (true) {
                    char character = (char) reader.read();
                    if ('|' == character) break;
                    message.append(character);
                }
                return message.toString();
            }
            catch (IOException e) {
                e.printStackTrace();
            }
        }
        else if (this.socket instanceof DatagramSocket) {
            try {
                byte[] buffer = new byte[messageSize];
                DatagramPacket packet = new DatagramPacket(buffer, buffer.length);
                ((DatagramSocket) socket).receive(packet);
                this.setRemoteAddress(packet.getSocketAddress());
                StringBuilder message = new StringBuilder();
                for (byte b : buffer) {
                    char character = (char)b;
                    if ('|' == character) break;    // Must be changed if the server's messageparser is changed
                    message.append(character);
                }
                return message.toString();
            }
            catch (IOException e) {
                e.printStackTrace();
            }
        }
        return null;
    }

    /**
     * @return Whether the connection is established or not.
     */
    public boolean isConnected() {
        return this.socket != null;
    }

    /**
     * Closes the connection when this is needed (TCP only)
     */
    public void close() {
        if (this.socket instanceof Socket) {
            try {
                ((Socket) this.socket).close();
            }
            catch (IOException e) {
                e.printStackTrace();
            }
        }
        else if (this.socket instanceof DatagramSocket) {
//            ((DatagramSocket)this.socket).close();
        }
    }

    /**
     * Sets the player that "owns" this <code>SocketWrapper</code>.
     *
     * @param player The owner of this object.
     */
    public void setPlayer(AbstractPlayer player) {
        this.player = player;
    }

    /**
     * Only used when running tests
     * 
     * @param messageSize The expected size of messages to receive
     */
    public void setMessageSize(int messageSize) {
        if (this.getSocket() instanceof DatagramSocket) {
            this.messageSize = messageSize;
        }
    }
}
