package server.framework.network.parser;

import server.framework.models.AbstractPlayer;

/**
 * An implementation of the <code>MessageParser</code> interface, that
 * creates and parses <code>String</code> messages. The format of these messages
 * is <action>":"[<player>"@"][<value>{","<value>}{";"<value>{","<value>}}]"|" (in EBNF-notation).
 *
 * @version 1.0
 * @author Eivind Sorteberg
 * @author Martin Jarrett
 */
public class StringParser implements MessageParser<String> {
    public static final int MAXIMUM_MESSAGE_SIZE = 64;

    // The different separators defined for the format
    private static final char SEP_ACTION = ':';
    private static final char SEP_SENDER = '@';
    private static final char SEP_VALUEPAIR = ';';
    private static final char SEP_VALUE = ',';
    private static final char END_OF_MESSAGE = '|';

    /**
     * Creates a string message in the specified format
     * from the given action identifier and values.
     *
     * @param action The action identifier.
     * @param values An array containing the message's values.
     * @param sender The sender of the action message.
     * @return A string message to send to the server.
     */
    public String createMessage(Action action, Object[][] values, AbstractPlayer sender) {
        String message = action.getName() + SEP_ACTION + getSenderString(sender);
        if (values != null && values.length > 0) {
            for (Object[] component : values) {
                for (Object attribute : component) {
                    message += attribute.toString() + SEP_VALUE;
                }
                // Exchange the last separator with an identifier sign
                message = changeLastLetter(message, SEP_VALUEPAIR);
            }
            message = message.substring(0, message.length() - 1);
        }
        message += END_OF_MESSAGE;
        return message;
    }

    /**
     * Parses a given string message into an <code>ActionPair</code>
     * wrapper.
     *
     * @param message The string message to parse.
     * @return An <code>ActionTriplet</code> containing the message's contents.
     */
    public ActionPair parseMessage(String message) {
        try {
            int actionSeparator = message.indexOf(SEP_ACTION);
            Action action = Action.getActionObject(message.substring(0, actionSeparator));
            String[] components = message.substring(actionSeparator + 1).split(Character.toString(SEP_VALUEPAIR));
            String[][] values = new String[components.length][];
            for (int i = 0; i < components.length; i++) {
                values[i] = components[i].split(Character.toString(SEP_VALUE));
            }
            return new ActionPair(action, values);
        } catch (StringIndexOutOfBoundsException e) {
            // Happens sometimes when connecting to localhost, nothing critical
            return new ActionPair(Action.empty_message, null);
        }
    }

    private static String changeLastLetter(String message, char newCharacter) {
        message = message.substring(0, message.length() - 1) + newCharacter;
        return message;
    }

    private static String getSenderString(AbstractPlayer sender) {
        if (sender == null) {
            return "";
        }
        else {
            return new StringBuilder(Integer.toString(sender.getId())).append(SEP_SENDER).toString();
        }
    }
}