package server.framework.threads;

import server.framework.models.AbstractPlayer;
import server.framework.models.AbstractSession;

/**
 * Implementors of the <code>EventHandler</code> interface are responsible
 * for handling game events that are caused by the session itself and by
 * clients connected to the session. This interface extends the
 * <code>Runnable</code> interface, and should contain a loop that runs
 * continously thoughout a game, generating events.
 *
 * In addition, whenever external game events occur, these should be handled in
 * the implementation of this interface.
 *
 * @version 1.0
 * @author Martin Jarrett
 * @author Eivind Sorteberg
 */
public interface EventHandler<P extends AbstractPlayer, S extends AbstractSession> extends Runnable {
    /**
     * Sets the session to which this thread is assigned.
     *
     * @param session The owner of this thread.
     */
    public void setSession(S session);

    /**
     * Method that should be called whenever an external game event occurs. How this event
     * is handled is decided by the implementation of this method.
     *
     * @param player The player that caused the event.
     * @param event The event.
     * @param values The values related to the event.
     */
    public void fireEventOccured(P player, Event event, Object[] values);

    /**
     * A method that should be called by the thread itself continuously. Whenever
     * a setting limit is reached, this thread should notify its models, so that these can act
     * accordingly.
     */
    public void checkSettings();

    /**
     * Called when a game is over, and the thread should be stopped. This
     * method should make sure the thread releases all its resources.
     */
    public void stop();

    /**
     * Method that ensures that the thread can perform pre-game calculations
     * before the game is initiated. Examples of such calculations are player
     * position generations.
     *
     * @return Whether the thread has completed its pre-game calculations or not.
     */
    public boolean isReady();
}
