package server.framework.threads;

import server.framework.models.AbstractPlayer;
import server.framework.models.AbstractSession;
import server.framework.network.parser.Action;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;

/**
 * This thread is responsible for bundling player positions into one
 * big position message, instead of transmitting many small position messages.
 *
 * @version 1.0
 * @author Martin Jarrett
 * @author Eivind Sorteberg
 */
public class GameThread implements Runnable {
    // The sleep time between each position transmission
    public static final int POSITION_INTERVAL = 100;

    // The list of player positions
    private HashMap<Integer, Object[]> positions;

    // List of "frozen" players
    private HashMap<Integer, Long> frozen;

    private AbstractSession session;
    private boolean running;

    /**
     * Initiates the <code>GameThread</code>'s lists and assigns the thread's session.
     *
     * @param session The session owning the game thread.
     */
    public GameThread(AbstractSession session) {
        this.session = session;
        this.positions = new HashMap<Integer, Object[]>();
        this.frozen = new HashMap<Integer, Long>();
    }

    /**
     * The loop that is responsible for creating and sending position bundles
     * with regular intervals. The frequency of these message transmissions is
     * determined by the <code>POSITION_INTERVAL</code> variable.
     */
    public void run() {
        this.running = true;

        while (running) {
            releaseFrozen();
            sendPositions();
            try {
                Thread.sleep(POSITION_INTERVAL);
            } catch (InterruptedException e) {
                e.printStackTrace();
            }
        }
    }

    /**
     * Creates the position bundle, and notifies the session that the message
     * is ready to be transmitted to all connected clients.
     */
    private void sendPositions() {
        List<Object[]> toSend = getPositions(positions, frozen);
        if (toSend.size() > 0) {
            this.session.sendAction(Action.position, toSend.toArray(new Object[toSend.size()][]), false);
        }
        this.positions.clear();
    }

    /**
     * Checks if any frozen players are ready to be released.
     */
    private void releaseFrozen() {
        List<Integer> toRemove = new ArrayList<Integer>();
        long currentTime = System.currentTimeMillis();
        for (int player : frozen.keySet()) {
            if (currentTime > frozen.get(player)) {
                toRemove.add(player);
            }
        }
        for (int player : toRemove) {
            this.frozen.remove(player);
        }
    }

    /**
     * Used to get the positions of the players.
     * Collects the positions of the players not frozen, and puts them
     * in an array. These positions are then removed from the position list,
     * so that the positions are not transmitted more than once.
     *
     * @param objects The players and their positions.
     * @param frozen The players that are not allowed to move.
     * @return An array containing the positions of the players that are allowed to move.
     */
    private static synchronized List<Object[]> getPositions(final HashMap<Integer, Object[]> objects, final HashMap<Integer, Long> frozen) {
        HashMap<Integer, Object[]> copy = new HashMap<Integer, Object[]>();
        copy.putAll(objects);   // To avoid concurrent modification
        List<Object[]> toSend = new ArrayList<Object[]>();
        for (int player : copy.keySet()) {
            if (!frozen.containsKey(player)) {
                Object[] position = copy.get(player);
                toSend.add(new Object[]{player, position[0], position[1]});
            }
        }
        return toSend;
    }

    /**
     * Called when the game is over. Stops the position bundle generation.
     */
    public void stop() {
        this.running = false;
    }

    /**
     * Adds a position to the position list.
     *
     * @param sender The player.
     * @param position The position.
     */
    public synchronized void addPosition(AbstractPlayer sender, Object[] position) {
        this.positions.put(sender.getId(), position);
    }

    /**
     * "Freezes" a player to a position for a given duration. Position updates are not forwarded from the player
     * for the next <duration> milliseconds. Used when a player dies, to ensure that the next received position (near
     * the trap) is not forwarded.
     *
     * @param player The player to freeze in place.
     * @param duration The duration of the freeze.
     */
    public void freezePlayer(AbstractPlayer player, int duration) {
        this.frozen.put(player.getId(), System.currentTimeMillis() + duration);
    }

    /**
     * Test method used to test the performance of the bundle generation method.
     * Runs 1 000 000 tests with 50 player positions, and calculates the maximum
     * and average time the method uses to complete.
     *
     * @param args Empty.
     */
    public static void main(String[] args) {
        int nof_runs = 1000000;
        int nof_players = 50;
        long sum = 0, max = 0, t0, diff;
        for (int i = 0; i < nof_runs; i++) {
            HashMap<Integer, Object[]> pmap = new HashMap<Integer, Object[]>();
            HashMap<Integer, Long> frozen = new HashMap<Integer, Long>(1);
            for (int p = 0; p < nof_players; p++) {
                int x = (int)(Math.random() * 1000);
                int y = (int)(Math.random() * 1000);
                pmap.put(p, new Object[]{x, y});
            }
            t0 = System.nanoTime();
            GameThread.getPositions(pmap, frozen);
            diff = System.nanoTime() - t0;
            sum += diff;
            if (diff > max) max = diff;
        }
        System.out.println("avg = " + (double)sum/(double)nof_runs);
        System.out.println("max = " + max);
    }
}
