package server.testmodules;

import server.framework.models.AbstractServer;
import server.framework.network.SocketWrapper;

/**
 * This test module tests the response time of the networks by sending very
 * small packets (4 bytes) with increasing intervals between each packet. When a set of packets
 * has been sent, the interval is increased with the amount specfied in the
 * <code>INTERVAL_INC</code> variable.
 *
 * The test results from this test are calculated by finding the minimum, maximum and
 * average response times for each interval set.
 *
 * @version 1.0
 * @author Martin Jarrett
 * @author Eivind Sorteberg
 */
public class PingTest extends TestModule {
    // The increasement of the intervals
    public static final int INTERVAL_INC = 50;

    // The initial interval between the packets
    public static final int LATENCY_START = 0;

    /**
     * Creates a ping test instance by calling the super constructor.
     *
     * @param socket The socket to use for the transmission.
     * @param server The server model.
     */
    public PingTest(SocketWrapper socket, AbstractServer server) {
        super(socket, server);
    }

    /**
     * @return The headers used for the HTML table generated by the test.
     */
    protected String[] getHeaders() {
        return new String[]{
                "Interval",
                "Latency (s)",
                "Max",
                "Min",
                "Sum"
        };
    }

    /**
     * Method that calculates the minimum, maximum and average response time
     * for the specified interval number.
     *
     * @param timers The set of measured response times for each packet.
     * @param interval_no The interval for which to calculate the values.
     * @return The minimum, maximum and average response time.
     */
    @SuppressWarnings({"PointlessArithmeticExpression"})
    protected Number[] calculateValues(Timer[] timers, int interval_no) {
        long interval = LATENCY_START + interval_no / NOF_PACKETS * INTERVAL_INC;
        long min = Long.MAX_VALUE;
        long max = 0;
        long avg = 0;
        long sum;
        long nofValid = 0;

        for (Timer timer : timers) {
            long time = timer.calculateTime();
            if (time == 0) System.err.println("Empty time");
            if (time < min) {
                min = time;
            }
            if (time > max) {
                max = time;
            }
            if (time > 0) {
                avg += time;
                nofValid++;
            }
        }
        avg -= (max + min);
        avg = nofValid > 2 ? avg / (nofValid - 2) : -1;
        sum = interval * 1000000 + avg;
        return new Number[]{interval, avg, max, min, sum};
    }

    /**
     * @return The title for the HTML file generated by the test.
     */
    protected String getTitle() {
        return "Ping tests";
    }

    /**
     * Creates the string to send.
     *
     * @param id The id of the packet.
     * @return A three digit string followed by a '|' character
     */
    protected String getStringToSend(int id) {
        return getThreeDigitString(id) + '|';
    }

    /**
     * Called when one interval has been completed. The interval between
     * the transmissions is then increased with the value specified in
     * <code>INTERVAL_INC</code>.
     */
    protected void notifyAboutIntervalComplete() {
        this.interval += INTERVAL_INC;
    }

    /**
     * @return The interval to use for the first set of transmissions.
     */
    protected long getInitialInterval() {
        return LATENCY_START;
    }

    /**
     * The first message to send.
     *
     * @return An empty string.
     */
    protected String getInitialMessage() {
        return "";
    }


    /**
     * Since the messages in this test only consist of the
     * id and the end character, the id string is simply the message.
     *
     * @param message The message that was received.
     * @return The message.
     */
    protected String findIdString(String message) {
        return message;
    }
}