package server.testmodules;

import server.framework.models.AbstractPlayer;
import server.framework.models.AbstractServer;
import server.framework.models.AbstractSession;
import server.framework.network.Communicator;
import server.framework.network.ServerSocketCommunicator;
import server.framework.network.SocketWrapper;
import server.brickblock.fwimpl.BBSession;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.HashMap;

/**
 * An abstract class containing common methods for test-classes.
 * Such methods can be measurement and report generating methods.
 *
 * @author Martin Jarrett
 * @author Eivind Sorteberg
 * @version 1.0
 */
public abstract class TestModule extends ServerSocketCommunicator {

    /**
     * Enumeration determining the test type.
     */
    public enum TestType {
        ping,
        speed;

        /**
         * Gets the test type.
         *
         * @param name The name to of the test that is unknown.
         * @return The test type of the test.
         */
        public static TestType getTestType(String name) {
            for (TestType type : values()) {
                if (type.name().equals(name)) {
                    return type;
                }
            }
            return null;
        }

        /**
         * Creates the <code>TestModule</code> type.
         *
         * @param type The test type.
         * @param socket The socket used in the test.
         * @param server The server the test is run on.
         * @return The type of <code>TestModule</code>  created.
         */
        public Communicator createTestModule(TestType type, SocketWrapper socket, AbstractServer server) {
            switch (type) {
                case ping:
                    return new PingTest(socket, server);
                case speed:
                    return new TransferSpeedTest(socket, server);
                default:
                    System.err.println("Illegal testtype: " + type);
                    System.exit(0);
                    return null;
            }
        }
    }

    // The number of packages sent with each interval
    public static final int NOF_PACKETS = 10;

    // The number of different intervals tested
    public static final int NOF_INTERVALS = 10;

    // The number of times the test is run
    public static final int NOF_RUNS = 5;

    // Text appended to the file name of the report to indicate the service provider
    public static final String PROVIDER = "telenor";

    // Text appended to the file name of the report to indicate the mobile network
    public static final String NETWORK = "edge";

    // Random number between 0 and 100, appended to the file name of the report to create unique file names
    public int random = (int)(Math.random()*100);

    //The string sent between server and client which is timed
    protected String message;

    // The counter of the number of times the test has completed a run
    private int counter = 0;

    protected ArrayList<HashMap<Integer, Timer>> allTests;

    // The variable indicating the currentRun of the test
    protected int currentRun;

    //The interval between each send from the server
    protected long interval;

    /**
     * Creates a <code>TestModule</code> object with the specified
     * attributes. Calls the parent constructor and initialize the
     * ArrayList of the tests and the currentRun to zero.
     *
     * @param socket The test client's <code>SocketWrapper</code> object.
     * @param server The test server.
     */
    public TestModule(SocketWrapper socket, AbstractServer server) {
        super(socket, server);
        this.allTests = new ArrayList<HashMap<Integer, Timer>>(NOF_RUNS);
        for (int i = 0; i < NOF_RUNS; i++) {
            allTests.add(new HashMap<Integer, Timer>());
        }
        this.currentRun = 0;
    }

    /**
     * Uses the parent class implementation.
     *
     * @param addresses See parent class.
     * @return See parent class.
     */
    public boolean connect(String[] addresses) {
        return super.connect(addresses);
    }

    /**
     * Uses the parent class implementation.
     *
     * @param addresses See parent class.
     */
    public void disconnect(String[] addresses) {
        super.disconnect(addresses);
    }

    /**
     * Notifies the sender thread that a new message is to be sent.
     *
     * @param message The message to send.
     * @param confirm Not used in this implementation.
     */
    public void sendMessage(Object message, boolean confirm) {
        super.sendMessage(message, confirm);
    }

    /**
     * Closes the connection.
     */
    public void close() {
        super.close();
    }

    /**
     * Uses the parent class implementation.
     */
    public void searchForNodes() {
    }

    /**
     *  Generates and returns the test report from the test results.
     *
     * @param printAllTimers Whether all results should be printed or just the main results.
     * @return The report as a String.
     */

    public String getReport(boolean printAllTimers) {
        String report =
                    "<html>\n" +
                    "   <head>\n" +
                    "      <link rel='stylesheet' type='text/css' href='style.css' />\n" +
                    "   </head>\n" +
                    "   <body>\n"  +
                    "   <h2>" + getTitle() + "</h2>\n";


        for (HashMap<Integer, Timer> runInfo : allTests) {
            report +=
                    "       <table>\n" +
                    "           <tr>\n";

            for (String header : getHeaders()) {
                report +=
                    "               <th>" + header + "</th>\n";
            }
            if (printAllTimers) {
                report +=
                    "               <th colspan='" + NOF_PACKETS + "'>All timers</th>\n";
            }

            report +=
                    "           </tr>\n";

            for (int key = 0; key < runInfo.size(); key += NOF_PACKETS) {
                Timer[] timers = new Timer[NOF_PACKETS];
                for (int i = 0; i < NOF_PACKETS; i++) {
                    timers[i] = runInfo.get(key + i);
                }
                Number[] values = calculateValues(timers, key);

                report +=
                    "           <tr>\n" +
                    "               <td class='id'>" + values[0] + "</td>\n";
                for (int i = 1; i < values.length; i++) {
                    if (values[i] instanceof Long) {
                        long value = (Long)values[i];
                        if (value >= 0) {
                            report += "                <td class='number'>" + getTimeInSeconds(value) + "</td>\n";
                        }
                        else {
                            report += "                <td class='lost'>N/A</td>\n";
                        }
                    }
                    else {
                        report += "                <td class='number'>" + values[i] + "</td>\n";
                    }
                }
                if (printAllTimers) {
                    long[] allTimers = getAllValues(timers);
                    for (long l : allTimers) {
                        report += "                <td class='number'>" + getTimeInSeconds(l) + "</td>\n";
                    }
                }
            }
            report +=
                    "           </tr>\n" +
                    "       </table>\n";
        }
        return report +=
                    "   </body>\n" +
                    "</html>";

    }

    /**
     * Method that is called when the receiver thread receives a new message.
     * When this happens, the server model is notified about the event.
     *
     * @param message The message that has been received.
     */
    @SuppressWarnings(value = {"unchecked"})
    public void notifyAboutMessageReceived(Object message) {
        try {
            AbstractPlayer sender = server.findPlayer(this.socket);
            if (sender == null) {
                sender = server.createPlayer(this.socket);
            }
            int sessionId = sender.getSessionId();
            if (sessionId < 0) {
                AbstractSession session = new BBSession(server, server.getTestType());
                session.addPlayer(sender);
            }

            String m = message.toString();
            if (Character.isDigit(m.charAt(0))) {
                m = findIdString(m);

                if (this.socket.equals(sender.getSocket())) {
                    int id = Integer.parseInt(m);

                    HashMap<Integer, Timer> currentRunInfo =
                            allTests.get(id <= getLastId() ? currentRun : currentRun - 1);
                    Timer timer = currentRunInfo.get(id);
                    timer.setReceive(System.nanoTime());
                    currentRunInfo.put(id, timer);
                }
            }
        }
        catch (NumberFormatException e) {
            System.err.println("Received: " + message);
        }
    }

    /**
     * Finds the id of a message.
     *
     * @param message The message to find an id from.
     * @return The id as a String.
     */
    protected abstract String findIdString(String message);

    /**
     * Calculates values from the Timers.
     *
     * @param timers The Timers to calculate from.
     * @param interval_no The interval number to calculate.
     * @return The array of calculated values.
     */
    protected abstract Number[] calculateValues(Timer[] timers, int interval_no);

    /**
     * Gets all values from the <code>Timer</code>s'.
     *
     * @param timers The array of <code>Timer</code>s to retrived the values from.
     * @return The array of values.
     */
    protected long[] getAllValues(Timer[] timers) {
        long[] values = new long[timers.length];
        for (int i = 0; i < values.length; i++) {
            values[i] = timers[i].calculateTime();
        }
        return values;
    }

    /**
     * Starts the test and initialize import variables.
     * The test is started by creating a new Sender thread.
     */
    public void runTest() {
        this.counter = 0;
        this.message = getInitialMessage();
        this.interval = getInitialInterval();
        new Thread(new Sender()).start();
    }

    /**
     * Notifies the <code>TestModule</code> that the run is completed
     * and the next run can be started. If the last run is completed,
     * the programs sleeps to ensure that the last messages arrive before
     * the report is printed.
     */
    protected void notifyAboutRunComplete() {
        if (currentRun+1 != NOF_RUNS) {
            currentRun++;
            System.out.println("RERUNNING: " + currentRun);
            runTest();
        }
        else {

            try {
                Thread.sleep(10000);
            }
            catch (InterruptedException e) {
                e.printStackTrace();
            }

            String report = getReport(true);

            try {
                System.out.println("Printing report");
                PrintWriter out = new PrintWriter(new BufferedWriter(new FileWriter(server.getTestType() + "_" + PROVIDER + "_" + NETWORK + "_" + server.getCommType() + "_" + random + ".html", false)));
                out.write(report);
                out.close();

            }
            catch (IOException e) {
                e.printStackTrace();
            }
        }
    }

    /**
     * Generates an id from the counter before the counter is incremented.
     *
     * @return An id to be used in the message.
     */
    public int getId() {
        int returnValue = counter;
        counter++;
        return returnValue;
    }

    /**
     * Gets the previous id.
     *
     * @return The previous id.
     */
    public int getLastId() {
        return counter - 1;
    }

    /**
     * Converts from nanoseconds to seconds.
     *
     * @param nanos The amount of nanoseconds to convert.
     * @return The amount of seconds as a String.
     */
    protected static String getTimeInSeconds(long nanos) {
        long seconds = nanos / 1000000000;
        long millis = (nanos - seconds * 1000000000) / 1000000;

        return seconds + "." + getThreeDigitString(millis);
    }

    /**
     * Converts from a number to a string of three digits by filling in
     * zeros before the number. Two zeros are added if the number is lower
     * than 10 (converts from 7 to 007) and one zero is added if the number
     * is between 10 and 100 (converts from 13 to 013).
     *
     * @param number The number to be converted.
     * @return The converted three digit String.
     */
    protected static String getThreeDigitString(long number) {
        return number < 10 ? ("00" + number) : ((number < 100) ? ("0" + number) : number + "");
    }

    /**
     * Notifies the <code>TestModules</code> that the interval is complete.
     */
    protected abstract void notifyAboutIntervalComplete();

    /**
     * Gets the initial interval to use between sends.
     *
     * @return The initial interval.
     */
    protected abstract long getInitialInterval();

    /**
     * Gets the initial message to send.
     *
     * @return The initial message.
     */
    protected abstract String getInitialMessage();

    /**
     * Gets the headers for the coloumns in the report.
     *
     * @return The coloumn headers.
     */
    protected abstract String[] getHeaders();

    /**
     * Gets the title for the report.
     *
     * @return The report title.
     */
    protected abstract String getTitle();

    /**
     * The internal class that times the time a message takes
     * to go from the server to the client and back.
     */
    protected class Timer {

        // The send timestamp.
        private long send;

        // The receival timestamp.
        private long receive;

        /**
         * Creates a <code>Timer</code> object with a send timestamp.
         *
         * @param send The timestamp of which the message was sent.
         */
        public Timer(long send) {
            this.send = send;
        }

        /**
         * Sets the receival time of the message.
         *
         * @param receive The timestamp of which the message was received.
         */
        public void setReceive(long receive) {
            this.receive = receive;
        }

        /**
         * Calculates the time the message took from the server to the client and back.
         *
         * @return The transmission time.
         */
        public long calculateTime() {
            return this.receive - this.send;
        }

        /**
         * Convert the send and receival timestamps to a String.
         *
         * @return The timestamp String.
         */
        public String toString() {
            return "[" + this.send + ", " + this.receive + "]";
        }
    }

    /**
     * The internal class responsible for sending messages from the server.
     */
    protected class Sender implements Runnable {

        /**
         * The run method that generates the message to be sent,
         * starts a <code>Timer</code>, sends the message, and completes
         * the test.
         */
        public void run() {
            int id = 0;
            HashMap<Integer, Timer> currentRunInfo = allTests.get(currentRun);
            while (id < (NOF_INTERVALS * NOF_PACKETS - 1)) {
                id = getId();
                String toSend = getStringToSend(id);
                receiverThread.setMessageSize(toSend.length());
                currentRunInfo.put(id, new Timer(System.nanoTime()));

                senderThread.sendMessage(toSend);

                if (id % NOF_PACKETS == NOF_PACKETS - 1) {
                    synchronized(this) {
                        notifyAboutIntervalComplete();
                    }
                }

                if (id != NOF_INTERVALS * NOF_PACKETS - 2) {
                    try {
                        Thread.sleep(interval);
                    }
                    catch (InterruptedException e) {
                        e.printStackTrace();
                    }
                }
            }
            try {
                Thread.sleep(500);
            }
            catch (InterruptedException e) {
                e.printStackTrace();
            }
//            allTests.add(currentRunInfo);
            notifyAboutRunComplete();
        }
    }

    /**
     * Gets the String to send.
     *
     * @param id The id of the message to send.
     * @return The String to send.
     */
    protected abstract String getStringToSend(int id);
}
