package server.testmodules;

import server.framework.models.AbstractServer;
import server.framework.network.SocketWrapper;

/**
 * This test module tests the transfer speed of the networks by sending packets
 * with increasing sizes. When a set of packets
 * has been sent, the size of the packets is increased with the amount specfied in the
 * <code>SIZE_INC</code> variable.
 *
 * The test results from this test are calculated by finding the minimum, maximum and
 * average transfer times for each size set, and calculating the transfer speed based
 * on the average.
 *
 * @version 1.0
 * @author Martin Jarrett
 * @author Eivind Sorteberg
 */
public class TransferSpeedTest extends TestModule {
    // The character separating the id from the rest of the message
    public static final char ID_CHAR = '!';

    // The initial size of the messages
    public static final int INITIAL_SIZE = 0;

    // How much the size is incremented each time a set is complete
    public static final int SIZE_INC = 10;

    // The interval between each transmission
    public static final int LATENCY_START = 200;

    /**
     * Creates a transfer speed test instance by calling the super constructor.
     *
     * @param socket The socket to use for the transmission.
     * @param server The server model.
     */
    public TransferSpeedTest(SocketWrapper socket, AbstractServer server) {
        super(socket, server);
    }

    /**
     * @return The headers used for the HTML table generated by the test.
     */
    protected String[] getHeaders() {
        return new String[]{
            "Data sent (bits)",
            "Transfer time (s)",
            "Max Transfer time (s)",
            "Min Transfer time (s)",
            "Transfer speed (bits/s)"
        };
    }

    /**
     * Method that calculates the minimum, maximum and average transfer time
     * for the specified interval number. The transfer speed is then calculated
     * based on the average transfer time.
     *
     * @param timers The set of measured response times for each packet.
     * @param interval_no The interval for which to calculate the values.
     * @return The minimum, maximum and average response time.
     */
    @SuppressWarnings({"PointlessArithmeticExpression"})
    protected Number[] calculateValues(Timer[] timers, int interval_no) {
        long datasize = 5 * 8 + (interval_no / NOF_PACKETS) * SIZE_INC * 8 + INITIAL_SIZE * 8;
        long avg = 0;
        long min = Long.MAX_VALUE;
        long max = 0;
        long nofValid = 0;
        float speed;

        for (Timer timer : timers) {
            long time = timer.calculateTime();
            if (time == 0) System.err.println("Empty time");
            if (time < min) {
                min = time;
            }
            if (time > max) {
                max = time;
            }
            if (time >= 0) {
                avg += time;
                nofValid++;
            }
        }
        avg -= (max + min);
        avg = nofValid > 2 ? avg / (nofValid - 2) : -1;
        speed = datasize / ((float)avg / 1000000000);
        return new Number[]{datasize, avg, max, min, speed};
    }

    /**
     * @return The title for the HTML file generated by the test.
     */
    protected String getTitle() {
        return "Transfer speed tests";
    }

    /**
     * Called when one size set has been completed. The size of the messages
     *  is then increased with the value specified in
     * <code>SIZE_INC</code>, and the correct number of 'x' characters
     * is added to the message to send.
     */
    protected void notifyAboutIntervalComplete() {
        for (int i = 0; i < SIZE_INC; i++) {
            message += 'x';
        }
    }

    /**
     * Returns the intital send interval between the messages. This value is constant
     * in this test and specified in LATENCY_START.
     *
     * @return LATENCY_START.
     */
    protected long getInitialInterval() {
        return LATENCY_START;
    }

    /**
     * Creates the initial message based on the INITIAL_SIZE value.
     *
     * @return The initial message to send.
     */
    protected String getInitialMessage() {
        StringBuilder builder = new StringBuilder();
        for (int i = 0; i < INITIAL_SIZE; i++) {
            builder.append('x');
        }
        return builder.toString();
    }

    /**
     * Creates the string to send by appending the specified number of
     * 'x' characters to the id and separator.
     *
     * @param id The id of the message.
     * @return A message on the form <id>'!'{'x'}
     */
    protected String getStringToSend(int id) {
        StringBuilder builder = new StringBuilder(getThreeDigitString(id));
        builder.append(ID_CHAR).append(this.message).append('|');
        return builder.toString();
    }

    /**
     * Extracts the id from the message by removing the x's and the '!'
     * character.
     *
     * @param message The message that was returned.
     * @return The id of the message.
     */
    protected String findIdString(String message) {
        return message.substring(0, message.indexOf(ID_CHAR));
    }
}